// Copyright 2015-2017 Espressif Systems (Shanghai) PTE LTD
//
// Copyright 2018 IS2T. This file has been modified by IS2T S.A.
// 1. Add SD card GPIO configuration
// 2. Fill ff_sdmmc_initialize() with SD card initialization code
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "diskio.h"
#include "ffconf.h"
#include "ff.h"
#include "sdmmc_cmd.h"
#include "esp_log.h"
#include "driver/sdmmc_host.h"
#include "driver/sdspi_host.h"

#define SD_CLK_GPIO    GPIO_NUM_14 // Shared with JTAG
#define SD_CMD_GPIO    GPIO_NUM_15 // Shared with JTAG
#define SD_D0_GPIO     GPIO_NUM_2  // Shared with LED/Camera
#define SD_D3_GPIO     GPIO_NUM_13 // Shared with JTAG/ECOM-COMM (SPI)

static sdmmc_card_t* s_cards[FF_VOLUMES] = { NULL };

static const char* TAG = "diskio_sdmmc";

sdmmc_card_t s_card;

DSTATUS ff_sdmmc_initialize (BYTE pdrv)
{
    esp_err_t err = ESP_OK;

    /*
     * Use 1-line mode, so ecom-comm (SPI) can be used at the same time, as the 4-line mode uses GPIO12 and GPIO13 which are
     * also used by the ecom-comm (SPI)
     */
    sdmmc_host_t host_config = SDMMC_HOST_DEFAULT();
    host_config.flags = SDMMC_HOST_FLAG_1BIT;
    host_config.max_freq_khz = SDMMC_FREQ_HIGHSPEED;

    /* This initializes the slot without card detect (CD) and write protect (WP) signals and in 1-line mode. */
    sdmmc_slot_config_t slot_config = SDMMC_SLOT_CONFIG_DEFAULT();
    slot_config.width = 1;

    /*
     * The below GPIOs should have external 10k pull-ups.
     * Internal pull-ups are not sufficient, but enabling them makes a difference on some boards
     */
    gpio_set_pull_mode(SD_CMD_GPIO, GPIO_PULLUP_ONLY);
    gpio_set_pull_mode(SD_D0_GPIO , GPIO_PULLUP_ONLY);
    gpio_set_pull_mode(SD_D3_GPIO , GPIO_PULLUP_ONLY); // Not used in 1-line SD mode, but we need to enable the internal pull-up

    /* Init the SD host */
    err = host_config.init();
    if (err != ESP_OK) {
        ESP_LOGD(TAG, "host init returned rc=0x%x", err);
        goto fail;
    }

    /* Configure SD slot */
    if (host_config.flags == SDMMC_HOST_FLAG_SPI) {
        err = sdspi_host_init_slot(host_config.slot,
                (const sdspi_slot_config_t*) &slot_config);
    } else {
        err = sdmmc_host_init_slot(host_config.slot,
                (const sdmmc_slot_config_t*) &slot_config);
    }
    if (err != ESP_OK) {
        ESP_LOGD(TAG, "slot_config returned rc=0x%x", err);
        goto fail;
    }

    /* Probe and initialize card */
    err = sdmmc_card_init(&host_config, &s_card);
    if (err != ESP_OK) {
        ESP_LOGD(TAG, "sdmmc_card_init failed 0x(%x)", err);
        goto fail;
    }

    return 0;

fail:
    host_config.deinit();
    return -1;
}

DSTATUS ff_sdmmc_status (BYTE pdrv)
{
    return 0;
}

DRESULT ff_sdmmc_read (BYTE pdrv, BYTE* buff, DWORD sector, UINT count)
{
    sdmmc_card_t* card = s_cards[pdrv];
    assert(card);
    esp_err_t err = sdmmc_read_sectors(card, buff, sector, count);
    if (err != ESP_OK) {
        ESP_LOGE(TAG, "sdmmc_read_blocks failed (%d)", err);
        return RES_ERROR;
    }
    return RES_OK;
}

DRESULT ff_sdmmc_write (BYTE pdrv, const BYTE* buff, DWORD sector, UINT count)
{
    sdmmc_card_t* card = s_cards[pdrv];
    assert(card);
    esp_err_t err = sdmmc_write_sectors(card, buff, sector, count);
    if (err != ESP_OK) {
        ESP_LOGE(TAG, "sdmmc_write_blocks failed (%d)", err);
        return RES_ERROR;
    }
    return RES_OK;
}

DRESULT ff_sdmmc_ioctl (BYTE pdrv, BYTE cmd, void* buff)
{
    sdmmc_card_t* card = s_cards[pdrv];
    assert(card);
    switch(cmd) {
        case CTRL_SYNC:
            return RES_OK;
        case GET_SECTOR_COUNT:
            *((DWORD*) buff) = card->csd.capacity;
            return RES_OK;
        case GET_SECTOR_SIZE:
            *((WORD*) buff) = card->csd.sector_size;
            return RES_OK;
        case GET_BLOCK_SIZE:
            return RES_ERROR;
    }
    return RES_ERROR;
}

void ff_diskio_register_sdmmc(BYTE pdrv, sdmmc_card_t* card)
{
    static const ff_diskio_impl_t sdmmc_impl = {
        .init = &ff_sdmmc_initialize,
        .status = &ff_sdmmc_status,
        .read = &ff_sdmmc_read,
        .write = &ff_sdmmc_write,
        .ioctl = &ff_sdmmc_ioctl
    };
    s_cards[pdrv] = card;
    ff_diskio_register(pdrv, &sdmmc_impl);
}

