/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <stdio.h>

#include "LLRCOMMAND_BLUETOOTH_impl.h"
#include "LLRCOMMAND_BLUETOOTH_internals.h"

#include "nvs.h"
#include "nvs_flash.h"
#include "controller.h"
#include "freertos/semphr.h"
#include "sni.h"


/* Macro definitions */
#define PROFILE_NUM      1
#define PROFILE_APP_ID   0

/* Type definitions */
typedef struct _device_cache_t {
    esp_bd_addr_t bda;
    esp_ble_addr_type_t ble_addr_type;
    struct _device_cache_t *next;
} device_cache_t;

typedef struct _gattc_profile_inst {
    esp_gattc_cb_t gattc_cb;
    uint16_t gattc_if;
    uint16_t app_id;
    uint16_t conn_id;
    uint16_t service_start_handle;
    uint16_t service_end_handle;
    uint16_t char_handle;
    esp_bd_addr_t remote_bda;
} gattc_profile_inst;

/* Static variables */
static device_cache_t *device_cache = NULL;

static SemaphoreHandle_t LLRCOMMAND_BLUETOOTH_set_param_sem;
static SemaphoreHandle_t LLRCOMMAND_BLUETOOTH_start_scan_sem;
static SemaphoreHandle_t LLRCOMMAND_BLUETOOTH_stop_scan_sem;
static SemaphoreHandle_t LLRCOMMAND_BLUETOOTH_notification_sem;

static uint8_t LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[LLRCOMMAND_BLUETOOTH_NOTIFICATION_BUFFER_SIZE];
static int32_t LLRCOMMAND_BLUETOOTH_immortal_notification_size = 0;
static uint8_t LLRCOMMAND_BLUETOOTH_notification_pending = 0;

static uint16_t LLRCOMMAND_BLUETOOTH_ble_connection_id = 0xFFFF;

static uint16_t used_mtu;

/* Static functions */
static void LLRCOMMAND_BLUETOOTH_notification_wakeup(void);
static void LLRCOMMAND_BLUETOOTH_notification_wait(void);
static void LLRCOMMAND_BLUETOOTH_add_device_to_cache(uint8_t *bda, esp_ble_addr_type_t ble_addr_type);
static void LLRCOMMAND_BLUETOOTH_get_device_from_cache(uint8_t *bda, esp_ble_addr_type_t *ble_addr_type);
static void LLRCOMMAND_BLUETOOTH_gap_cb(esp_gap_ble_cb_event_t event, esp_ble_gap_cb_param_t *param);
static void LLRCOMMAND_BLUETOOTH_event_handler(esp_gattc_cb_event_t event, esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param);
static void LLRCOMMAND_BLUETOOTH_gattc_cb(esp_gattc_cb_event_t event, esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param);

static esp_ble_scan_params_t ble_scan_params = {
    .scan_type              = BLE_SCAN_TYPE_ACTIVE,
    .own_addr_type          = BLE_ADDR_TYPE_PUBLIC,
    .scan_filter_policy     = BLE_SCAN_FILTER_ALLOW_ALL,
    .scan_interval          = 0x50,
    .scan_window            = 0x30
};

static gattc_profile_inst gl_profile_tab[PROFILE_NUM] = {
        [PROFILE_APP_ID] = {
                .gattc_cb = LLRCOMMAND_BLUETOOTH_event_handler,
                .gattc_if = ESP_GATT_IF_NONE,
        },
};


int32_t LLRCOMMAND_BLUETOOTH_IMPL_init(uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    vSemaphoreCreateBinary(LLRCOMMAND_BLUETOOTH_set_param_sem);
    xSemaphoreTake(LLRCOMMAND_BLUETOOTH_set_param_sem, 0);

    vSemaphoreCreateBinary(LLRCOMMAND_BLUETOOTH_start_scan_sem);
    xSemaphoreTake(LLRCOMMAND_BLUETOOTH_start_scan_sem, 0);

    vSemaphoreCreateBinary(LLRCOMMAND_BLUETOOTH_stop_scan_sem);
    xSemaphoreTake(LLRCOMMAND_BLUETOOTH_stop_scan_sem, 0);

    vSemaphoreCreateBinary(LLRCOMMAND_BLUETOOTH_notification_sem);

    llrcommand_bluetooth_current_state.is_connected = false;
    llrcommand_bluetooth_current_state.is_scanning = false;

    do {
        esp_bt_controller_config_t bt_cfg = BT_CONTROLLER_INIT_CONFIG_DEFAULT();

        api_status = esp_bt_controller_mem_release(ESP_BT_MODE_CLASSIC_BT);
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluetooth controller release classic bt memory failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_bt_controller_init(&bt_cfg);
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluetooth controller initialize failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_bt_controller_enable(ESP_BT_MODE_BLE);
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluetooth controller enable failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_bluedroid_init();
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluedroid init failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_bluedroid_enable();
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluedroid enable failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_ble_gap_register_callback(LLRCOMMAND_BLUETOOTH_gap_cb);
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] gap register failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_ble_gattc_register_callback(LLRCOMMAND_BLUETOOTH_gattc_cb);
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] gattc register failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_ble_gattc_app_register(PROFILE_APP_ID);
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] gattc app register failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_ble_gatt_set_local_mtu(ESP_GATT_MAX_MTU_SIZE);
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] esp_ble_gatt_set_local_mtu with error (err = %d)\n", api_status);
#endif
            break;
        }

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] init complete successfully\n");
#endif
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    } while (0);

    return status;
}

int32_t LLRCOMMAND_BLUETOOTH_IMPL_deinit(uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    do {
        api_status = esp_bluedroid_disable();
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluedroid disable failed with error (err = %d)\n",  api_status);
#endif
            break;
        }

        api_status = esp_bluedroid_deinit();
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluedroid de-init failed with error (err = %d)\n",  api_status);
#endif
            break;
        }

        api_status = esp_bt_controller_disable();
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluetooth controller disable failed with error (err = %d)\n", api_status);
#endif
            break;
        }

        api_status = esp_bt_controller_deinit();
        if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] Bluetooth controller de-initialize failed with error (err = %d)\n", api_status);
#endif
            break;
        }

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] De-init successfull\n");
#endif
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    } while (0);

    return status;
}

int32_t LLRCOMMAND_BLUETOOTH_IMPL_scanStart(uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    if (true == llrcommand_bluetooth_current_state.is_scanning) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[WARNING] Already scanning...\n");
#endif
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    } else {
        llrcommand_bluetooth_current_state.is_scanning = true;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] setting up the scan parameters...\n");
#endif
        do {
            api_status = esp_ble_gap_set_scan_params(&ble_scan_params);
            if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                printf("[ERROR] Set scan parameters failed with error: (err = %d)\n", api_status);
#endif
                break;
            } else {
                xSemaphoreTake(LLRCOMMAND_BLUETOOTH_set_param_sem, portMAX_DELAY);
            }

            api_status = esp_ble_gap_start_scanning(10);
            if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                printf("[ERROR] Scanning the peer devices failed with error: (err = %d)\n", api_status);
#endif
                break;
            } else {
                xSemaphoreTake(LLRCOMMAND_BLUETOOTH_start_scan_sem, portMAX_DELAY);
            }

            status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
        } while (0);
    }

    return status;
}

int32_t LLRCOMMAND_BLUETOOTH_IMPL_scanStop(uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    llrcommand_bluetooth_current_state.is_scanning = false;

    api_status = esp_ble_gap_stop_scanning();
    if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[ERROR] Scan stop failed with error: (err = %d)\n", api_status);
#endif
    } else {
        xSemaphoreTake(LLRCOMMAND_BLUETOOTH_stop_scan_sem, portMAX_DELAY);
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    }

    return status;
}

int32_t LLRCOMMAND_BLUETOOTH_IMPL_connect(int32_t deviceAddrType, int8_t* deviceAddr, int32_t deviceAddrLength, uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;
    esp_ble_addr_type_t ble_addr_type;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    if (true == llrcommand_bluetooth_current_state.is_scanning) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] %s stop scanning before connect.\n", __func__);
#endif
        LLRCOMMAND_BLUETOOTH_IMPL_scanStop(0);
    }

    LLRCOMMAND_BLUETOOTH_get_device_from_cache((uint8_t *)deviceAddr, &ble_addr_type);

    api_status = esp_ble_gattc_open(gl_profile_tab[PROFILE_APP_ID].gattc_if, (uint8_t *)deviceAddr, ble_addr_type, true);
    if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[ERROR] Connect failed with error: (err = %d)\n", api_status);
#endif
    } else {
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    }

    return status;
}

int32_t LLRCOMMAND_BLUETOOTH_IMPL_write(int32_t handle, int8_t* data, int32_t offset, int32_t length, uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    gl_profile_tab[PROFILE_APP_ID].char_handle = (uint8_t)(handle & 0xFF);

    api_status = esp_ble_gattc_write_char(gl_profile_tab[PROFILE_APP_ID].gattc_if,
            gl_profile_tab[PROFILE_APP_ID].conn_id,
            gl_profile_tab[PROFILE_APP_ID].char_handle,
            length,
            (uint8_t *)data,
            ESP_GATT_WRITE_TYPE_RSP,
            ESP_GATT_AUTH_REQ_NONE);
    if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[ERROR] Write failed with error: (err = %d)\n", api_status);
#endif
    } else {
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    }

    return status;
}

int32_t LLRCOMMAND_BLUETOOTH_IMPL_read(int32_t handle, uint8_t retry) {

    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    gl_profile_tab[PROFILE_APP_ID].char_handle = (uint16_t)handle;

    api_status = esp_ble_gattc_read_char(gl_profile_tab[PROFILE_APP_ID].gattc_if,
            gl_profile_tab[PROFILE_APP_ID].conn_id,
            gl_profile_tab[PROFILE_APP_ID].char_handle,
            ESP_GATT_AUTH_REQ_NONE);
    if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[ERROR] Read failed with error: (err = %d)\n", api_status);
#endif
    } else {
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    }

    return status;
}

static int32_t LLRCOMMAND_BLUETOOTH_IMPL_notification_java_thread = -1;

int32_t LLRCOMMAND_BLUETOOTH_IMPL_notifyEvent(int8_t* data, int32_t length, uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;

    if (-1 == LLRCOMMAND_BLUETOOTH_IMPL_notification_java_thread)
    {
        LLRCOMMAND_BLUETOOTH_IMPL_notification_java_thread = SNI_getCurrentJavaThreadID();
    }
    if (NULL != LLRCOMMAND_BLUETOOTH_notification_sem)
    {
        xSemaphoreTake(LLRCOMMAND_BLUETOOTH_notification_sem, portMAX_DELAY);
        if (LLRCOMMAND_BLUETOOTH_notification_pending)
        {
            if (LLRCOMMAND_BLUETOOTH_NOTIFICATION_BUFFER_SIZE != length)
            {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                printf("[WARNING] %s: Java Immortal buffer size != C notification buffer size (%d != %d)\n", __func__, length, LLRCOMMAND_BLUETOOTH_NOTIFICATION_BUFFER_SIZE);
#endif
            }
            else
            {
                if (LLRCOMMAND_BLUETOOTH_immortal_notification_size <= length)
                {
                    memcpy(data, LLRCOMMAND_BLUETOOTH_immortal_notification_ptr, LLRCOMMAND_BLUETOOTH_immortal_notification_size);
                    status = LLRCOMMAND_BLUETOOTH_immortal_notification_size;
                    LLRCOMMAND_BLUETOOTH_notification_pending = 0;
                    LLRCOMMAND_BLUETOOTH_immortal_notification_size = 0;
                }
                else
                {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                    printf("[ERROR] %s: Java Immortal buffer size < C notification size (%d != %d)\n", __func__, length, LLRCOMMAND_BLUETOOTH_immortal_notification_size);
#endif
                }
            }
        } else {
            // Put the thread to sleep
            LLRCOMMAND_BLUETOOTH_notification_wait();
            status = LLRCOMMAND_BLUETOOTH_IMPL_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
        }
        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_notification_sem);
    } else {
        // bluetooth init not already called, wait for it.
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[WARNING] %s: Bluetooth not already initialized\n", __func__);
#endif
        LLRCOMMAND_BLUETOOTH_notification_wait();
        status = LLRCOMMAND_BLUETOOTH_IMPL_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
    }

    return status;
}

int32_t LLRCOMMAND_BLUETOOTH_IMPL_disconnect(uint8_t retry) {
    int32_t status = LLRCOMMAND_BLUETOOTH_STATUS_UNDEFINED_ERROR;
    int32_t api_status;

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s\n", __func__);
#endif

    api_status = esp_ble_gattc_close(gl_profile_tab[PROFILE_APP_ID].gattc_if,
            gl_profile_tab[PROFILE_APP_ID].conn_id);
    if (api_status != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[ERROR] Disconnect failed with error: (err = %d)\n", api_status);
#endif
    } else {
        status = LLRCOMMAND_BLUETOOTH_IMPL_NO_ERROR;
    }

    return status;
}

static void LLRCOMMAND_BLUETOOTH_event_handler(esp_gattc_cb_event_t event, esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param) {
    esp_ble_gattc_cb_param_t *p_data = (esp_ble_gattc_cb_param_t *)param;

    switch (event) {
    case ESP_GATTC_REG_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_REG_EVT\n");
#endif
        break;

    case ESP_GATTC_UNREG_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_UNREG_EVT\n");
#endif
        break;

    case ESP_GATTC_CONNECT_EVT:
    {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_CONNECT_EVT conn_id %d, if %d\n", p_data->connect.conn_id, gattc_if);
#endif
        gl_profile_tab[PROFILE_APP_ID].conn_id = p_data->connect.conn_id;
        memcpy(gl_profile_tab[PROFILE_APP_ID].remote_bda, p_data->connect.remote_bda, sizeof(esp_bd_addr_t));
        esp_err_t mtu_ret = esp_ble_gattc_send_mtu_req (gattc_if, p_data->connect.conn_id);
        if (mtu_ret != ESP_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] config MTU error, error code = %x\n", mtu_ret);
#endif
            break;
        }
        break;
    }

    case ESP_GATTC_OPEN_EVT:
        if (param->open.status == ESP_GATT_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] %s: ESP_GATTC_OPEN_EVT ==> CONNECTED\n", __func__);
#endif
            LLRCOMMAND_BLUETOOTH_gatt_connection_up(p_data);
        } else {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] %s: ESP_GATTC_OPEN_EVT ==> NOT CONNECTED\n", __func__);
#endif
        }
        xSemaphoreTake(LLRCOMMAND_BLUETOOTH_notification_sem, portMAX_DELAY);
        if (LLRCOMMAND_BLUETOOTH_notification_pending) {
            // a notification has already been pushed and will be dropped
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[WARNING] %s: /!\\ notification dropped /!\\\n", __func__);
#endif
        }
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[0] = LLRCOMMAND_BLUETOOTH_CONNECT_RESULT_OPCODE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[1] = LLRCOMMAND_BLUETOOTH_CONNECT_RESULT_SIZE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[2] = (param->open.status == ESP_GATT_OK) ? LLRCOMMAND_BLUETOOTH_STATUS_CONNECTED : LLRCOMMAND_BLUETOOTH_STATUS_NOT_CONNECTED;
        LLRCOMMAND_BLUETOOTH_immortal_notification_size = LLRCOMMAND_BLUETOOTH_CONNECT_RESULT_SIZE + 2;
        LLRCOMMAND_BLUETOOTH_notification_wakeup();
        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_notification_sem);
        break;

    case ESP_GATTC_CFG_MTU_EVT:
        if (param->cfg_mtu.status != ESP_GATT_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] config mtu failed, error status = %x\n", param->cfg_mtu.status);
#endif
            break;
        }
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_CFG_MTU_EVT, Status %d, MTU %d, conn_id %d\n", param->cfg_mtu.status, param->cfg_mtu.mtu, param->cfg_mtu.conn_id);
#endif
        used_mtu = param->cfg_mtu.mtu;
        break;

    case ESP_GATTC_SEARCH_RES_EVT:
    {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_SEARCH_RES_EVT\n");
#endif
        break;
    }

    case ESP_GATTC_SEARCH_CMPL_EVT:
        if (p_data->search_cmpl.status != ESP_GATT_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] search service failed, error status = %x\n", p_data->search_cmpl.status);
#endif
            break;
        }
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_SEARCH_CMPL_EVT\n");
#endif
        break;

    case ESP_GATTC_REG_FOR_NOTIFY_EVT:
    {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_REG_FOR_NOTIFY_EVT\n");
#endif
        if (p_data->reg_for_notify.status != ESP_GATT_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] ESP_GATTC_REG_FOR_NOTIFY_EVT failed: error status = %d\n", p_data->reg_for_notify.status);
#endif
            break;
        } else {
            uint16_t count = 0;
            esp_gatt_status_t ret_status = esp_ble_gattc_get_attr_count(gattc_if,
                    gl_profile_tab[PROFILE_APP_ID].conn_id,
                    ESP_GATT_DB_DESCRIPTOR,
                    gl_profile_tab[PROFILE_APP_ID].service_start_handle,
                    gl_profile_tab[PROFILE_APP_ID].service_end_handle,
                    gl_profile_tab[PROFILE_APP_ID].char_handle,
                    &count);
            if (ret_status != ESP_GATT_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                printf("[ERROR] esp_ble_gattc_get_attr_count error\n");
#endif
                break;
            }
        }
        break;
    }

    case ESP_GATTC_NOTIFY_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_NOTIFY_EVT\n");
#endif
        xSemaphoreTake(LLRCOMMAND_BLUETOOTH_notification_sem, portMAX_DELAY);
        if (LLRCOMMAND_BLUETOOTH_notification_pending) {
            // a notification has already been pushed and will be dropped
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[WARNING] %s: /!\\ notification dropped /!\\\n", __func__);
#endif
        }
        if (p_data->notify.is_notify) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] Notify event\n");
#endif
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[0] = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_OPCODE;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[1] = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_SIZE + p_data->notify.value_len;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[2] = p_data->notify.conn_id >> 8;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[3] = p_data->notify.conn_id;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[4] = LLRCOMMAND_BLUETOOTH_NOTIFICATION_OPCODE;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[5] = p_data->notify.handle >> 8;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[6] = p_data->notify.handle;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[7] = p_data->notify.value_len >> 8;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[8] = p_data->notify.value_len;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[9] = 0;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[10] = 0;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[11] = p_data->notify.handle >> 8;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[12] = p_data->notify.handle;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[13] = used_mtu >> 8;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[14] = used_mtu;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[15] = ESP_GATT_OK;
            memcpy(&LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[16], p_data->notify.value, p_data->notify.value_len);
            LLRCOMMAND_BLUETOOTH_immortal_notification_size = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_SIZE + p_data->notify.value_len + 2;
            LLRCOMMAND_BLUETOOTH_notification_wakeup();
            xSemaphoreGive(LLRCOMMAND_BLUETOOTH_notification_sem);
        } else {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] Indicate event\n");
#endif
        }
        break;

    case ESP_GATTC_WRITE_DESCR_EVT:
        if (p_data->write.status != ESP_GATT_OK) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] write descr failed, error status = %x\n", p_data->write.status);
#endif
            break;
        }
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] write descr success\n");
#endif
        break;

    case ESP_GATTC_SRVC_CHG_EVT:
    {
        esp_bd_addr_t bda;
        memcpy(bda, p_data->srvc_chg.remote_bda, sizeof(esp_bd_addr_t));
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_SRVC_CHG_EVT, bd_addr: %x %x %x %x %x %x:\n",
                bda[0], bda[1], bda[2], bda[3], bda[4], bda[5]);
#endif
        break;
    }

    case ESP_GATTC_WRITE_CHAR_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_WRITE_CHAR_EVT\n");
#endif
        xSemaphoreTake(LLRCOMMAND_BLUETOOTH_notification_sem, portMAX_DELAY);
        if (LLRCOMMAND_BLUETOOTH_notification_pending) {
            // a notification has already been pushed and will be dropped
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[WARNING] %s: /!\\ notification dropped /!\\\n", __func__);
#endif
        }
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[0] = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_OPCODE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[1] = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_SIZE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[2] = p_data->write.conn_id >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[3] = p_data->write.conn_id;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[4] = LLRCOMMAND_BLUETOOTH_WRITE_OPCODE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[5] = p_data->write.handle >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[6] = p_data->write.handle;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[7] = 0;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[8] = 0;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[9] = p_data->write.offset >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[10] = p_data->write.offset;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[11] = p_data->write.handle >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[12] = p_data->write.handle;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[13] = used_mtu >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[14] = used_mtu;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[15] = p_data->write.status;
        LLRCOMMAND_BLUETOOTH_immortal_notification_size = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_SIZE + 2;
        LLRCOMMAND_BLUETOOTH_notification_wakeup();
        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_notification_sem);
        break;

    case ESP_GATTC_READ_CHAR_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_READ_CHAR_EVT\n");
#endif
        xSemaphoreTake(LLRCOMMAND_BLUETOOTH_notification_sem, portMAX_DELAY);
        if (LLRCOMMAND_BLUETOOTH_notification_pending) {
            // a notification has already been pushed and will be dropped
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[WARNING] %s: /!\\ notification dropped /!\\\n", __func__);
#endif
        }
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[0] = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_OPCODE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[1] = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_SIZE + p_data->read.value_len;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[2] = p_data->read.conn_id >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[3] = p_data->read.conn_id;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[4] = LLRCOMMAND_BLUETOOTH_READ_OPCODE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[5] = p_data->read.handle >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[6] = p_data->read.handle;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[7] = p_data->read.value_len >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[8] = p_data->read.value_len;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[9] = 0;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[10] = 0;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[11] = p_data->read.handle >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[12] = p_data->read.handle;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[13] = used_mtu >> 8;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[14] = used_mtu;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[15] = p_data->read.status;
        memcpy(&LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[16], p_data->read.value, p_data->read.value_len);
        LLRCOMMAND_BLUETOOTH_immortal_notification_size = LLRCOMMAND_BLUETOOTH_OP_COMPLETE_SIZE + p_data->read.value_len + 2;
        LLRCOMMAND_BLUETOOTH_notification_wakeup();
        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_notification_sem);
        break;

    case ESP_GATTC_DISCONNECT_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_DISCONNECT_EVT, reason = %d\n", p_data->disconnect.reason);
#endif
        xSemaphoreTake(LLRCOMMAND_BLUETOOTH_notification_sem, portMAX_DELAY);
        if (LLRCOMMAND_BLUETOOTH_notification_pending) {
            // a notification has already been pushed and will be dropped
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[WARNING] %s: /!\\ notification dropped /!\\\n", __func__);
#endif
        }
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[0] = LLRCOMMAND_BLUETOOTH_CONNECT_RESULT_OPCODE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[1] = LLRCOMMAND_BLUETOOTH_CONNECT_RESULT_SIZE;
        LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[2] = LLRCOMMAND_BLUETOOTH_STATUS_NOT_CONNECTED;
        LLRCOMMAND_BLUETOOTH_immortal_notification_size = LLRCOMMAND_BLUETOOTH_CONNECT_RESULT_SIZE + 2;
        LLRCOMMAND_BLUETOOTH_notification_wakeup();
        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_notification_sem);
        break;

    case ESP_GATTC_CLOSE_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GATTC_CLOSE_EVT\n");
#endif
        break;

    default:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[WARNING] %s: untreated event %d\n", __func__, event);
#endif
        break;
    }
}

static void LLRCOMMAND_BLUETOOTH_gattc_cb(esp_gattc_cb_event_t event, esp_gatt_if_t gattc_if, esp_ble_gattc_cb_param_t *param) {
    /* If event is register event, store the gattc_if for each profile */
    if (event == ESP_GATTC_REG_EVT) {
        if (param->reg.status == ESP_GATT_OK) {
            gl_profile_tab[param->reg.app_id].gattc_if = gattc_if;
        } else {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] reg app failed, app_id %04x, status %d\n",
                    param->reg.app_id,
                    param->reg.status);
#endif
            return;
        }
    }

    /* If the gattc_if equal to profile A, call profile A cb handler,
     * so here call each profile's callback */
    do {
        int idx;
        for (idx = 0; idx < PROFILE_NUM; idx++) {
            if (gattc_if == ESP_GATT_IF_NONE || /* ESP_GATT_IF_NONE, not specify a certain gatt_if, need to call every profile cb function */
                    gattc_if == gl_profile_tab[idx].gattc_if) {
                if (gl_profile_tab[idx].gattc_cb) {
                    gl_profile_tab[idx].gattc_cb(event, gattc_if, param);
                }
            }
        }
    } while (0);
}

static void LLRCOMMAND_BLUETOOTH_gap_cb(esp_gap_ble_cb_event_t event, esp_ble_gap_cb_param_t *param) {
    uint8_t ble_device_name_len = 0;
    uint8_t *ble_device_name = NULL;

    switch (event) {
    case ESP_GAP_BLE_SCAN_PARAM_SET_COMPLETE_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GAP_BLE_SCAN_PARAM_SET_COMPLETE_EVT\n");
#endif
        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_set_param_sem);
        break;

    case ESP_GAP_BLE_SCAN_START_COMPLETE_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] ESP_GAP_BLE_SCAN_START_COMPLETE_EVT\n");
#endif
        if (param->scan_start_cmpl.status != ESP_BT_STATUS_SUCCESS) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] scan start failed, error status = %x\n", param->scan_start_cmpl.status);
#endif
            break;
        }
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] scan start success\n");
#endif
        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_start_scan_sem);
        break;

    case ESP_GAP_BLE_SCAN_RESULT_EVT:
    {
        esp_ble_gap_cb_param_t *scan_result = (esp_ble_gap_cb_param_t *)param;
        switch (scan_result->scan_rst.search_evt)
        {
        case ESP_GAP_SEARCH_INQ_RES_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] ESP_GAP_SEARCH_INQ_RES_EVT searched Adv Data Len %d, Scan Response Len %d bda: %x %x %x %x %x %x, Addr type: %d :\n",
                    scan_result->scan_rst.adv_data_len, scan_result->scan_rst.scan_rsp_len,
                    scan_result->scan_rst.bda[0], scan_result->scan_rst.bda[1], scan_result->scan_rst.bda[2], scan_result->scan_rst.bda[3],
                    scan_result->scan_rst.bda[4], scan_result->scan_rst.bda[5], scan_result->scan_rst.ble_addr_type);
#endif
            LLRCOMMAND_BLUETOOTH_add_device_to_cache(scan_result->scan_rst.bda, scan_result->scan_rst.ble_addr_type);
            xSemaphoreTake(LLRCOMMAND_BLUETOOTH_notification_sem, portMAX_DELAY);
            if (LLRCOMMAND_BLUETOOTH_notification_pending) {
                // a notification has already been pushed and will be dropped
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                printf("[WARNING] %s: /!\\ notification dropped /!\\\n", __func__);
#endif
            }
            ble_device_name = esp_ble_resolve_adv_data(scan_result->scan_rst.ble_adv, ESP_BLE_AD_TYPE_NAME_CMPL, &ble_device_name_len);
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[0] = LLRCOMMAND_BLUETOOTH_SCAN_RESULT_OPCODE;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[1] = LLRCOMMAND_BLUETOOTH_SCAN_RESULT_SIZE + ble_device_name_len;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[2] = scan_result->scan_rst.ble_addr_type;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[3] = scan_result->scan_rst.ble_evt_type;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[4] = scan_result->scan_rst.flag;
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[5] = scan_result->scan_rst.rssi;
            memcpy(&LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[6], scan_result->scan_rst.bda, sizeof(esp_bd_addr_t));
            LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[6 + sizeof(esp_bd_addr_t)] = ble_device_name_len;
            if (ble_device_name_len > 0) {
                memcpy(&LLRCOMMAND_BLUETOOTH_immortal_notification_ptr[6 + sizeof(esp_bd_addr_t) + 1], ble_device_name, ble_device_name_len);
            }
            LLRCOMMAND_BLUETOOTH_immortal_notification_size = LLRCOMMAND_BLUETOOTH_SCAN_RESULT_SIZE + ble_device_name_len + 2;
            LLRCOMMAND_BLUETOOTH_notification_wakeup();
            llrcommand_bluetooth_current_state.is_scanning = true;
            xSemaphoreGive(LLRCOMMAND_BLUETOOTH_notification_sem);
            break;

        case ESP_GAP_SEARCH_INQ_CMPL_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] scan search complete\n");
#endif
            llrcommand_bluetooth_current_state.is_scanning = false;
            break;

        default:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[WARNING] %s: event %d => untreated event %d\n", __func__, event, scan_result->scan_rst.search_evt);
#endif
            break;
        }
        break;
    }

    case ESP_GAP_BLE_SCAN_STOP_COMPLETE_EVT:
        if (param->scan_stop_cmpl.status != ESP_BT_STATUS_SUCCESS) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] scan stop failed, error status = %x\n", param->scan_stop_cmpl.status);
#endif
            break;
        }
        llrcommand_bluetooth_current_state.is_scanning = false;
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] stop scan successfully\n");
#endif

#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO]  Ready to connect to the remote device.\n");
#endif

        xSemaphoreGive(LLRCOMMAND_BLUETOOTH_stop_scan_sem);
        break;

    case ESP_GAP_BLE_ADV_STOP_COMPLETE_EVT:
        if (param->adv_stop_cmpl.status != ESP_BT_STATUS_SUCCESS) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] adv stop failed, error status = %x\n", param->adv_stop_cmpl.status);
#endif
            break;
        }
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] stop adv successfully\n");
#endif
        break;

    case ESP_GAP_BLE_UPDATE_CONN_PARAMS_EVT:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] update connection params status = %d, min_int = %d, max_int = %d,conn_int = %d,latency = %d, timeout = %d\n",
                param->update_conn_params.status,
                param->update_conn_params.min_int,
                param->update_conn_params.max_int,
                param->update_conn_params.conn_int,
                param->update_conn_params.latency,
                param->update_conn_params.timeout);
#endif
        break;

    default:
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[WARNING] %s: untreated event %d\n", __func__, event);
#endif
        break;
    }
}

esp_gatt_status_t LLRCOMMAND_BLUETOOTH_gatt_connection_up(esp_ble_gattc_cb_param_t *p_conn_status) {
    LLRCOMMAND_BLUETOOTH_ble_connection_id = p_conn_status->connect.conn_id;
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
    printf("[INFO] %s: connection is up with id %d\n", __func__, LLRCOMMAND_BLUETOOTH_ble_connection_id);
#endif
    return ESP_GATT_OK;
}

static void LLRCOMMAND_BLUETOOTH_notification_wakeup() {
    if (-1 != LLRCOMMAND_BLUETOOTH_IMPL_notification_java_thread) {
        if (SNI_OK == SNI_resumeJavaThread(LLRCOMMAND_BLUETOOTH_IMPL_notification_java_thread)){
            LLRCOMMAND_BLUETOOTH_notification_pending = 1;
        } else {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[ERROR] %s: Java thread resume fail\n", __func__);
#endif
        }
    } else {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[ERROR] %s: no Java notification thread to resume.\n", __func__);
#endif
    }
}

static void LLRCOMMAND_BLUETOOTH_notification_wait() {
    int32_t suspend_result = SNI_suspendCurrentJavaThread(0);

    if (suspend_result != SNI_OK)
    {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[ERROR] %s: Java thread suspend fail.\n", __func__);
#endif
    }
}

static void LLRCOMMAND_BLUETOOTH_add_device_to_cache(uint8_t *bda, esp_ble_addr_type_t ble_addr_type)
{
    if (device_cache != NULL) {
        device_cache_t *tmp = device_cache;

        do {
            if (memcmp(tmp->bda, bda, sizeof(tmp->bda)) == 0) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                printf("[INFO] Device with bda: %x %x %x %x %x %x already exists in cache\n",
                        bda[0], bda[1], bda[2], bda[3], bda[4], bda[5]);
#endif
                break;
            }

            tmp = tmp->next;
        } while (tmp != NULL);

        if (tmp == NULL) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[INFO] Adding device with bda: %x %x %x %x %x %x and ble_addr_type: %d to the cache\n",
                    bda[0], bda[1], bda[2], bda[3], bda[4], bda[5], ble_addr_type);
#endif
            tmp = (device_cache_t *)malloc(sizeof(*tmp));
            memcpy(tmp->bda, bda, sizeof(tmp->bda));
            tmp->ble_addr_type = ble_addr_type;
            tmp->next = device_cache;
            device_cache = tmp;
        }
    } else {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[INFO] Adding device with bda: %x %x %x %x %x %x and ble_addr_type: %d to the cache\n",
                bda[0], bda[1], bda[2], bda[3], bda[4], bda[5], ble_addr_type);
#endif
        device_cache = (device_cache_t *)malloc(sizeof(*device_cache));
        memcpy(device_cache->bda, bda, sizeof(device_cache->bda));
        device_cache->ble_addr_type = ble_addr_type;
        device_cache->next = NULL;
    }
}

static void LLRCOMMAND_BLUETOOTH_get_device_from_cache(uint8_t *bda, esp_ble_addr_type_t *ble_addr_type)
{
    if (device_cache != NULL) {
        device_cache_t *tmp = device_cache;

        do {
            if (memcmp(tmp->bda, bda, sizeof(tmp->bda)) == 0) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
                printf("[INFO] Device with bda: %x %x %x %x %x %x was found in cache. Will return the stored ble_addr_type: %d\n",
                        bda[0], bda[1], bda[2], bda[3], bda[4], bda[5], tmp->ble_addr_type);
#endif
                *ble_addr_type = tmp->ble_addr_type;
                break;
            }

            tmp = tmp->next;
        } while (tmp != NULL);

        if (tmp == NULL) {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
            printf("[WARNING] Device with bda: %x %x %x %x %x %x was not found in cache. Will return the default ble_addr_type: %d\n",
                 bda[0], bda[1], bda[2], bda[3], bda[4], bda[5], BLE_ADDR_TYPE_RANDOM);
#endif
            *ble_addr_type = BLE_ADDR_TYPE_RANDOM;
        }
    } else {
#ifdef LLRCOMMAND_BLUETOOTH_DEBUG
        printf("[WARNING] Cache is empty. Will return the default ble_addr_type: %d for device with bda: %x %x %x %x %x %x\n",
                BLE_ADDR_TYPE_RANDOM, bda[0], bda[1], bda[2], bda[3], bda[4], bda[5]);
#endif
        *ble_addr_type = BLE_ADDR_TYPE_RANDOM;
    }
}
