/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * This implementation uses a dedicated hardware timer for time implementation,
 * instead of internal FreeRTOS timer which is a 32bits timer.
 */

/* Includes ------------------------------------------------------------------*/

#include <microej_time.h>
#include <stdlib.h>
#include <stdio.h>

#include "freertos/FreeRTOS.h"
#include "freertos/timers.h"
#include "freertos/task.h"
#include "freertos/semphr.h"

#include "LLMJVM_impl.h"
#include "microej.h"

/* Defines -------------------------------------------------------------------*/

/*
 * Useful macros and definitions
 */

// ID for the FreeRTOS Timer
#define WAKE_UP_TIMER_ID	42

/* Globals -------------------------------------------------------------------*/

/*
 * Shared variables
 */
//Absolute time in ms at which timer will be launched
static int64_t LLMJVM_FREERTOS_next_wake_up_time = INT64_MAX;

// Timer for scheduling next alarm
static TimerHandle_t LLMJVM_FREERTOS_wake_up_timer;

// Binary semaphore to wakeup microJVM
static SemaphoreHandle_t LLMJVM_FREERTOS_Semaphore;

/* Private functions ---------------------------------------------------------*/

// Since LLMJVM_schedule() prototype does not match the timer callback prototype,
// we create a wrapper around it and check the ID of the timer.
static void wake_up_timer_callback(TimerHandle_t timer)
{
    uint32_t id = (uint32_t ) pvTimerGetTimerID(timer);
	if(id == WAKE_UP_TIMER_ID)
	{
		LLMJVM_schedule();
	}
}

/* Public functions ----------------------------------------------------------*/

/*
 * Implementation of functions from LLMJVM_impl.h
 * and other helping functions.
 */

// Creates the timer used to callback the LLMJVM_schedule() function.
// After its creation, the timer is idle.
int32_t LLMJVM_IMPL_initialize()
{
    // Create a timer to scheduler alarm for the VM
    LLMJVM_FREERTOS_wake_up_timer = xTimerCreate(NULL, 100, pdFALSE, (void*) WAKE_UP_TIMER_ID, wake_up_timer_callback);

	if(LLMJVM_FREERTOS_wake_up_timer == 0)
	{
		return LLMJVM_ERROR;
	}

	vSemaphoreCreateBinary(LLMJVM_FREERTOS_Semaphore);

	if(LLMJVM_FREERTOS_Semaphore == NULL)
	{
		return LLMJVM_ERROR;
	}

	// Created semaphore is in a state such that the first call to 'take' the semaphore would pass.
	// Just execute a take to init it at 0.
	// Note: From FreeRTOS version 7.6.0, the function xSemaphoreCreateBinary can be used to avoid
	// the call to xSemaphoreTake.
	xSemaphoreTake(LLMJVM_FREERTOS_Semaphore, portMAX_DELAY);

	return LLMJVM_OK;
}

// Once the task is started, save a handle to it.
int32_t LLMJVM_IMPL_vmTaskStarted()
{
	return LLMJVM_OK;
}

// Schedules requests from the VM
int32_t LLMJVM_IMPL_scheduleRequest(int64_t absoluteTime)
{
	int64_t relativeTime; // relative from current time to 'absoluteTime'
	int64_t relativeTick;
    portBASE_TYPE xTimerChangePeriodResult;
    portBASE_TYPE xTimerStartResult;

	//First check if absolute time is lower than current schedule time
	if(absoluteTime < LLMJVM_FREERTOS_next_wake_up_time)
    {

        // Save new alarm absolute time
        LLMJVM_FREERTOS_next_wake_up_time = absoluteTime;

        // Stop current timer (no delay)
        xTimerStop(LLMJVM_FREERTOS_wake_up_timer, 0);

        relativeTime = absoluteTime - LLMJVM_IMPL_getCurrentTime(MICROEJ_TRUE);
        // Determine relative time/tick
        relativeTick = microej_time_timeToTick(relativeTime);

		if(relativeTick > 0)
		{
            // Schedule the new alarm
			xTimerChangePeriodResult = xTimerChangePeriod(LLMJVM_FREERTOS_wake_up_timer, relativeTick, 0);
			xTimerStartResult = xTimerStart(LLMJVM_FREERTOS_wake_up_timer, 0);

            if( (xTimerChangePeriodResult == pdPASS) && (xTimerStartResult == pdPASS) )
                return LLMJVM_OK;
            else
                return LLMJVM_ERROR;
		}
        else
        {
            // 'absoluteTime' has been reached already, notify the VM now
            return LLMJVM_schedule();
        }
	}
    // else : nothing to do. An sooner alarm is scheduled.

	return LLMJVM_OK;
}

// Suspends the VM task if the pending flag is not set
int32_t LLMJVM_IMPL_idleVM()
{
	portBASE_TYPE res = xSemaphoreTake(LLMJVM_FREERTOS_Semaphore, portMAX_DELAY);

	return res == pdTRUE ? LLMJVM_OK : LLMJVM_ERROR;
}

// Wakes up the VM task and reset next wake up time
int32_t LLMJVM_IMPL_wakeupVM()
{
	portBASE_TYPE res;
	if(xPortInIsrContext()){
		portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
		res = xSemaphoreGiveFromISR(LLMJVM_FREERTOS_Semaphore, &xHigherPriorityTaskWoken);
		if( xHigherPriorityTaskWoken != pdFALSE ){
			// Force a context switch here.
			portYIELD_FROM_ISR();
		}
	}
	else {
		res = xSemaphoreGive(LLMJVM_FREERTOS_Semaphore);
	}

	LLMJVM_FREERTOS_next_wake_up_time = INT64_MAX;

	return res == pdTRUE ? LLMJVM_OK : LLMJVM_ERROR;
}

// Clear the pending wake up flag
int32_t LLMJVM_IMPL_ackWakeup()
{
	return LLMJVM_OK;
}

int32_t LLMJVM_IMPL_getCurrentTaskID()
{
	return (int32_t) xTaskGetCurrentTaskHandle();
}

// Sets application time
void LLMJVM_IMPL_setApplicationTime(int64_t t)
{
	microej_time_setapplicationtime(t);
}

// Gets the system or the application time in milliseconds
int64_t LLMJVM_IMPL_getCurrentTime(uint8_t system)
{
	return microej_time_getcurrenttime(system);
}

// Gets the current system time in nanoseconds
int64_t LLMJVM_IMPL_getTimeNanos()
{
	return microej_time_gettimenanos();
}

int32_t LLMJVM_IMPL_shutdown(void)
{
	// nothing to do
	return LLMJVM_OK;
}

#include "sni.h"
jdouble Java_com_is2t_microjvm_test_MJVMPortValidation_testFPU__DD(jdouble a, jdouble b)
{
	return a * b;
}
jfloat Java_com_is2t_microjvm_test_MJVMPortValidation_testFPU__FF(jfloat a, jfloat b)
{
	return a * b;
}
