/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#ifndef WIFI_ESP32_DRIVER_H
#define WIFI_ESP32_DRIVER_H

#include <stdbool.h>
#include <stdint.h>

#include "NETIF_interface.h"

/** @brief Wifi interface name */
#define WLAN_INTERFACE_NAME	"wifi_interface"

/**
 * @brief function to initialize the software module
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t WIFI_ESP32_init_f(void);

/**
 * @brief function to initialize the wifi module
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t WIFI_ESP32_initialize_f(void);

/**
 * @brief this function apply IP configuration from network interface
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t WIFI_ESP32_apply_configuration_f(void);

/**
 * @brief this function start the wifi module
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t WIFI_ESP32_start_f(void);

/**
 * @brief this function stop the wifi module
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t WIFI_ESP32_stop_f(void);

/**
 * @brief this function return if the network interface is started or not
 *
 * @param[out] _puc_netif_started '1' returned if started, '0' otherwise
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t WIFI_ESP32_netif_is_started_f(unsigned char * _puc_netif_started);

/**
 * @brief this function checks if the WiFi is started
 *
 * @return true if success else false
 */
bool WIFI_ESP32_is_started_f(void);

/**
 * @brief this function joins a WiFi network
 *
 * @param[in] _pc_ssid ssid of the network
 * @param[in] _i_ssid_length ssid length
 * @param[in] _pc_passphrase password needed for network connection
 * @param[in] _i_passphrase_length password length
 * @param[in] _i_security_mode security mode
 *
 * @return true if success else false
 */
bool WIFI_ESP32_join_f(const char * _pc_ssid, int _i_ssid_length, const char * _pc_passphrase, int _i_passphrase_length, int _i_security_mode);

/**
 * @brief this function leaves a WiFi network
 *
 * @return true if success else false
 */
bool WIFI_ESP32_leave_f(void);

/**
 * @brief this function returns the bssid of the WiFi network
 *
 * @param[out] _puc_bssid the buffer to be filled with the bssid
 * @param[in]  _i_bssid_length length to be filled in the buffer
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_bssid_f(unsigned char * _puc_bssid, int _i_bssid_length);

/**
 * @brief this function returns the channel of the WiFi network
 *
 * @param[out] _ps_channel the buffer to be filled
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_channel_f(short * _ps_channel);

/**
 * @brief this function returns the rssi of the WiFi network
 *
 * @param[out] _pf_rssi the buffer to be filled
 * @param[in]  _i_rssi_length length to be filled in the buffer
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_rssi_f(float * _pf_rssi, int _i_rssi_length);

/**
 * @brief this function returns the security mode of the WiFi network
 *
 * @param[out] _ps_mode the buffer to be filled
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_security_mode_f(short * _ps_mode);

/**
 * @brief this function returns the ssid of the WiFi network
 *
 * @param[out] _pc_ssid the buffer to be filled
 * @param[in]  _i_ssid_length length to be filled in the buffer
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ssid_f(char * _pc_ssid, int _i_ssid_length);

/**
 * @brief this function returns the WPS modes of the WiFi network
 *
 * @param[out] _ps_wps_modes the buffer to be filled
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_wps_modes_f(short * _ps_wps_modes);

/**
 * @brief this function enables the SoftAP for the module
 *
 * @param[in] _pc_ssid the ssid of the SoftAP
 * @param[in] _i_ssid_length the ssid length
 * @param[in] _pc_passphrase the password of the SoftAP
 * @param[in] _i_passphrase_length the password length
 *
 * @return true if success else false
 */
bool WIFI_ESP32_enable_softap_f(const char * _pc_ssid, int _i_ssid_length, const char * _pc_passphrase, int _i_passphrase_length);

/**
 * @brief this function changes the name of the SoftAP
 *
 * @param[in] _pc_name the ssid of the SoftAP
 * @param[in] _i_name_length the ssid length
 *
 * @return true if success else false
 */
bool WIFI_ESP32_set_name_softap_f(const char * _pc_name, int _i_name_length);

/**
 * @brief this function disables the SoftAP
 *
 * @return true if success else false
 */
bool WIFI_ESP32_disable_softap_f(void);

/**
 * @brief this function returns the access point count discovered
 *
 * @param[out] _ps_ap_count the access point count to be returned
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ap_count_f(short * const _ps_ap_count);

/**
 * @brief this function returns the ssid from the saved list
 *
 * @param[in]  _ui_index the index from the saved list
 * @param[out] _puc_ssid the buffer to be returned
 * @param[in]  _i_ssid_length the length to copy in the return buffer
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ap_ssid_f(unsigned int _ui_index, unsigned char * _puc_ssid, int _i_ssid_length);

/**
 * @brief this function returns the bssid from the saved list
 *
 * @param[in]  _ui_index the index from the saved list
 * @param[out] _puc_bssid the buffer to be returned
 * @param[in]  _i_bssid_length the length to copy in the return buffer
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ap_bssid_f(unsigned int _ui_index, unsigned char * _puc_bssid, int _i_bssid_length);

/**
 * @brief this function returns the channel from the saved list
 *
 * @param[in]  _ui_index the index from the saved list
 * @param[out] _pi_channel the buffer to be returned
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ap_channel_f(unsigned int _ui_index, int * _pi_channel);

/**
 * @brief this function returns the rssi from the saved list
 *
 * @param[in]  _ui_index the index from the saved list
 * @param[out] _pf_rssi the buffer to be returned
 * @param[in]  _i_rssi_length the length to copy in the return buffer
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ap_rssi_f(unsigned int _ui_index, float * _pf_rssi, int _i_rssi_length);

/**
 * @brief this function returns the security mode from the saved list
 *
 * @param[in]  _ui_index the index from the saved list
 * @param[out] _pi_mode the buffer to be returned
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ap_security_mode_f(unsigned int _ui_index, int * _pi_mode);

/**
 * @brief this function returns the wps mode from the saved list
 *
 * @param[in]  _ui_index the index from the saved list
 * @param[out] _pi_wps_cap the buffer to be returned
 *
 * @return true if success else false
 */
bool WIFI_ESP32_get_ap_wps_modes_f(unsigned int _ui_index, int * _pi_wps_cap);

/**
 * @brief this function returns the client state
 *
 * @return true if the client is started else false
 */
bool WIFI_ESP32_get_client_state_f(void);

#endif /* WIFI_ESP32_DRIVER_H */
