/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <stdio.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_system.h"
#include "esp_spi_flash.h"
#include "esp_task_wdt.h"
#include "nvs_flash.h"
#include "microjvm_main.h"
#include "esp_ota_ops.h"

/******************************************************
 *                    Constants
 ******************************************************/

/* MicroJvm task constants */
#define MICROJVM_STACK_SIZE                (21 * 1024)
#define JAVA_TASK_PRIORITY                 ( 8 ) /** Should be > tskIDLE_PRIORITY & < configTIMER_TASK_PRIORITY */
#define JAVA_TASK_STACK_SIZE               (MICROJVM_STACK_SIZE / 4)
#define JAVA_TASK_NAME	                   "MicroJvm"

/* System monitor tasks constants */
#define SYSTEM_MONITOR_STACK_SIZE          (4 * 1024)
#define SYSTEM_MONITOR_TASK_STACK_SIZE     (SYSTEM_MONITOR_STACK_SIZE / 4)
#define SYSTEM_MONITOR_WAKEUP_PERIOD_MS    (CONFIG_TASK_WDT_TIMEOUT_S * 800) /** Set the system monitor threads delay time to 20% less than the WDT timer */

/* System monitor core 0 task constants */
#define SYSTEM_MONITOR_CORE0_TASK_PRIORITY ( CONFIG_TIMER_TASK_PRIORITY - 1 ) /** Should be > tskIDLE_PRIORITY & < configTIMER_TASK_PRIORITY */
#define SYSTEM_MONITOR_CORE0_TASK_NAME	   "SystemMonitorCore0"

#if portNUM_PROCESSORS > 1
/* System monitor core 1 task constants */
#define SYSTEM_MONITOR_CORE1_TASK_PRIORITY ( CONFIG_TIMER_TASK_PRIORITY - 2 ) /** Should be > tskIDLE_PRIORITY & < configTIMER_TASK_PRIORITY */
#define SYSTEM_MONITOR_CORE1_TASK_NAME	   "SystemMonitorCore1"
#endif

/******************************************************
 *               Static Function Declarations
 ******************************************************/

void __attribute__((weak)) LLCOMM_stack_initialize(void)
{
	/* Does nothing by default, override when ECOM-COMM is used */
}

static void xSystemMonitorCore0Function(void * pvParameters)
{
	esp_err_t status = ESP_OK;

	/* Subscribe this task to TWDT, then check if it is subscribed */
	status |= esp_task_wdt_add(NULL);
	status |= esp_task_wdt_status(NULL);

	/* Enter the task loop only if TWDT subscribe is OK */
	if (status == ESP_OK)
	{
		/* Enter the task loop. This task kicks the WDT on a specific interval, it can be extended with monitoring functionality. */
		do {
			/* Kick the WDT and exit the task loop if not OK */
			if (esp_task_wdt_reset() != ESP_OK)
			{
				break;
			}

			/* Delay the task */
			vTaskDelay(SYSTEM_MONITOR_WAKEUP_PERIOD_MS / portTICK_PERIOD_MS);
		} while (1);
	}

	/* If task exited, delete it */
	vTaskDelete(xTaskGetCurrentTaskHandle());
}

#if portNUM_PROCESSORS > 1
static void xSystemMonitorCore1Function(void * pvParameters)
{
	esp_err_t status = ESP_OK;

	/* Subscribe this task to TWDT, then check if it is subscribed */
	status |= esp_task_wdt_add(NULL);
	status |= esp_task_wdt_status(NULL);

	/* Enter the task loop only if TWDT subscribe is OK */
	if (status == ESP_OK)
	{
		/* Enter the task loop. This task kicks the WDT on a specific interval, it can be extended with monitoring functionality. */
		do {
			/* Kick the WDT and exit the task loop if not OK */
			if (esp_task_wdt_reset() != ESP_OK)
			{
				break;
			}

			/* Delay the task */
			vTaskDelay(SYSTEM_MONITOR_WAKEUP_PERIOD_MS / portTICK_PERIOD_MS);
		} while (1);
	}

	/* If task exited, delete it */
	vTaskDelete(xTaskGetCurrentTaskHandle());
}
#endif

static void xJavaTaskFunction(void * pvParameters)
{
	LLCOMM_stack_initialize();
	microjvm_main();
	vTaskDelete(xTaskGetCurrentTaskHandle());
}

void app_main()
{
    /* Print chip information */
    esp_chip_info_t chip_info;
    esp_chip_info(&chip_info);
    printf("This is ESP32 chip with %d CPU cores, WiFi%s%s, ",
            chip_info.cores,
            (chip_info.features & CHIP_FEATURE_BT) ? "/BT" : "",
            (chip_info.features & CHIP_FEATURE_BLE) ? "/BLE" : "");

    printf("silicon revision %d, ", chip_info.revision);

    printf("%dMB %s flash\n", spi_flash_get_chip_size() / (1024 * 1024),
            (chip_info.features & CHIP_FEATURE_EMB_FLASH) ? "embedded" : "external");

    /* Initialize NVS */
    esp_err_t ret = nvs_flash_init();
    if (ret == ESP_ERR_NVS_NO_FREE_PAGES) {
        ESP_ERROR_CHECK(nvs_flash_erase());
        ret = nvs_flash_init();
    }
    ESP_ERROR_CHECK( ret );

    /* Print partition information */
    const esp_partition_t *partition = esp_ota_get_running_partition();
    if (partition != NULL) {
        switch (partition->subtype) {
        case ESP_PARTITION_SUBTYPE_APP_FACTORY:
            printf("Running application from factory partition\n");
            break;
        case ESP_PARTITION_SUBTYPE_APP_OTA_0:
            printf("Running application from ota_0 partition\n");
            break;
        case ESP_PARTITION_SUBTYPE_APP_OTA_1:
            printf("Running application from ota_1 partition\n");
            break;
        default:
            printf("Running application from unknown partition\n");
            break;
        }
    } else {
        printf("Cannot retrieve running partition\n");
    }

    /* Start the WDT kicking threads */
    xTaskCreatePinnedToCore(xSystemMonitorCore0Function, SYSTEM_MONITOR_CORE0_TASK_NAME, SYSTEM_MONITOR_TASK_STACK_SIZE, NULL, SYSTEM_MONITOR_CORE0_TASK_PRIORITY, NULL, 0);
#if portNUM_PROCESSORS > 1
    xTaskCreatePinnedToCore(xSystemMonitorCore1Function, SYSTEM_MONITOR_CORE1_TASK_NAME, SYSTEM_MONITOR_TASK_STACK_SIZE, NULL, SYSTEM_MONITOR_CORE1_TASK_PRIORITY, NULL, 1);
#endif

    /* Start the MicroJvm thread */
    xTaskCreate(xJavaTaskFunction, JAVA_TASK_NAME, JAVA_TASK_STACK_SIZE, NULL, JAVA_TASK_PRIORITY, NULL);
}
