/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#ifndef NETWORK_INTERFACE_H
#define NETWORK_INTERFACE_H

#include <stdint.h>
#include "NET_delegate_task.h"

/** @brief max length of name interface */
#define NETIF_MAX_NAME_LENGTH       (20)

/** @brief max number of ip address for each interface */
#define NETIF_MAX_IP_ADDRESS        (1)

/** @brief max number of byte to represent an ip address */
#define NETIF_MAX_IP_LENGTH         (4)

/** @brief max number of byte to represent an mac address */
#define NETIF_MAX_MAC_LENGTH        (6)

// v6 is not supported yet, but here are the v4 and v6 constants
// for the getVMInterfaceAddress protocol between Java and the native stacks

/** @brief ipv4 address info size (tag (1) + IP (4) + prefix (1) + hasBroadcast (1) + broadcast IP (4)) */
#define IPV4_ADDR_INFO_SIZE 11
/** ipv6 address info size (tag (1) + IP (16) + prefix (1) */
#define IPV6_ADDR_INFO_SIZE 18
/** @brief ipv4 address tag */
#define IPV4_ADDR_TAG 4
/** @brief ipv6 address tag */
#define IPV6_ADDR_TAG 6

/** @brief list of module constant */
typedef enum
{
	NETIF_ERROR,
	NETIF_NO_ERROR,
	NETIF_BE_PATIENT,
	NETIF_CLOSE_NOTIFY
}NETIF_status_t;

/** @brief list of option interface */
typedef enum
{
	NETIF_IS_LOOPBACK,
	NETIF_IS_POINT_TO_POINT,
	NETIF_IS_UP,
	NETIF_IS_SUPPORT_BROADCAST,
	NETIF_MAX_OPTION,   /* no line after this */	
}NETIF_flag_t;

/** @brief represent network interface object */
typedef struct
{
	unsigned short us_id;                                                    /**< id of network interface */
	unsigned char uc_is_started;                                             /**< flag to know if interface is started */
	unsigned char uc_is_static;                                              /**< flag to know if configuration is static or dhcp */
	char pc_name[NETIF_MAX_NAME_LENGTH];                                     /**< interface name */
	unsigned char puc_flag[NETIF_MAX_OPTION];                                /**< interface flag */
	unsigned char ppuc_ip_addr[NETIF_MAX_IP_ADDRESS][NETIF_MAX_IP_LENGTH];   /**< IPV4 */
	unsigned char puc_netmask[NETIF_MAX_IP_LENGTH];                          /**< netmask */
	unsigned char puc_gateway_ip_addr[NETIF_MAX_IP_LENGTH];                  /**< gateway */
	unsigned char puc_dns_ip[NETIF_MAX_IP_LENGTH];                           /**< dns */
	unsigned char puc_hw_addr[NETIF_MAX_MAC_LENGTH];                         /**< mac addr */
	
	NETIF_status_t (*pp_initialize_f)(void);
	
	NETIF_status_t (*pp_start_f)(void);
	
	NETIF_status_t (*pp_stop_f)(void);
	
	NETIF_status_t (*pp_apply_configuration_f)(void);
																		
	NETIF_status_t (*pp_is_started_f)(unsigned char * _puc_netif_started);
}NETIF_t;

/**
 * @brief function to init module
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_init_f(void);

/**
 * @brief function to start interface
 *
 * @param[in] _pc_interface_name interface name
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_start_f(char * _pc_interface_name);

/**
 * @brief function to stop interface
 *
 * @param[in] _pc_interface_name interface name
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_stop_f(char * _pc_interface_name);

/**
 * @brief function to apply configuration 
 *
 * @param[in] _pc_interface_name interface name
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_apply_configuration_f(char * _pc_interface_name);

/**
 * @brief function to get dns ip
 *
 * @param[in] _pc_interface_name interface name
 * @param[in] _puc_dns ip dns
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_dns_f(char * _pc_interface_name, unsigned char * _puc_dns);

/**
 * @brief function to set dns ip
 *
 * @param[in] _pc_interface_name interface name
 * @param[in] _puc_dns ip dns
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_dns_f(char * _pc_interface_name, unsigned char * _puc_dns);

/**
 * @brief function to get gateway ip
 *
 * @param[in] _pc_interface_name interface name
 * @param[in] _puc_gateway ip gateway
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_gateway_f(char * _pc_interface_name, unsigned char * _puc_gateway);

/**
 * @brief function to set gateway ip
 *
 * @param[in] _pc_interface_name interface name
 * @param[in] _puc_gateway ip gateway
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_gateway_f(char * _pc_interface_name, unsigned char * _puc_gateway);

/**
 * @brief function to set network mask
 *
 * @param[in] _pc_interface_name interface name
 * @param[in] _puc_netmask ip gateway
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_netmask_f(char * _pc_interface_name, unsigned char * _puc_netmask);

/**
 * @brief function to get network mask
 *
 * @param[in] _pc_interface_name interface name
 * @param[in] _puc_netmask ip gateway
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_netmask_f(char * _pc_interface_name, unsigned char * _puc_netmask);

/**
 * @brief function to know if interface is started
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _puc_is_started fill with 0 or 1 according to state of interface 
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_is_started_f(char * _pc_interface_name, unsigned char * _puc_is_started);

/**
 * @brief function to get is static flag
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _puc_is_static fill with 0 or 1 according to state of interface configuration
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_is_static_f(char * _pc_interface_name, unsigned char * _puc_is_static);

/**
 * @brief function to set is static flag
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _uc_is_static fill with 0 or 1 according to state of interface configuration
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_is_static_f(char * _pc_interface_name, unsigned char _uc_is_static);

/**
 * @brief function to register interface
 *
 * @param[out] _ppst_registered_interface pointer change to a pointer on interface registered in private pool
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_registered_interface_f(NETIF_t ** _ppst_registered_interface);

/**
 * @brief function to unregister interface 
 *
 * @param[in] _pst_interface interface to unregister
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_release_f(NETIF_t * const _pst_interface);

/**
 * @brief function to get count of registered interface 
 *
 * @param[out] _pus_count count of registered interface
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_count_registered_interface_f(unsigned short * _pus_count);

/**
 * @brief function to get count of ip address on one interface 
 *
 * @param[in] _us_interface_id interface id
 * @param[out] _pus_count count of address on interface
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_count_address_on_interface_f(unsigned short _us_interface_id, unsigned short * _pus_count);

/**
 * @brief function to get interface name according to id
 *
 * @param[in] _us_interface_id interface id
 * @param[out] _pc_name interface name
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_interface_name_f(unsigned short _us_interface_id, char * _pc_name);

/**
 * @brief function to get interface ip according to id interface and address id
 *
 * @param[in] _us_interface_id interface id
 * @param[in] _us_address_id address id
 * @param[out] _puc_ip array of 4 bytes (ipv4)
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_interface_ip_f(unsigned short _us_interface_id, unsigned short _us_address_id, unsigned char * _puc_ip);

/**
 * @brief function to set interface ip according to id interface and address id
 *
 * @param[in] _us_interface_id interface id
 * @param[in] _us_address_id address id
 * @param[out] _puc_ip array of 4 bytes (ipv4)
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_interface_ip_f(unsigned short _us_interface_id, unsigned short _us_address_id, unsigned char * _puc_ip);

/**
 * @brief function to get interface ip according to interface name and address id
 *
 * @param[in] _us_interface_id interface id
 * @param[in] _us_address_id address id
 * @param[out] _puc_ip array of 4 bytes (ipv4)
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_interface_ip_by_name_f(char * _pc_interface_name, unsigned short _us_address_id, unsigned char * _puc_ip);

/**
 * @brief function to set interface ip according to interface name and address id
 *
 * @param[in] _us_interface_id interface id
 * @param[in] _us_address_id address id
 * @param[out] _puc_ip array of 4 bytes (ipv4)
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_interface_ip_by_name_f(char * _pc_interface_name, unsigned short _us_address_id, unsigned char * _puc_ip);
	
/**
 * @brief function to get flag is loopback
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _puc_flag flag output
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_is_loopback_f(char * _pc_interface_name, unsigned char * _puc_flag);

/**
 * @brief function to get flag is up
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _puc_flag flag output
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_is_up_f(char * _pc_interface_name, unsigned char * _puc_flag);

/**
 * @brief function to set flag is up
 *
 * @param[in] _pc_interface_name interface name
 * @param[in] _uc_flag flag intput
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_is_up_f(char * _pc_interface_name, unsigned char _uc_flag);

/**
 * @brief function to get flag is point to point
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _puc_flag flag output
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_is_point_to_point_f(char * _pc_interface_name, unsigned char * _puc_flag);

/**
 * @brief function to get flag support multicast
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _puc_flag flag output
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_support_multicast_f(char * _pc_interface_name, unsigned char * _puc_flag);

/**
 * @brief function to get flag support multicast by interface id
 *
 * @param[in] _us_interface_id interface id
 * @param[out] _puc_flag flag output
 *
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_support_multicast_by_id_f(unsigned short _us_interface_id, unsigned char * _puc_flag);

/**
 * @brief function to get hw address
 *
 * @param[in] _pc_interface_name interface name
 * @param[out] _puc_hw_address array of 6 bytes (mac addr)
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_hw_addr_f(char * _pc_interface_name, unsigned char * _puc_hw_address);

/**
 * @brief function to get host by addr
 *
 * @param[in] _puc_ip_address array of 4 bytes (ipv4 addr)
 * @param[out] _pc_interface_name interface name
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_host_by_addr_f(unsigned char * _puc_ip_address, char * _pc_interface_name);

/**
 * @brief function to get interface id by addrs 
 *
 * @param[in] _puc_ip_address array of 4 bytes (ipv4 addr)
 * @param[out] _pc_interface_name interface name
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_get_interface_id_by_addr_f(unsigned char * _puc_ip_address, unsigned short * _pus_interface_id);

/** 
 * @brief generic initialize function
 * 
 * @see NETIF_status_t
 */
NETIF_status_t NETIF_initialize_f(void);

/**
 * @brief function to set the default interface
 *
 * @param[in] _us_if_id default interface id
 * 
 * @return @see NETIF_status_t
 */
NETIF_status_t NETIF_set_default_interface_f(unsigned short _us_if_id);

#endif /* NETWORK_INTERFACE_H */
