/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#ifndef NET_ADAPT_DELEGATE_TASK_H
#define NET_ADAPT_DELEGATE_TASK_H

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "freertos/semphr.h"

/** @brief list of module constant */
typedef enum
{
	NET_ADAPT_NO_ERROR,
	NET_ADAPT_BE_PATIENT,
	NET_ADAPT_ERROR,
}NET_ADAPT_status_t;

#define NET_ADAPT_YIELD(tick)                                           //taskYIELD()
#define NET_ADAPT_MAILBOX_DECLARE(name)                                 QueueHandle_t (name)
#define NET_ADAPT_INIT_MAILBOX(mailbox,size_of_msg,number_of_msg)       mailbox = xQueueCreate((number_of_msg),(size_of_msg))  
#define NET_ADAPT_DELAY(ms)                                             (vTaskDelay((ms) / portTICK_RATE_MS))             
#define NET_ADAPT_INFITIVE_TIMEOUT                                      (portMAX_DELAY)
#define NET_ADAPT_MAILBOX_EMPTY(mailbox)                                (uxQueueMessagesWaiting((mailbox)) == 0)

#define NET_ADAPT_SYNC_OBJ_DECLARE(name)                                SemaphoreHandle_t (name)
#define NET_ADAPT_INIT_SYNC_OBJ(sync_obj)								vSemaphoreCreateBinary(sync_obj)
#define NET_ADAPT_LOCK_SYNC_OBJ(sync_obj)								xSemaphoreTake(sync_obj, portMAX_DELAY)
#define NET_ADAPT_UNLOCK_SYNC_OBJ(sync_obj)								xSemaphoreGive(sync_obj)


/** @brief delay to repost a net event */
#define NET_ADAPT_DELEGATE_EVENT_REPOST_DELAY		 (25)

/**
 * @brief function to initialise the NET delegate module adaptation
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_init_f(void);

/**
 * @brief function to get the caller java thread id
 *
 * @param[in,out] _pul_java_thread pointer on java thread
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_get_java_thread(unsigned long * const _pul_java_thread);

/**
 * @brief function to suspend the current java thread 
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_suspend_java_thread(void);

/**
 * @brief function to resume a java thread
 *
 * @param[in] _ul_java_thread java thread id
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_resume_java_thread(const unsigned long _ul_java_thread);

/**
 * @brief this function is call to post event action
 *
 * @param[in,out] _pv_mailbox pointer on mailbox
 * @param[in]     _pv_event event to post
 * @param[in]     _ul_timeout timeout to wait to post event
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_post_event_action_f(void * const _pv_mailbox,
                                                 void * const _pv_event,
                                                 const unsigned long _ul_timeout);

/**
 * @brief this function is call to wait event action
 *
 * this function must idle the task until an event is present in mailbox
 *
 * @param[in,out] _pv_mailbox pointer on mailbox
 * @param[out]    _ppv_event event to post
 * @param[in]     _ul_timeout timeout to wait to post event
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wait_event(void * const _pv_mailbox,
                                        void ** const _ppv_event,
                                        const unsigned long _ul_timeout);
																						
/**
 * @brief this function to start interface
 *
 * @param[in,out] _pc_buffer_in interface name
 * @param[in] _i_interface_name_length_in interface name buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_start_f(char * const _pc_interface_name_in,
																			int _i_interface_name_length_in,
																			short * const _ps_error_out);
																						
/**
 * @brief this function to stop interface
 *
 * @param[in,out] _pc_buffer_in interface name
 * @param[in] _i_interface_name_length_in interface name buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_stop_f(char * const _pc_interface_name_in,
																		int _i_interface_name_length_in,
                                    short * const _ps_error_out);
																						
/**
 * @brief this function to apply configuration
 *
 * @param[in,out] _pc_buffer_in interface name
 * @param[in] _i_interface_name_length_in interface name buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_apply_configuration_f(char * const _pc_interface_name_in,
																										int _i_interface_name_length_in,
																										short * const _ps_error_out);
												   
/**
 * @brief this function to disable soft ap
 *
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_disable_soft_ap_f(short * const _ps_error_out);

/**
 * @brief this function to enable soft ap
 *
 * @param[in,out] _pc_ssid_in wifi ssid
 * @param[in] _i_ssid_length_in ssid buffer length
 * @param[in,out] _pc_passphrase_in wifi passphrase
 * @param[in] _i_passphrase_length_in passphrase buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_enable_soft_ap_f(char * const _pc_ssid_in,
																										int _i_ssid_length_in,
																										char * const _pc_passphrase_in,
																										int _i_passphrase_length_in,
																										short * const _ps_error_out);
																										
/**
 * @brief this function set name of soft ap
 *
 * @param[in,out] _pc_name_in device name
 * @param[in] _i_name_length_in device name buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_set_name_soft_ap_f(char * const _pc_name_in,
																					           int _i_name_length_in,
																										 short * const _ps_error_out);

/**
 * @brief this function to get bssid
 *
 * @param[in,out] _pc_bssid_in wifi bssid
 * @param[in] _i_bssid_length_in bssid buffer length																										
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_bssid_f(char * const _pc_bssid_in,
																							int _i_bssid_length_in,
												                      short * const _ps_error_out);

/**
 * @brief this function to get channel
 *
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_channel_f(short * const _ps_error_out);

/**
 * @brief this function to get rssi
 *
 * @param[in,out] _pf_rssi_in wifi rssi
 * @param[in] _i_rssi_length_in rssi buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_rssi_f(float * const _pf_rssi_in,
																							int _i_rssi_length_in,
																							short * const _ps_error_out);

/**
 * @brief this function to get security mode
 *
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_security_mode_f(short * const _ps_error_out);

/**
 * @brief this function to get ssid
 *
 * @param[in,out] _pc_ssid_in wifi ssid
 * @param[in] _i_ssid_length_in ssid buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_ssid_f(char * const _pc_ssid_in,
																							int _i_ssid_length_in,
																							short * const _ps_error_out);

/**
 * @brief this function to get wps modes
 *
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_wps_modes_f(short * const _ps_error_out);
																							
/**
 * @brief this function to join wifi network
 *
 * @param[in,out] _pc_ssid_in wifi ssid
 * @param[in] _i_ssid_length_in ssid buffer length
 * @param[in,out] _pc_passphrase_in wifi passphrase
 * @param[in] _i_ssid_length_in passphrase buffer length
 * @param[in] _i_security_mode_in the security mode
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_join_f(char * const _pc_ssid_in,
																					int _i_ssid_length_in,
																					char * const _pc_passphrase_in,
																					int _i_passphrase_length_in,
																					int _i_security_mode_in,
																					short * const _ps_error_out);

 /**
 * @brief this function to leave wifi network
 *
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_leave_f(short * const _ps_error_out);

/**
 * @brief Get the number of available access points
 *
 * @param[in,out] _ps_error_out The number of access points in the list if success, < 0 otherwise
 *
 * @return @see NET_ADAPT_status_t
 */																				 
NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_count_f(short * const _ps_error_out);

/**
 * @brief this function to get ssid
 *
 * @param[in] _i_index_in index in the access point array
 * @param[in,out] _puc_ssid_in wifi ssid
 * @param[in] _i_ssid_length_in ssid buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_ssid_f(int _i_index_in,
																								unsigned char * const _puc_ssid_in,
																								int _i_ssid_length_in,
																								short * const _ps_error_out);
																							
/**
 * @brief this function to get bssid
 *
 * @param[in] _i_index_in index in the access point array
 * @param[in,out] _puc_bssid_in wifi bssid
 * @param[in] _i_bssid_length_in bssid buffer length																										
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_bssid_f(int _i_index_in,
																									unsigned char * const _puc_bssid_in,
																									int _i_bssid_length_in,
																									short * const _ps_error_out);

/**
 * @brief this function to get channel
 *
 * @param[in] _i_index_in index in the access point array
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_channel_f(int _i_index_in,
																										short * const _ps_error_out);

/**
 * @brief this function to get rssi
 *
 * @param[in] _i_index_in index in the access point array
 * @param[in,out] _pf_rssi_in wifi rssi
 * @param[in] _i_rssi_length_in rssi buffer length
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_rssi_f(int _i_index_in,
																								float * const _pf_rssi_in,
																								int _i_rssi_length_in,
																								short * const _ps_error_out);

/**
 * @brief this function to get security mode
 *
 * @param[in] _i_index_in index in the access point array
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_security_mode_f(int _i_index_in,
																													short * const _ps_error_out);

/**
 * @brief this function to get wps modes
 *
 * @param[in] _i_index_in index in the access point array
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_wifi_get_ap_wps_modes_f(int _i_index_in,
																											short * const _ps_error_out);

/**
 * @brief this function get the client state
 *
 * @param[in,out] _ps_error_out error or result
 *
 * @return @see NET_ADAPT_status_t
 */
NET_ADAPT_status_t NET_ADAPT_get_client_state_f(short * const _ps_error_out);

#endif /* NET_ADAPT_DELEGATE_TASK_H */
