/*
 * C
 *
 * Copyright 2017-2018 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */

#ifndef  ASYNC_NETCONN_H
#define  ASYNC_NETCONN_H

/**
 * @file
 * @brief Asynchronous network API based on netconn
 * @author MicroEJ Developer Team
 * @version 0.0.1
 * @date 22 May 2018
 */

#include <stdint.h>
#include <sni.h>
#include "async_netconn_configuration.h"

#ifdef __cplusplus
	extern "C" {
#endif

/** @brief Netconn operations list. */
typedef enum {
  SELECT_READ,
  SELECT_WRITE
} select_operation;

/**
 * @brief External function used to retrieve currentTime (defined in LLMJVM)
 */
extern int64_t LLMJVM_IMPL_getCurrentTime__Z(uint8_t system);

/**
 * Gets the system time in milliseconds (int64_t). This time is independent from any user considerations
 * and cannot be changed.
 */
#define async_netconn_get_current_time_ms()    LLMJVM_IMPL_getCurrentTime__Z(1) // 1 means that system time is required

/**
 * @brief Check the netconn for the given file descriptor and operation without blocking.
 *
 * @param[in] fd the file descriptor.
 * @param[in] operation the operation (read or write) we want to monitor with the netconn callbacks.
 *
 * @return netconn check result for the fd and operation
 */
int32_t non_blocking_select(int32_t fd, select_operation operation);

/**
 * @brief It uses netconn callbacks to wait for an operation for the given file descriptor.
 * This function will suspend the execution of the current Java thread using
 * SNI_suspendCurrentJavaThreadWithCallback(). Once the netconn callbacks are triggered the Java
 * thread is resumed and the given SNI callback is called.
 *
 * @param[in] fd the file descriptor.
 * @param[in] operation the operation (read or write) we want to monitor with the netconn callbacks.
 * @param[in] timeout_ms timeout in millisecond
 * @param[in] callback the SNI callback to call when the Java thread is resumed or timeout occurs.
 *
 * @return 0 on success, -1 on failure.
 */
int32_t async_netconn(int32_t fd, select_operation operation, int64_t timeout_ms, SNI_callback *callback);

/**
 * @brief Initializes the requests FIFOs.
 * This function must be called prior to any call of async_netconn().
 * It can be called several times.
 */
void async_netconn_request_fifo_init(void);

/**
 * @brief The entry point for the async_netconn task.
 * This function must be called from a dedicated task.
 */
void async_netconn_task_main(void);

/**
 * @brief This function checks if the async_netconn task is waiting for an event
 * and notifies the task in that case
 */
void async_netconn_try_notify_events(void);

/**
 * @brief Initialize the async_netconn component. This function must be called prior to any call of
 * async_netconn().
 *
 * @return 0 on success, -1 on failure.
 */
int32_t async_netconn_init(void);

/**
 * @brief Enter critical section for the async_netconn component.
 */
void async_netconn_lock(void);

/**
 * @brief Exit critical section for the async_netconn component.
 */
void async_netconn_unlock(void);

/**
 * @brief Wait for an async_netconn event.
 *
 * @param[in] timeout timeout in millisecond to wait for an event
 */
void async_netconn_wait_events(int32_t timeout);

/**
 * @brief Notify the caller than an async_netconn event was received.
 */
void async_netconn_notify_events(void);

#ifdef __cplusplus
	}
#endif

#endif // ASYNC_NETCONN_H
