/*
 * C
 *
 * Copyright 2017-2018 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */

#ifndef  ASYNC_NETCONN_SOCKETS_H
#define  ASYNC_NETCONN_SOCKETS_H

#include <lwip/sockets.h>

/**
 * @file
 * @brief Asynchronous network API based on netconn
 * @author MicroEJ Developer Team
 * @version 0.0.1
 * @date 22 May 2018
 */

#include <stdint.h>
#include <stddef.h>
#include <sni.h>
#include "async_netconn_configuration.h"

#ifdef __cplusplus
	extern "C" {
#endif

/**
 * @brief netconn socket structure
 */
typedef struct async_netconn_sock {
	/** sockets currently are built on netconns, each socket has one netconn */
	struct netconn *conn;
	/** data that was left from the previous read */
	void *lastdata;
	/** offset in the data that was left from the previous read */
	uint16_t lastoffset;
	/** number of times data was received, set by async_netconn_event_callback() */
	int16_t rcvevent;
	/** number of times data was ACKed (free send buffer), set by async_netconn_event_callback() */
	uint16_t sendevent;
	/** error happened for this socket, set by async_netconn_event_callback() */
	uint16_t errevent;
	/** last error that occurred on this socket (in fact, all our errnos fit into an uint8_t) */
	uint8_t err;
} async_netconn_sock;

/**
 * @brief Socket timeout cache structure
 */
typedef struct async_netconn_sock_timeout {
	int32_t timeout;
	int64_t absolute_timeout;
} async_netconn_sock_timeout;

/**
 * @brief Accept a connection on a socket
 *
 * @param[in] s the socket file descriptor
 * @param[out] addr a pointer to a sockaddr structure
 * @param[inout] addrlen a value-result argument: the caller must initialize it to contain the size
 *               (in bytes) of the structure pointed to by addr; on return it will contain the
 *               actual size of the peer address
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_accept(int32_t s, struct sockaddr *addr, socklen_t *addrlen);

/**
 * @brief Bind a name to a socket
 *
 * @param[in] s the socket file descriptor
 * @param[in] name a pointer to a sockaddr structure
 * @param[in] namelen specifies the size, in bytes, of the address structure pointed to by name
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_bind(int32_t s, const struct sockaddr *name, socklen_t namelen);

/**
 * @brief Close a file descriptor
 *
 * @param[in] s the socket file descriptor
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_close(int32_t s);

/**
 * @brief Initiate a connection on a socket
 *
 * @param[in] s the socket file descriptor
 * @param[in] name a pointer to a sockaddr structure
 * @param[in] namelen specifies the size, in bytes, of the address structure pointed to by name
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_connect(int32_t s, const struct sockaddr *name, socklen_t namelen);

/**
 * @brief Listen for connections on a socket
 *
 * @param[in] s the socket file descriptor
 * @param[in] backlog the maximum length to which the queue of pending connections for s may grow
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_listen(int32_t s, int32_t backlog);

/**
 * @brief Receive a message from a socket
 *
 * @param[in] s the socket file descriptor
 * @param[out] mem the buffer where the message is placed
 * @param[in] len the maximum size that can be received in mem buffer
 * @param[in] flags receive flags
 * @param[out] from the sockaddr from where the message is received
 * @param[out] fromlen the length of the sockaddr
 *
 * @return On success, the length of the received message.  On error, -1 is returned
 */
int32_t async_netconn_recvfrom(int32_t s, void *mem, size_t len, int32_t flags, struct sockaddr *from, socklen_t *fromlen);

/**
 * @brief Receive a message from a socket
 *
 * @param[in] s the socket file descriptor
 * @param[out] mem the buffer where the message is placed
 * @param[in] len the maximum size that can be received in mem buffer
 * @param[in] flags receive flags
 *
 * @return On success, the length of the received message.  On error, -1 is returned
 */
int32_t async_netconn_recv(int32_t s, void *mem, size_t len, int32_t flags);

/**
 * @brief Send a message on a socket
 *
 * @param[in] s the socket file descriptor
 * @param[in] data the buffer containing the data to be sent
 * @param[in] size the data size
 * @param[in] flags send flags
 *
 * @return On success, the length of the sent message.  On error, -1 is returned
 */
int32_t async_netconn_send(int32_t s, const void *data, size_t size, int32_t flags);

/**
 * @brief Send a message on a socket
 *
 * @param[in] s the socket file descriptor
 * @param[in] data the buffer containing the data to be sent
 * @param[in] size the data size
 * @param[in] flags send flags
 * @param[in] to sockaddr where to send the message
 * @param[in] tolen sockaddr length
 *
 * @return On success, the length of the sent message.  On error, -1 is returned
 */
int32_t async_netconn_sendto(int32_t s, const void *data, size_t size, int32_t flags, const struct sockaddr *to, socklen_t tolen);

/**
 * @brief Creates an endpoint for communication and returns a file descriptor that refers to that
 *        endpoint
 *
 * @param[in] domain specifies a communication domain
 * @param[in] type specifies the communication semantics
 * @param[in] protocol specifies a particular protocol to be used with the socket
 *
 * @return On success, a file descriptor for the new socket is returned.  On error, -1 is returned
 */
int32_t async_netconn_socket(int32_t domain, int32_t type, int32_t protocol);

/**
 * @brief Shut down part of a full-duplex connection
 *
 * @param[in] s the socket file descriptor
 * @param[in] how if how is SHUT_RD, further receptions will be disallowed.
 *            If how is SHUT_WR, further transmissions will be disallowed.
 *            If how is SHUT_RDWR, further receptions and transmissions will be disallowed
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_shutdown(int32_t s, int32_t how);

/**
 * @brief Get name of connected peer socket
 *
 * @param[in] s the socket file descriptor
 * @param[out] name a pointer to a sockaddr structure
 * @param[inout] namelen should be initialized to indicate the amount of space (in bytes) pointed to
 *               by name. On return it contains the actual size of the peer name
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_getpeername(int32_t s, struct sockaddr *name, socklen_t *namelen);

/**
 * @brief Get socket name
 *
 * @param[in] s the socket file descriptor
 * @param[out] name a pointer to a sockaddr structure
 * @param[inout] namelen should be initialized to indicate the amount of space (in bytes) pointed to
 *               by name. On return it contains the actual size of the socket name
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
int32_t async_netconn_getsockname(int32_t s, struct sockaddr *name, socklen_t *namelen);

/**
 * @brief Get options on sockets
 *
 * @param[in] s the socket file descriptor
 * @param[in] level the level at which the option resides
 * @param[in] optname the name of the option
 * @param[out] optval identify a buffer in which the value for the requested option(s) are to be
 *             returned
 * @param[inout] optlen is a value-result argument, initially containing the size of the buffer
 *               pointed to by optval, and modified on return to indicate the actual size of the
 *               value returned
 *
 * @return On success, zero is returned for the standard options.  On error, -1 is returned
 */
int32_t async_netconn_getsockopt(int32_t s, int32_t level, int32_t optname, void *optval, socklen_t *optlen);

/**
 * @brief Set options on sockets
 *
 * @param[in] s the socket file descriptor
 * @param[in] level the level at which the option resides
 * @param[in] optname the name of the option
 * @param[in] optval used to access option values
 * @param[in] optlen used to access option values
 *
 * @return On success, zero is returned for the standard options.  On error, -1 is returned
 */
int32_t async_netconn_setsockopt(int32_t s, int32_t level, int32_t optname, const void *optval, socklen_t optlen);

/**
 * @brief Control device
 *
 * @param[in] s the socket file descriptor
 * @param[in] cmd the command
 * @param[in] argp the command arguments
 *
 * @return On success, zero is returned for the standard options.  On error, -1 is returned
 */
int32_t async_netconn_ioctl(int32_t s, long cmd, void *argp);

/**
 * @brief Manipulate file descriptor
 *
 * @param[in] s the socket file descriptor
 * @param[in] cmd the command
 * @param[in] val the value of the command
 *
 * @return On success, zero is returned for the standard options.  On error, -1 is returned
 */
int32_t async_netconn_fcntl(int32_t s, int32_t cmd, int32_t val);

/**
 * @brief Check for events on a given socket
 *
 * @param[in] s the socket file descriptor
 * @param[in] read_events 1 if the file descriptor is checked for read events, 0 otherwise
 * @param[in] write_events 1 if the file descriptor is checked for write events, 0 otherwise
 *
 * @return 1 if there are events (read/write) available on the socket, otherwise 0
 */
int32_t async_netconn_check_events(int32_t s, uint8_t read_events, uint8_t write_events);

/**
 * @brief Initialize the socket timeout cache
 */
void async_netconn_init_socket_timeout_cache(void);

/**
 * @brief Get a socket timeout from the cache
 *
 * @param[in] s the socket file descriptor
 *
 * @return socket timeout on success, a negative value on error.
 */
int32_t async_netconn_get_socket_timeout_from_cache(int32_t s);

/**
 * @brief Get a socket absolute timeout from the cache (time when the async_netconn task should unblock the select)
 *
 * @param[in] s the socket file descriptor
 *
 * @return socket absolute timeout on success, a negative value on error.
 */
int64_t async_netconn_get_socket_absolute_timeout_from_cache(int32_t s);

/**
 * @brief Set the given socket's timeout in the cache
 *
 * @param[in] s the socket file descriptor
 * @param[in] timeout socket timeout
 */
void async_netconn_set_socket_timeout_in_cache(int32_t s, int32_t timeout);

/**
 * @brief Set the given socket's absolute timeout in the cache (time when the async_netconn task should unblock the select)
 *
 * @param[in] s the socket file descriptor
 * @param[in] absolute_timeout socket absolute timeout
 */
void async_netconn_set_socket_absolute_timeout_in_cache(int32_t s, int64_t absolute_timeout);

/**
 * @brief Remove a socket timeout from the cache
 *
 * @param[in] s the socket file descriptor
 */
void async_netconn_rem_socket_timeout_from_cache(int32_t s);

/**
 * @brief Return the last error on the socket
 *
 * @param[in] s the socket file descriptor
 *
 * @return The socket's last error
 */
int32_t async_netconn_errno(int32_t s);

#ifdef __cplusplus
	}
#endif

#endif // ASYNC_NETCONN_SOCKETS_H
