/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include "freertos/FreeRTOS.h"
#include "freertos/semphr.h"
#include "pool.h"
#include "NETIF_interface.h"
#include "NET_delegate_task.h"

/** @brief enable assert for this module */
#ifdef ENABLE_ASSERT_NETIF
#define ASSERT_NETIF_INTERFACE(x) do{if(!(x)){printf("assert_netif\n");}}while(0) 
#else
#define ASSERT_NETIF_INTERFACE(x) ((void) 0)
#endif

/** @brief define max network interface registerable */
#define MAX_INTERFACE     (3)

/** @brief define usable macro to enter and acxit of critical section */
#define ENTER_CRITICAL_SECTION     (xSemaphoreTake(gst_sem, portMAX_DELAY))
#define END_OF_CRITICAL_SECTION    (xSemaphoreGive(gst_sem))

/** @brief private pool interface */
static NETIF_t gpst_pool_netif[MAX_INTERFACE];
static POOL_item_status_t gpst_pool_netif_item_status[MAX_INTERFACE];
static POOL_ctx_t gst_pool_netif_ctx =
{
	gpst_pool_netif,
	gpst_pool_netif_item_status,
	sizeof(NETIF_t),
	sizeof(gpst_pool_netif)/sizeof(NETIF_t)
};

/** @brief private mutex on driver */
static SemaphoreHandle_t gst_sem;

/** @brief private default interface id */
static unsigned short guc_default_id_interface;

/**
 * @brief function to get any flag
 *
 * @param[in] _pc_name interface name
 * @param[out] _puc_flag flag output 
 * @param[in] _uc_option option id
 * 
 * @return @see NETIF_status_t
 */
static NETIF_status_t get_flag_f(char * _pc_name, unsigned char * _puc_flag, unsigned char _uc_option);

/**
 * @brief function to get any flag
 *
 * @param[in] _us_interface_id interface id
 * @param[out] _puc_flag flag output
 * @param[in] _uc_option option id
 *
 * @return @see NETIF_status_t
 */
static NETIF_status_t get_flag_by_id_f(unsigned short _us_interface_id, unsigned char * _puc_flag, unsigned char _uc_option);

/**
 * @brief function to set any flag
 *
 * @param[in] _pc_name interface name
 * @param[in] _uc_flag flag input 
 * @param[in] _uc_option option id
 * 
 * @return @see NETIF_status_t
 */
static NETIF_status_t set_flag_f(char * _pc_name, unsigned char _uc_flag, unsigned char _uc_option);

/**
 * @brief function to get index interface on pool according to it name
 *
 * @param[in]  _pc_if_name interface name
 * @param[out] _pus_id index interface
 *
 * @return true if success else false
 */
static bool get_index_if_from_name_f(char * _pc_if_name, unsigned short * _pus_id);


NETIF_status_t NETIF_init_f(void)
{	
	NETIF_status_t e_return;
	
	/* create mutex */
	gst_sem = xSemaphoreCreateMutex();
	
	/* init default interface */
	guc_default_id_interface = 0;
	
	/* test if success creation */
	if (gst_sem != NULL)
	{
		e_return = NETIF_NO_ERROR;
		
		/* Initialise the private pool */
		memset(gpst_pool_netif, 0, sizeof(gpst_pool_netif));
		memset(gpst_pool_netif_item_status, POOL_FREE, sizeof(gpst_pool_netif_item_status));
		
		for (unsigned short us_i = 0;us_i<MAX_INTERFACE;us_i++)
		{
			gpst_pool_netif[us_i].us_id = us_i;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}


NETIF_status_t NETIF_is_started_f(char * _pc_interface_name, unsigned char * _puc_is_started)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_is_started != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		if((_puc_is_started != NULL) && (gpst_pool_netif[us_i].pp_is_started_f != NULL))
		{
			e_return = gpst_pool_netif[us_i].pp_is_started_f(_puc_is_started);
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_stop_f(char * _pc_interface_name)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		if (gpst_pool_netif[us_i].pp_stop_f != NULL)
		{
			// stop the network interface
			e_return = gpst_pool_netif[us_i].pp_stop_f();
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_apply_configuration_f(char * _pc_interface_name)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		if (gpst_pool_netif[us_i].pp_apply_configuration_f != NULL)
		{
			e_return = gpst_pool_netif[us_i].pp_apply_configuration_f();
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_dns_f(char * _pc_interface_name, unsigned char * _puc_dns)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_dns != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		memcpy(_puc_dns, gpst_pool_netif[us_i].puc_dns_ip, sizeof(gpst_pool_netif[us_i].puc_dns_ip));		
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_set_dns_f(char * _pc_interface_name, unsigned char * _puc_dns)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_dns != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		memcpy(gpst_pool_netif[us_i].puc_dns_ip, _puc_dns, sizeof(gpst_pool_netif[us_i].puc_dns_ip));		
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_set_gateway_f(char * _pc_interface_name, unsigned char * _puc_gateway)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_gateway != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		memcpy(gpst_pool_netif[us_i].puc_gateway_ip_addr, _puc_gateway, sizeof(gpst_pool_netif[us_i].puc_gateway_ip_addr));		
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_gateway_f(char * _pc_interface_name, unsigned char * _puc_gateway)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_gateway != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		memcpy(_puc_gateway, gpst_pool_netif[us_i].puc_gateway_ip_addr, sizeof(gpst_pool_netif[us_i].puc_gateway_ip_addr));		
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_set_netmask_f(char * _pc_interface_name, unsigned char * _puc_netmask)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_netmask != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		memcpy(gpst_pool_netif[us_i].puc_netmask, _puc_netmask , sizeof(gpst_pool_netif[us_i].puc_netmask));		
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_netmask_f(char * _pc_interface_name, unsigned char * _puc_netmask)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_netmask != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		memcpy(_puc_netmask, gpst_pool_netif[us_i].puc_netmask, sizeof(gpst_pool_netif[us_i].puc_netmask));		
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_start_f(char * _pc_interface_name)
{	
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		if (gpst_pool_netif[us_i].pp_start_f != NULL)
		{
			e_return = gpst_pool_netif[us_i].pp_start_f();
		}

	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_registered_interface_f(NETIF_t ** _ppst_registered_interface)
{
	NETIF_status_t e_return;
	POOL_status_t e_pool_return;
	
	/* test entry function */
	if (_ppst_registered_interface != NULL)
	{
		/* allocate in pool */
		e_pool_return = POOL_reserve_f(&gst_pool_netif_ctx, (void**)_ppst_registered_interface);

		/* test if alloc is ok */
		if (e_pool_return == POOL_NO_ERROR)
		{
			e_return = NETIF_NO_ERROR;
		}
		else
		{
			e_return = NETIF_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}


NETIF_status_t NETIF_unregister_f(NETIF_t * const _pst_interface)
{
	NETIF_status_t e_return;
	
	/* release in pool */
	if (POOL_free_f(&gst_pool_netif_ctx, (void*)_pst_interface) != POOL_NO_ERROR)
	{
		e_return = NETIF_ERROR;
	}
	else
	{
		e_return = NETIF_NO_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_count_registered_interface_f(unsigned short * _pus_count)
{
	NETIF_status_t e_return;
	
	/* test entry function */
	if (_pus_count != NULL)
	{
		*_pus_count = 0;
		
		/* walk on array status in pool and count item not free */
		for (int i=0;i<MAX_INTERFACE;i++)
		{
			if (gpst_pool_netif_item_status[i] != POOL_FREE)
			{
				*_pus_count += 1;
			}
		}
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_count_address_on_interface_f(unsigned short _us_interface_id, unsigned short * _pus_count)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_pus_count != NULL) && (_us_interface_id < MAX_INTERFACE))
	{
		*_pus_count = 0;
		
		/* walk on array status in pool and count item not free */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			if (gpst_pool_netif[us_i].us_id == _us_interface_id)
			{
				*_pus_count = NETIF_MAX_IP_ADDRESS;
				
				break; /* don't loose time exit */
			}
		}
		
		/* test if interface found */ 
		if (us_i >= MAX_INTERFACE)
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_interface_name_f(unsigned short _us_interface_id, char * _pc_name)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_pc_name != NULL) && (_us_interface_id < MAX_INTERFACE))
	{	
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			if (gpst_pool_netif[us_i].us_id == _us_interface_id)
			{
				strncpy(_pc_name, gpst_pool_netif[us_i].pc_name, sizeof(gpst_pool_netif[us_i].pc_name));
				break; /* don't loose time exit */
			}
		}
		
		/* test if interface found */ 
		if (us_i >= MAX_INTERFACE)
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}


NETIF_status_t NETIF_get_interface_ip_by_name_f(char * _pc_interface_name, unsigned short _us_address_id, unsigned char * _puc_ip)
{
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_ip != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		e_return = NETIF_get_interface_ip_f(us_i,_us_address_id,_puc_ip);
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_set_interface_ip_by_name_f(char * _pc_interface_name, unsigned short _us_address_id, unsigned char * _puc_ip)
{
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_ip != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		e_return = NETIF_set_interface_ip_f(us_i,_us_address_id,_puc_ip);
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_interface_ip_f(unsigned short _us_interface_id, unsigned short _us_address_id, unsigned char * _puc_ip)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_puc_ip != NULL) && (_us_interface_id < MAX_INTERFACE) && (_us_address_id <= NETIF_MAX_IP_ADDRESS))
	{
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			if (gpst_pool_netif[us_i].us_id == _us_interface_id)
			{
				memcpy(_puc_ip,&gpst_pool_netif[us_i].ppuc_ip_addr[_us_address_id],sizeof(gpst_pool_netif[us_i].ppuc_ip_addr[_us_address_id]));

				break; /* don't loose time exit */
			}
		}
		
		/* test if interface found */ 
		if (us_i >= MAX_INTERFACE)
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_set_interface_ip_f(unsigned short _us_interface_id, unsigned short _us_address_id, unsigned char * _puc_ip)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_puc_ip != NULL) && (_us_interface_id < MAX_INTERFACE) && (_us_address_id <= NETIF_MAX_IP_ADDRESS))
	{
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			if (gpst_pool_netif[us_i].us_id == _us_interface_id)
			{
				memcpy(&gpst_pool_netif[us_i].ppuc_ip_addr[_us_address_id],_puc_ip,sizeof(gpst_pool_netif[us_i].ppuc_ip_addr[_us_address_id]));
				
				break; /* don't loose time exit */
			}
		}
		
		/* test if interface found */ 
		if (us_i >= MAX_INTERFACE)
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_is_static_f(char * _pc_interface_name, unsigned char * _puc_is_static)
{
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	ASSERT_NETIF_INTERFACE(_puc_is_static != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		*_puc_is_static = gpst_pool_netif[us_i].uc_is_static;
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);	
}

NETIF_status_t NETIF_set_is_static_f(char * _pc_interface_name, unsigned char _uc_is_static)
{
	NETIF_status_t e_return = NETIF_ERROR;
	unsigned short us_i;
	
	ASSERT_NETIF_INTERFACE(_pc_interface_name != NULL);
	
	if (get_index_if_from_name_f(_pc_interface_name, &us_i))
	{
		 gpst_pool_netif[us_i].uc_is_static = _uc_is_static;
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);	
}
 
NETIF_status_t NETIF_get_is_loopback_f(char * _pc_interface_name, unsigned char * _puc_flag)
{
	return (get_flag_f(_pc_interface_name, _puc_flag, NETIF_IS_LOOPBACK));
}


NETIF_status_t NETIF_get_is_up_f(char * _pc_interface_name, unsigned char * _puc_flag)
{
	return (get_flag_f(_pc_interface_name, _puc_flag, NETIF_IS_UP));
}

NETIF_status_t NETIF_set_is_up_f(char * _pc_interface_name, unsigned char _uc_flag)
{
	return (set_flag_f(_pc_interface_name, _uc_flag, NETIF_IS_UP));
}

NETIF_status_t NETIF_get_is_point_to_point_f(char * _pc_interface_name, unsigned char * _puc_flag)
{
	return (get_flag_f(_pc_interface_name, _puc_flag, NETIF_IS_POINT_TO_POINT));
}

NETIF_status_t NETIF_get_support_multicast_f(char * _pc_interface_name, unsigned char * _puc_flag)
{
	return (get_flag_f(_pc_interface_name, _puc_flag, NETIF_IS_SUPPORT_BROADCAST));
}

NETIF_status_t NETIF_get_support_multicast_by_id_f(unsigned short _us_interface_id, unsigned char * _puc_flag)
{
	return (get_flag_by_id_f(_us_interface_id, _puc_flag, NETIF_IS_SUPPORT_BROADCAST));
}

NETIF_status_t NETIF_get_hw_addr_f(char * _pc_interface_name, unsigned char * _puc_hw_address)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_puc_hw_address != NULL) && (_pc_interface_name != NULL))
	{
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			if (strcmp(gpst_pool_netif[us_i].pc_name,_pc_interface_name) == 0)
			{
				memcpy(_puc_hw_address,&gpst_pool_netif[us_i].puc_hw_addr, sizeof(gpst_pool_netif[us_i].puc_hw_addr));
				
				break; /* don't loose time exit */
			}
		}
		
		/* test if interface found */ 
		if (us_i >= MAX_INTERFACE)
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_host_by_addr_f(unsigned char * _puc_ip_address, char * _pc_interface_name)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	unsigned short us_j;
	
	/* test entry function */
	if ((_puc_ip_address != NULL) && (_pc_interface_name != NULL))
	{
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			for (us_j=0;us_j<NETIF_MAX_IP_ADDRESS;us_j++)
			{
				if (memcmp(gpst_pool_netif[us_i].ppuc_ip_addr[us_j], _puc_ip_address, NETIF_MAX_IP_LENGTH) == 0)
				{
					/* if address is found in interface copy name to return host */
					strcpy(_pc_interface_name, gpst_pool_netif[us_i].pc_name);
				}
			}
		}
		
		/* test if interface found */ 
		if ((us_i >= MAX_INTERFACE) || (us_j >= NETIF_MAX_IP_ADDRESS))
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

static NETIF_status_t get_flag_f(char * _pc_name, unsigned char * _puc_flag, unsigned char _uc_option)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_pc_name != NULL) && (_puc_flag != NULL) && (_uc_option < NETIF_MAX_OPTION))
	{
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			if (strcmp(gpst_pool_netif[us_i].pc_name,_pc_name) == 0)
			{
				* _puc_flag = (unsigned char)gpst_pool_netif[us_i].puc_flag[_uc_option];
				
				break; /* don't loose time exit */
			}
		}
		
		/* test if interface found */ 
		if (us_i >= MAX_INTERFACE)
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

static NETIF_status_t get_flag_by_id_f(unsigned short _us_interface_id, unsigned char * _puc_flag, unsigned char _uc_option)
{
	NETIF_status_t e_return;

	/* test entry function */
	if ((_us_interface_id < MAX_INTERFACE) && (_puc_flag != NULL) && (_uc_option < NETIF_MAX_OPTION))
	{
		*_puc_flag = (unsigned char)gpst_pool_netif[_us_interface_id].puc_flag[_uc_option];
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}

	return (e_return);
}

static NETIF_status_t set_flag_f(char * _pc_name, unsigned char _uc_flag, unsigned char _uc_option)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_pc_name != NULL) && (_uc_option < NETIF_MAX_OPTION))
	{
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			if (strcmp(gpst_pool_netif[us_i].pc_name,_pc_name) == 0)
			{
				gpst_pool_netif[us_i].puc_flag[_uc_option] = _uc_flag;
				break; /* don't loose time exit */
			}
		}
		
		/* test if interface found */ 
		if (us_i >= MAX_INTERFACE)
		{
			e_return = NETIF_ERROR;
		}
		else
		{
			e_return = NETIF_NO_ERROR;
		}
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_interface_id_by_addr_f(unsigned char * _puc_ip_address, unsigned short * _pus_interface_id)
{
	NETIF_status_t e_return;
	unsigned short us_i;
	
	/* test entry function */
	if ((_puc_ip_address != NULL) && (_pus_interface_id != NULL))
	{
		/* if address == 0.0.0.0 then return the first available media */
		if ((_puc_ip_address[0] == 0) && (_puc_ip_address[1] == 0) &&
			  (_puc_ip_address[2] == 0) && (_puc_ip_address[3] == 0))
		{
		    /* walk on pool */
            for (us_i=0;us_i<MAX_INTERFACE;us_i++)
            {
                for (unsigned short us_j =0;us_j<NETIF_MAX_IP_ADDRESS;us_j++)
                {
                    unsigned char invalid_ip_address[NETIF_MAX_IP_LENGTH] = {0};

                    if (memcmp(gpst_pool_netif[us_i].ppuc_ip_addr[us_j],invalid_ip_address, NETIF_MAX_IP_LENGTH) != 0)
                    {
                        * _pus_interface_id = us_i;

                        return (NETIF_NO_ERROR); /* don't loose time exit */
                    }
                }
            }

            /* No media has a valid ip address != 0.0.0.0, so return 0 media */
			* _pus_interface_id = 0;
			return NETIF_NO_ERROR;
		}
		
		/* walk on pool */
		for (us_i=0;us_i<MAX_INTERFACE;us_i++)
		{
			for (unsigned short us_j =0;us_j<NETIF_MAX_IP_ADDRESS;us_j++)
			{
				if (memcmp(gpst_pool_netif[us_i].ppuc_ip_addr[us_j],_puc_ip_address, NETIF_MAX_IP_LENGTH) == 0)
				{
					* _pus_interface_id = us_i;
					
					return (NETIF_NO_ERROR); /* don't loose time exit */
				}
			}
		}
		
		/* test if interface found */ 
		e_return = NETIF_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_initialize_f(void)
{
	NETIF_status_t e_return = NETIF_NO_ERROR;
	unsigned short us_i;
	
	ENTER_CRITICAL_SECTION;	

	/* walk on pool to initialize all media registered */
	for (us_i=0;us_i<MAX_INTERFACE;us_i++)
	{
		if ((gpst_pool_netif_item_status[us_i] == POOL_USED) && (gpst_pool_netif[us_i].pp_initialize_f() != NETIF_NO_ERROR))
		{
			e_return = NETIF_ERROR;
		}
	}
	
	END_OF_CRITICAL_SECTION;
	
	return (e_return);
}


NETIF_status_t NETIF_set_default_interface_f(unsigned short _us_if_id)
{
	NETIF_status_t e_return;
	
	if (_us_if_id < MAX_INTERFACE)
	{
		guc_default_id_interface = _us_if_id;
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}
	
	return (e_return);
}

NETIF_status_t NETIF_get_default_interface_f(unsigned short * _pus_if_id)
{
	NETIF_status_t e_return;
	
	if (_pus_if_id != NULL)
	{
		*_pus_if_id = guc_default_id_interface;
		e_return = NETIF_NO_ERROR;
	}
	else
	{
		e_return = NETIF_ERROR;
	}

	return (e_return);
}

/**
 * @brief function to get index interface on pool according to it name
 *
 * @param[in]  _pc_if_name interface name
 * @param[out] _pus_id index interface
 *
 * @return true if success else false
 */
static bool get_index_if_from_name_f(char * _pc_if_name, unsigned short * _pus_id)
{
	bool b_return;
	
	ASSERT_NETIF_INTERFACE(_pc_if_name!= NULL);
	ASSERT_NETIF_INTERFACE(_pus_id!= NULL);
	
	unsigned short us_i;

	/* walk on pool */
	for (us_i=0;us_i<MAX_INTERFACE;us_i++)
	{
		/* compare name */
		if (strcmp(gpst_pool_netif[us_i].pc_name,_pc_if_name) == 0)
		{
			break; /* don't loose time exit */
		}
	}
	
	/* test if interface found */ 
	if (us_i >= MAX_INTERFACE)
	{
		b_return = false;
	}
	else
	{
		b_return = true;
	}
	
	/* affect return value */
	*_pus_id = us_i;

	return (b_return);
}
