/*
 * C
 *
 * Copyright 2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <string.h>
#include <stdbool.h>
#include "pool.h"
#include "NET_ADAPT_delegate_task.h"
#include "NET_delegate_task.h"
#include <stdio.h>
#include "LLNET_CONSTANTS.h"

//#define NET_DEBUG
//#define NET_ERROR

#ifdef NET_ERROR
#include <stdio.h>
#endif

#include "WIFI_ESP32_driver.h"

/** @brief private status for initialisation */
enum
{
	NO_INIT,
	INIT_PASSED,
};

#ifdef NET_DEBUG
#define ASSERT_NET(x) do{if(!(x)){printf("assert in NET_delegate_task\n");while(1)}}while(0)
#else
#define ASSERT_NET(x) ((void) 0)
#endif

/** @brief private init toggle */
static unsigned char guc_init_passed = NO_INIT;

/** @brief private mailbox */
static NET_ADAPT_MAILBOX_DECLARE(gpu32_mailbox);

/** @brief private net event sync object */
static NET_ADAPT_SYNC_OBJ_DECLARE(gpu32_net_events_sync_obj);

/** @brief private pool event */
static NET_event_t gpst_pool_event[NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX];
static POOL_item_status_t gpst_pool_event_item_status[NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX];
static POOL_ctx_t gst_pool_event_ctx =
{
	gpst_pool_event,
	gpst_pool_event_item_status,
	sizeof(NET_event_t),
	sizeof(gpst_pool_event)/sizeof(NET_event_t)
};

/** @brief net events dispatcher event definition */
typedef struct
{
	NET_event_t* net_event;
	int64_t net_event_scheduled_time;
} NET_delegate_events_dispatcher_events_t;

static int32_t events_list_nb = 0;
static NET_delegate_events_dispatcher_events_t events_list[NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX];

/** @brief current event processed */
static NET_event_t * gpst_current_event;

/** @brief external function used to retrieve currentTime (same as MicroJvm) */
extern int64_t LLMJVM_IMPL_getCurrentTime__Z(uint8_t system);

static uint8_t NET_is_event_in_events_list(NET_event_t* net_event)
{
	uint8_t found = 0;

	if (net_event!= NULL)
	{
		for(int32_t i = -1; ++i < NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX;)
		{
			if(events_list[i].net_event == net_event)
			{
				// Event found in events list
				found = 1;
				break;
			}
		}
	}

	return found;
}

static void NET_add_events_list(NET_event_t* net_event, int64_t delay)
{
	NET_delegate_events_dispatcher_events_t* free_event = NULL;

	if((events_list_nb < NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX) && (net_event!= NULL))
	{
		for(int32_t i = -1; ++i < NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX;)
		{
			if(events_list[i].net_event == NULL)
			{
				// free event found
				free_event = &events_list[i];
				break;
			}
		}

		if(free_event != NULL)
		{
			// take a look at events ready to be posted
			free_event->net_event = net_event;
			free_event->net_event_scheduled_time = LLMJVM_IMPL_getCurrentTime__Z(1) + delay;
			++events_list_nb;
		}
	}
}

static void NET_remove_events_list(NET_delegate_events_dispatcher_events_t* event)
{
	if(event != NULL)
	{
		event->net_event = NULL;
		--events_list_nb;
	}
}

static NET_delegate_events_dispatcher_events_t* NET_get_events_list()
{
	int64_t currentTime = LLMJVM_IMPL_getCurrentTime__Z(1);
	int64_t max_time_elapsed = 0x8000000000000000;
	int32_t max_time_elapsed_index = -1;
	int32_t i = -1;

	for(i = -1; ++i < NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX;)
	{
		if(events_list[i].net_event != NULL)
		{
			int64_t diff = currentTime - events_list[i].net_event_scheduled_time;
			if(diff > max_time_elapsed)
			{
				max_time_elapsed_index = i;
				max_time_elapsed = diff;
			}
		}
	}
	if(max_time_elapsed_index >= 0)
	{
		return &events_list[max_time_elapsed_index];
	}
	return NULL;
}

NET_status_t NET_init_f(void)
{
	NET_status_t e_return;
	NET_ADAPT_status_t e_adapt_return;

	/* Initialise the private mailbox */
	NET_ADAPT_INIT_MAILBOX(gpu32_mailbox, sizeof(NET_event_t*), NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX);

	/* Initialise the private mailbox */
	NET_ADAPT_INIT_SYNC_OBJ(gpu32_net_events_sync_obj);

	/* Initialise the private pool */
	memset(gpst_pool_event, 0, sizeof(gpst_pool_event));
	memset(gpst_pool_event_item_status, POOL_FREE, sizeof(gpst_pool_event_item_status));

	/* init adaptation layer */
	e_adapt_return = NET_ADAPT_init_f();

	if (e_adapt_return == NET_ADAPT_NO_ERROR)
	{
		/* init passed */
		guc_init_passed = INIT_PASSED;

		e_return = NET_NO_ERROR;
	}
	else
	{
		e_return = NET_ADAPTATION_ERROR_INIT;
	}


	return (e_return);
}

NET_status_t NET_get_event(NET_event_t ** _pst_event)
{
	NET_status_t e_return;
	POOL_status_t e_pool_return;
	LLNET_DEBUG_ASSERT(guc_init_passed == INIT_PASSED);

	/* test init */
	if (guc_init_passed != NO_INIT)
	{
		/* test entry parameter */
		if (_pst_event != NULL)
		{
			/* allocate in pool */
			e_pool_return = POOL_reserve_f(&gst_pool_event_ctx, (void**)_pst_event);

			/* test if alloc is ok */
			if (e_pool_return == POOL_NO_ERROR)
			{
				e_return = NET_NO_ERROR;
			}
			else
			{
				#ifdef NET_ERROR
				printf("NET_ERROR [%s:%d] : get event no space availabe\n", __func__, __LINE__);
				#endif
				e_return = NET_NO_SPACE_AVAILABLE;
			}
		}
		else
		{
				#ifdef NET_ERROR
				printf("NET_ERROR [%s:%d] : get event parameter error\n", __func__, __LINE__);
				#endif
			e_return = NET_PARAMETER_ERROR;
		}
	}
	else
	{
		#ifdef NET_ERROR
		printf("NET_ERROR [%s:%d] : get event module not initialized\n", __func__, __LINE__);
		#endif
		e_return = NET_MODULE_NOT_INITIALIZED;
	}

	return (e_return);
}

NET_status_t NET_release_event(NET_event_t * _pst_event)
{
	NET_status_t e_return;

	/* test init */
	if (guc_init_passed != NO_INIT)
	{
		/* test entry parameter */
		if (_pst_event != NULL)
		{
			/* release in pool */
			if (POOL_free_f(&gst_pool_event_ctx, (void*)_pst_event) != POOL_NO_ERROR)
			{
				#ifdef NET_ERROR
				printf("NET_ERROR [%s:%d] : release event error during freeing\n", __func__, __LINE__);
				#endif
				e_return = NET_ERROR_DURING_FREEING_POOL;
			}
			else
			{
				e_return = NET_NO_ERROR;
			}
		}
		else
		{
			#ifdef NET_ERROR
			printf("NET_ERROR [%s:%d] : release event parameter error\n", __func__, __LINE__);
			#endif
			e_return = NET_PARAMETER_ERROR;
		}
	}
	else
	{
		#ifdef NET_ERROR
		printf("NET_ERROR [%s:%d] : release event module not initialized\n", __func__, __LINE__);
		#endif
		e_return = NET_MODULE_NOT_INITIALIZED;
	}

	return (e_return);
}

NET_status_t NET_post_event(NET_event_t * _pst_event)
{
	NET_status_t e_return;
	NET_ADAPT_status_t e_result;

	if (guc_init_passed != NO_INIT)
	{
		if (_pst_event != NULL)
		{
			/* update java thread into the event */
			e_result = NET_ADAPT_get_java_thread(&_pst_event->ul_id_java_thread);

			if (e_result == NET_ADAPT_NO_ERROR)
			{
				/* suspend java thread caller */
				if (NET_ADAPT_suspend_java_thread() == NET_ADAPT_NO_ERROR)
				{
					#ifdef NET_DEBUG
						printf("NET_DEBUG [%s:%d] : Java thread %lu post event %d \n", __func__, __LINE__, _pst_event->ul_id_java_thread, _pst_event->e_id_event);
					#endif
					e_result = NET_ADAPT_post_event_action_f(&gpu32_mailbox, _pst_event, 0);

					if (e_result == NET_ADAPT_NO_ERROR)
					{
						e_return = NET_NO_ERROR;
					}
					else
					{
						e_return = NET_NO_SPACE_AVAILABLE;
					}
				}
				else
				{
					e_return = NET_CANNOT_SUSPEND_JVM_THREAD;
				}
			}
			else
			{
				e_return = NET_CANNOT_GET_JVM_THREAD;
			}
		}
		else
		{
			#ifdef NET_ERROR
				printf("NET_ERROR [%s:%d] : post event error parameter\n", __func__, __LINE__);
			#endif
			e_return = NET_PARAMETER_ERROR;
		}
	}
	else
	{
		#ifdef NET_ERROR
		printf("NET_ERROR [%s:%d] : post event module not initialized\n", __func__, __LINE__);
		#endif
		e_return = NET_MODULE_NOT_INITIALIZED;
	}

	return (e_return);
}

NET_status_t NET_repost_event(NET_event_t * _pst_event)
{
	NET_status_t e_return;
	NET_ADAPT_status_t e_result;

	if (guc_init_passed != NO_INIT)
	{
		if (_pst_event != NULL)
		{
			#ifdef NET_DEBUG
				printf("NET_DEBUG [%s:%d] : Java thread %lu post event %d \n", __func__, __LINE__, _pst_event->ul_id_java_thread, _pst_event->e_id_event);
			#endif
			e_result = NET_ADAPT_post_event_action_f(&gpu32_mailbox, _pst_event, 0);

			if (e_result == NET_ADAPT_NO_ERROR)
			{
				e_return = NET_NO_ERROR;
			}
			else
			{
				e_return = NET_NO_SPACE_AVAILABLE;
			}
		}
		else
		{
			#ifdef NET_ERROR
				printf("NET_ERROR [%s:%d] : post event error parameter\n", __func__, __LINE__);
			#endif
			e_return = NET_PARAMETER_ERROR;
		}
	}
	else
	{
		#ifdef NET_ERROR
		printf("NET_ERROR [%s:%d] : post event module not initialized\n", __func__, __LINE__);
		#endif
		e_return = NET_MODULE_NOT_INITIALIZED;
	}

	return (e_return);
}

NET_status_t NET_retrieve_event(NET_event_t ** _pst_event, unsigned long _ul_id_java_thread)
{
	NET_status_t e_return;
	unsigned char uc_found = 0;
	int i;

	/* test init */
	if (guc_init_passed != NO_INIT)
	{
		/* test entry parameter */
		if (_pst_event != NULL)
		{
			/* search the java id in all pool event */
			for (i=0;(i<sizeof(gpst_pool_event)/sizeof(NET_event_t)) && (uc_found != 1);i++)
			{
				if (gpst_pool_event[i].ul_id_java_thread == _ul_id_java_thread)
				{
					/* event found */
					uc_found = 1;
				}
			}

			if (uc_found != 0)
			{
				/* return the address of event found
                 * due to the loop for i exit with i + 1 value */
				* _pst_event = &gpst_pool_event[i-1];
				e_return = NET_NO_ERROR;
			}
			else
			{
				/* error event not found */
				e_return = NET_EVENT_NOT_FOUND;
				#ifdef NET_ERROR
				printf("NET_ERROR [%s:%d] : retrieve event not found\n", __func__, __LINE__);
				#endif
			}
		}
		else
		{
				#ifdef NET_ERROR
				printf("NET_ERROR [%s:%d] : retrieve event parameter error\n", __func__, __LINE__);
				#endif
			e_return = NET_PARAMETER_ERROR;
		}
	}
	else
	{
		#ifdef NET_ERROR
		printf("NET_ERROR [%s:%d] : get retrieve module not initialized\n", __func__, __LINE__);
		#endif
		e_return = NET_MODULE_NOT_INITIALIZED;
	}

	return (e_return);
}

//#define NET_DELEGATE_TRACE_EVENTS

#ifdef NET_DELEGATE_TRACE_EVENTS
#define PASSED_EVENTS_SIZE	256
volatile NET_id_event_t passed_events[PASSED_EVENTS_SIZE];
volatile int passed_events_index = -1;
#endif


void NET_run_f(void)
{
	NET_ADAPT_status_t e_result;
	bool b_repost;

	if (guc_init_passed != NO_INIT)
	{
		b_repost = false;

		/* forever */
		for(;;)
		{
			uint32_t net_queue_timeout = NET_ADAPT_INFITIVE_TIMEOUT;
			NET_ADAPT_LOCK_SYNC_OBJ(gpu32_net_events_sync_obj);
			NET_delegate_events_dispatcher_events_t* repost_event = NET_get_events_list();
			if(repost_event != NULL)
			{
				int64_t currentTime = LLMJVM_IMPL_getCurrentTime__Z(1);
				if(repost_event->net_event_scheduled_time > currentTime)
				{
					net_queue_timeout = (uint32_t)(repost_event->net_event_scheduled_time - currentTime);
					NET_ADAPT_UNLOCK_SYNC_OBJ(gpu32_net_events_sync_obj);
					/* wait a message (indefinite timeout) */
					e_result = NET_ADAPT_wait_event(&gpu32_mailbox, (void**)&gpst_current_event, net_queue_timeout);
				}
				else
				{
					gpst_current_event = repost_event->net_event;
					NET_remove_events_list(repost_event);
					e_result = NET_ADAPT_NO_ERROR;
					NET_ADAPT_UNLOCK_SYNC_OBJ(gpu32_net_events_sync_obj);
				}
			}
			else
			{
					NET_ADAPT_UNLOCK_SYNC_OBJ(gpu32_net_events_sync_obj);
					/* wait a message (indefinite timeout) */
					e_result = NET_ADAPT_wait_event(&gpu32_mailbox, (void**)&gpst_current_event, net_queue_timeout);
			}

			#ifdef NET_DEBUG
				printf("NET_DEBUG [%s:%d] : event %d received !!!\n", __func__, __LINE__, gpst_current_event->e_id_event);
			#endif

			/* test received function */
			if (e_result != NET_ADAPT_NO_ERROR)
			{
				/* log error NET_UNEXPECTED_ERROR_IN_RECEIVED; */
				#ifdef NET_ERROR
					printf("NET_ERROR [%s:%d] : run unexpected error in received\n", __func__, __LINE__);
				#endif
				continue;
			}
			else
			{
				#ifdef NET_DELEGATE_TRACE_EVENTS
					if(++passed_events_index == PASSED_EVENTS_SIZE)
					{
						passed_events_index = 0;
					}
					passed_events[passed_events_index] = gpst_current_event->e_id_event;
				#endif

				/* process event */
				switch (gpst_current_event->e_id_event)
				{
					case (NET_APPLY_CONFIGURATION_EVENT) :
						NET_ADAPT_apply_configuration_f(gpst_current_event->u_ctx_event.st_apply_configuration_ctx_t.pc_interface_name_in,
																						gpst_current_event->u_ctx_event.st_apply_configuration_ctx_t.i_interface_name_length_in,
						                                &gpst_current_event->u_ctx_event.st_apply_configuration_ctx_t.s_error_out);
						break;
					case (NET_START_EVENT) :
						NET_ADAPT_start_f(gpst_current_event->u_ctx_event.st_start_ctx_t.pc_interface_name_in,
															gpst_current_event->u_ctx_event.st_start_ctx_t.i_interface_name_length_in,
						                  &(gpst_current_event->u_ctx_event.st_start_ctx_t.s_error_out));
						break;
					case (NET_STOP_EVENT) :
						NET_ADAPT_stop_f(gpst_current_event->u_ctx_event.st_stop_ctx_t.pc_interface_name_in,
															gpst_current_event->u_ctx_event.st_stop_ctx_t.i_interface_name_length_in,
															&gpst_current_event->u_ctx_event.st_stop_ctx_t.s_error_out);
						break;
					case (NET_WIFI_DISABLE_SOFT_AP_EVENT) :
						NET_ADAPT_wifi_disable_soft_ap_f(&gpst_current_event->u_ctx_event.st_wifi_disable_soft_ap_ctx_t.s_error_out);
						break;
					case (NET_WIFI_ENABLE_SOFT_AP_EVENT) :
						NET_ADAPT_wifi_enable_soft_ap_f(gpst_current_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.pc_ssid_in,
																						gpst_current_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.i_ssid_length_in,
						                                gpst_current_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.pc_passphrase_in,
																						gpst_current_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.i_passphrase_length_in,
														                &gpst_current_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.s_error_out);
						break;
					case (NET_WIFI_SET_NAME_SOFT_AP_EVENT) :
						NET_ADAPT_wifi_set_name_soft_ap_f(gpst_current_event->u_ctx_event.st_wifi_set_name_soft_ap_ctx_t.pc_name_in,
						                                  gpst_current_event->u_ctx_event.st_wifi_set_name_soft_ap_ctx_t.i_name_length_in,
														                  &gpst_current_event->u_ctx_event.st_wifi_set_name_soft_ap_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_BSSID_EVENT) :
						NET_ADAPT_wifi_get_bssid_f(gpst_current_event->u_ctx_event.st_wifi_get_bssid_ctx_t.pc_bssid_in,
																				gpst_current_event->u_ctx_event.st_wifi_get_bssid_ctx_t.i_bssid_length_in,
																				&gpst_current_event->u_ctx_event.st_wifi_get_bssid_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_CHANNEL_EVENT) :
						NET_ADAPT_wifi_get_channel_f(&gpst_current_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_RSSI_EVENT) :
						NET_ADAPT_wifi_get_rssi_f(gpst_current_event->u_ctx_event.st_wifi_get_rssi_ctx_t.pf_rssi_in,
																			gpst_current_event->u_ctx_event.st_wifi_get_rssi_ctx_t.i_rssi_length_in,
														          &gpst_current_event->u_ctx_event.st_wifi_get_rssi_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_SECURITY_MODE_EVENT) :
						NET_ADAPT_wifi_get_security_mode_f(&gpst_current_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_SSID_EVENT) :
						NET_ADAPT_wifi_get_ssid_f(gpst_current_event->u_ctx_event.st_wifi_get_ssid_ctx_t.pc_ssid_in,
																			gpst_current_event->u_ctx_event.st_wifi_get_ssid_ctx_t.i_ssid_length_in,
														          &gpst_current_event->u_ctx_event.st_wifi_get_ssid_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_WPS_MODES_EVENT) :
						NET_ADAPT_wifi_get_wps_modes_f(&gpst_current_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out);
						break;
					case (NET_WIFI_JOIN_EVENT) :
						NET_ADAPT_wifi_join_f(gpst_current_event->u_ctx_event.st_wifi_join_ctx_t.pc_ssid_in,
																	gpst_current_event->u_ctx_event.st_wifi_join_ctx_t.i_ssid_length_in,
						                      gpst_current_event->u_ctx_event.st_wifi_join_ctx_t.pc_passphrase_in,
																	gpst_current_event->u_ctx_event.st_wifi_join_ctx_t.i_passphrase_length_in,
																	gpst_current_event->u_ctx_event.st_wifi_join_ctx_t.i_security_mode_in,
														      &gpst_current_event->u_ctx_event.st_wifi_join_ctx_t.s_error_out);
						break;
					case (NET_WIFI_LEAVE_EVENT) :
						NET_ADAPT_wifi_leave_f(&gpst_current_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_AP_COUNT_EVENT) :
						NET_ADAPT_wifi_get_ap_count_f(&gpst_current_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_AP_SSID_EVENT) :
						NET_ADAPT_wifi_get_ap_ssid_f(gpst_current_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.i_index_in,
																					gpst_current_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.puc_ssid_in,
																					gpst_current_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.i_ssid_length_in,
																					&gpst_current_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_AP_BSSID_EVENT) :
						NET_ADAPT_wifi_get_ap_bssid_f(gpst_current_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.i_index_in,
																					gpst_current_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.puc_bssid_in,
																					gpst_current_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.i_bssid_length_in,
																					&gpst_current_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_AP_CHANNEL_EVENT) :
						NET_ADAPT_wifi_get_ap_channel_f(gpst_current_event->u_ctx_event.st_wifi_get_ap_channel_ctx_t.i_index_in,
																						&gpst_current_event->u_ctx_event.st_wifi_get_ap_channel_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_AP_RSSI_EVENT) :
						NET_ADAPT_wifi_get_ap_rssi_f(gpst_current_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.i_index_in,
																					gpst_current_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.pf_rssi_in,
																					gpst_current_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.i_rssi_length_in,
																					&gpst_current_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_AP_SECURITY_MODE_EVENT) :
						NET_ADAPT_wifi_get_ap_security_mode_f(gpst_current_event->u_ctx_event.st_wifi_get_ap_security_mode_ctx_t.i_index_in,
																									&gpst_current_event->u_ctx_event.st_wifi_get_ap_security_mode_ctx_t.s_error_out);
						break;

					case (NET_WIFI_GET_AP_WPS_MODES_EVENT) :
						NET_ADAPT_wifi_get_ap_wps_modes_f(gpst_current_event->u_ctx_event.st_wifi_get_ap_wps_modes_ctx_t.i_index_in,
																							&gpst_current_event->u_ctx_event.st_wifi_get_ap_wps_modes_ctx_t.s_error_out);
						break;
					case (NET_WIFI_GET_CLIENT_STATE) :
							NET_ADAPT_get_client_state_f(&gpst_current_event->u_ctx_event.st_get_client_state_ctx_t.s_error_out);
  					break;
					default :
						;
						#ifdef NET_ERROR
						printf("NET_ERROR [%s:%d] : run unknown event received\n", __func__, __LINE__);
						#endif
						/* theoretically impossible */

				}

				if (b_repost)
				{
					#ifdef NET_DEBUG
						printf("NET_DEBUG [%s:%d] : event %d must be delayed !!!\n", __func__, __LINE__, gpst_current_event->e_id_event);
					#endif
					b_repost = false;
					// Schedule a repost of the event
					NET_ADAPT_LOCK_SYNC_OBJ(gpu32_net_events_sync_obj);
					if(!NET_is_event_in_events_list(gpst_current_event))
					{
						// Event not found in the events list (not already notified), add it
						NET_add_events_list(gpst_current_event, NET_ADAPT_DELEGATE_EVENT_REPOST_DELAY);
					}
					else
					{
						// Event has been already added by a notify call, post it in the mailbox
						int32_t i = -1;
						NET_delegate_events_dispatcher_events_t* current_net_event;
						uint8_t found = 0;
						for(i = -1; ++i < NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX;)
						{
							current_net_event = &events_list[i];
							if((current_net_event != NULL) && (current_net_event->net_event == gpst_current_event))
							{
								found = 1;
								break;
							}
						}
						if(found)
						{
							NET_remove_events_list(current_net_event);
						}
						NET_ADAPT_post_event_action_f(&gpu32_mailbox, gpst_current_event, 0);
					}
					NET_ADAPT_UNLOCK_SYNC_OBJ(gpu32_net_events_sync_obj);
				}
				else
				{
					#ifdef NET_DEBUG
						printf("NET_DEBUG [%s:%d] : event %d processed !!!\n", __func__, __LINE__, gpst_current_event->e_id_event);
					#endif
					/* resume JVM thread caller */
					if (NET_ADAPT_resume_java_thread(gpst_current_event->ul_id_java_thread) != NET_ADAPT_NO_ERROR)
					{
							#ifdef NET_ERROR
								printf("NET_ERROR [%s:%d] : cannot resume java thread\n", __func__, __LINE__);
							#endif
					}
					else
					{
						gpst_current_event = NULL;

						/* yield cpu if needed */
						NET_ADAPT_YIELD(10);
					}
				}
			}
		}
	}
	else
	{
		#ifdef NET_ERROR
		printf("NET_ERROR [%s:%d] : run module not initialized\n", __func__, __LINE__);
		#endif
	}
}

static NET_status_t NET_is_event_ready_for_operation(NET_event_t* event, unsigned short _us_socket_id, NET_id_operation_t _operation, uint8_t* is_ready)
{
	NET_status_t e_return = NET_NO_ERROR;
	unsigned short current_socket_id = *((unsigned short*)(&(event->u_ctx_event)));

	if(is_ready != NULL)
	{
		*is_ready = 0;

		if(_us_socket_id == current_socket_id)
		{
			if((_operation == NET_ACCEPT_CONNECT) &&
					(event->e_id_event == NET_ACCCEPT_EVENT))
			{
				*is_ready = 1;
			}
			else if((_operation == NET_READ) &&
					((event->e_id_event == NET_READ_EVENT) ||
					(event->e_id_event == NET_READ_BYTE_BUFFER_NATIVE_EVENT) ||
					(event->e_id_event == NET_RECEIVE_EVENT)))
			{
				*is_ready = 1;
			}
			else if((_operation == NET_WRITE) &&
					((event->e_id_event == NET_WRITE_EVENT) ||
					(event->e_id_event == NET_WRITE_BYTE_BUFFER_NATIVE_EVENT) ||
					(event->e_id_event == NET_SEND_EVENT)))
			{
				*is_ready = 1;
			}
		}
	}
	else
	{
		e_return = NET_PARAMETER_ERROR;
	}

	return e_return;
}

NET_status_t NET_notify_ready_for_operation(unsigned short _us_socket_id, NET_id_operation_t _operation)
{
	NET_status_t e_return = NET_EVENT_NOT_FOUND;
	int32_t i = -1;
	NET_delegate_events_dispatcher_events_t* current_net_event = NULL;
	uint8_t found = 0;

	NET_ADAPT_LOCK_SYNC_OBJ(gpu32_net_events_sync_obj);

	for(i = -1; ++i < NET_DELEGATE_MAX_NUMBER_OF_EVENT_IN_MAILBOX;)
	{
		current_net_event = &events_list[i];
		if((current_net_event != NULL) && (current_net_event->net_event != NULL))
		{
			NET_is_event_ready_for_operation(current_net_event->net_event, _us_socket_id, _operation, &found);
			if(found)
			{
				break;
			}
		}
	}

	if(found)
	{
		// NET event need to be notified
		NET_event_t* notified_net_event = current_net_event->net_event;
		NET_remove_events_list(current_net_event);
		if(NET_ADAPT_post_event_action_f(&gpu32_mailbox, notified_net_event, 0) == NET_ADAPT_NO_ERROR)
		{
			e_return = NET_NO_ERROR;
		}
	}
	else
	{
		e_return = NET_NO_ERROR;
	}

	NET_ADAPT_UNLOCK_SYNC_OBJ(gpu32_net_events_sync_obj);

	return (e_return);
}
