/*
 * C
 *
 * Copyright 2017-2018 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */

/**
 * @file
 * @brief Asynchronous network based on nettconn implementation
 * @author MicroEJ Developer Team
 * @version 0.0.1
 * @date 22 May 2018
 */

#include "async_netconn.h"
#include "async_netconn_configuration.h"
#include <string.h>
#include <lwip/api.h>
#include <netinet/in.h>
#include <stdbool.h>
#include <unistd.h>
#include "LLNET_Common.h"
#include "LLNET_ERRORS.h"
#include "osal.h"

#ifdef __cplusplus
	extern "C" {
#endif

/**
 * Sanity check between the expected version of the configuration and the actual version of
 * the configuration.
 * If an error is raised here, it means that a new version of the CCO has been installed and
 * the configuration async_netconn_configuration.h must be updated based on the one provided
 * by the new CCO version.
 */
#if ASYNC_NETCONN_CONFIGURATION_VERSION != 0

	#error "Version of the configuration file async_netconn_configuration.h is not compatible with this implementation."

#endif

/** @brief  An asynchronous netconn request */
typedef struct async_netconn_request {
	int32_t fd;
	int32_t java_thread_id;
	int64_t absolute_timeout_ms;
	select_operation operation;
	struct async_netconn_request *next;
} async_netconn_request;

/*
 * See implementations for descriptions.
 */
static void async_netconn_do_select(void);
static void async_netconn_update_notified_requests(void);
static async_netconn_request *async_netconn_allocate_request(void);
static async_netconn_request *async_netconn_free_used_request(async_netconn_request *request, async_netconn_request *previous_request_in_used_fifo);
static void async_netconn_free_unused_request(async_netconn_request *request);
static int32_t async_netconn_send_new_request(async_netconn_request *request, SNI_callback *callback);
void async_netconn_request_fifo_init(void);

/**
 * @brief Pool of requests. Used to reserve MAX_NB_ASYNC_NETCONN async netconn requests.
 */
static async_netconn_request all_requests[MAX_NB_ASYNC_NETCONN];

/**
 * @brief Linked-list of free requests that can be allocated using async_netconn_allocate_request().
 */
static async_netconn_request *free_requests_fifo;

/**
 * @brief Linked-list of used requests.
 */
static async_netconn_request *used_requests_fifo;

/**
 * @brief Check the netconn for the given file descriptor and operation without blocking.
 *
 * @param[in] fd the file descriptor.
 * @param[in] operation the operation (read or write) we want to monitor with the netconn callbacks.
 *
 * @return netconn check result for the fd and operation
 */
int32_t non_blocking_select(int32_t fd, select_operation operation) {
	if (operation == SELECT_READ) {
		return async_netconn_check_events(fd, 1, 0);
	} else {
		return async_netconn_check_events(fd, 0, 1);
	}
}

/**
 * @brief It uses netconn callbacks to wait for an operation for the given file descriptor.
 *
 * @param[in] fd the file descriptor.
 * @param[in] operation the operation (read or write) we want to monitor with the netconn callbacks.
 * @param[in] timeout_ms timeout in millisecond
 * @param[in] callback the SNI callback to call when the Java thread is resumed or timeout occurs.
 *
 * @return 0 on success, -1 on failure.
 */
int32_t async_netconn(int32_t fd, select_operation operation, int64_t timeout_ms, SNI_callback *callback) {
	async_netconn_request *request = async_netconn_allocate_request();
	if (request == NULL) {
		// No request available :-(
		return -1;
	}

	int32_t java_thread_id = SNI_getCurrentJavaThreadID();
	if (java_thread_id == SNI_ERROR) {
		// Not called from the VM task
		async_netconn_free_unused_request(request);
		return -1;
	}

	LLNET_DEBUG_TRACE("async_netconn: async_netconn on fd=0x%X operation=%s thread 0x%X timeout %lld\n", fd, operation == SELECT_READ ? "read" : "write", java_thread_id, timeout_ms);
	request->java_thread_id = java_thread_id;
	request->fd = fd;
	request->operation = operation;
	if (timeout_ms != 0) {
		request->absolute_timeout_ms = async_netconn_get_current_time_ms() + timeout_ms;
	} else {
		request->absolute_timeout_ms = 0; // infinite timeout
	}

	async_netconn_set_socket_absolute_timeout_in_cache(fd, request->absolute_timeout_ms);

	int32_t res;
	res = async_netconn_send_new_request(request, callback);

	return res;
}

/**
 * @brief Initializes the requests FIFOs.
 * This function must be called prior to any call of async_netconn().
 * It can be called several times.
 */
void async_netconn_request_fifo_init(void) {
	async_netconn_lock();

	// Init free requests FIFO
	free_requests_fifo = &all_requests[0];
	for (int32_t i = 0; i < MAX_NB_ASYNC_NETCONN - 1; i++) {
		all_requests[i].next = &all_requests[i + 1];
	}
	all_requests[MAX_NB_ASYNC_NETCONN - 1].next = NULL;

	// Init used requests FIFO
	used_requests_fifo = NULL;

	async_netconn_unlock();
}

/**
 * @brief The entry point for the async_netconn task.
 * This function must be called from a dedicated task.
 */
void async_netconn_task_main(void) {
	llnet_init();

	async_netconn_request_fifo_init();

	do {
		// Execute a netconn select().
		async_netconn_do_select();

		// Update the received request depending on the netconn results.
		async_netconn_update_notified_requests();
	} while (true);
}

/**
 * @brief This function checks if the async_netconn task is waiting for an event
 * and notifies the task in that case
 */
void async_netconn_try_notify_events(void) {
	if (used_requests_fifo) {
		async_netconn_notify_events();
	}
}

/**
 * Execute a netconn select().
 */
static void async_netconn_do_select(void) {
	async_netconn_request *request;
	// Used to save the lower timeout found in the requests.
	int64_t min_absolute_timeout_ms = INT64_MAX;
	// Used to compute the minimum relative timeout used for the select.
	int64_t min_relative_timeout_ms;

	// --------------------------------------------------
	// Save the minimum timeout found in the request list
	// --------------------------------------------------
	request = used_requests_fifo;
	while (request != NULL) {
		int64_t request_absolute_timeout_ms = request->absolute_timeout_ms;
		if ((request_absolute_timeout_ms != 0) && (request->absolute_timeout_ms < min_absolute_timeout_ms)) {
			// Save the lowest timeout
			min_absolute_timeout_ms = request_absolute_timeout_ms;
		}
		request = request->next;
	}

	// -----------------------------
	//  Compute select timeout value
	// -----------------------------
	if (min_absolute_timeout_ms != INT64_MAX) {
		// At least one request has a timeout.
		min_relative_timeout_ms = min_absolute_timeout_ms - async_netconn_get_current_time_ms();
		// Saturate the relative timeout to a positive value
		if (min_relative_timeout_ms < 0) {
			// 0 means no timeout
			min_relative_timeout_ms = 0;
		}
	} else {
		// No request has timeout
		min_relative_timeout_ms = OSAL_INFINITE_TIME;
	}

	// -------------------------------------------------------------------------------------
	//  Do the select (actually waiting for an event on any socket managed by async_netconn)
	// -------------------------------------------------------------------------------------
	LLNET_DEBUG_TRACE("async_netconn: select (timeout msec=%d)\n", (int32_t)min_relative_timeout_ms);
	async_netconn_wait_events((int32_t)min_relative_timeout_ms);
	LLNET_DEBUG_TRACE("async_netconn: select unblocked\n");
}

/**
 * @brief After the netconn callbacks were triggered, update the status of the requests
 * that have been notified by the netconn or have reached the timeout.
 */
static void async_netconn_update_notified_requests(void) {
	async_netconn_request *request;
	async_netconn_request *previous_request = NULL;
	int64_t current_time_ms = async_netconn_get_current_time_ms();

	// Browse all the requests to find which have been modified
	request = used_requests_fifo;
	while (request != NULL) {
		bool request_timeout_reached;

		// Check if timeout has been reached.
		if ((request->absolute_timeout_ms != 0) && (request->absolute_timeout_ms <= current_time_ms)) {
			request_timeout_reached = true;
		} else {
			request_timeout_reached = false;
		}

		if ((non_blocking_select(request->fd, request->operation)) || request_timeout_reached) {
			// Request done.
			LLNET_DEBUG_TRACE("async_netconn: request done for fd=0x%X operation=%s notify thread 0x%X (%s)\n", request->fd, request->operation == SELECT_READ ? "read" : "write", request->java_thread_id, request_timeout_reached == true ? "timeout" : "no timeout");
			SNI_resumeJavaThread(request->java_thread_id);
			request = async_netconn_free_used_request(request, previous_request);
			// Previous_request is still the same because we have removed request from the used FIFO
		} else {
			previous_request = request;
			request = request->next;
		}
	}
}

/**
 * @brief Remove the given request from the used FIFO and put it in the free FIFO.
 *
 * @param[in/out] request used async_netconn request to free
 * @param[in/out] previous_request_in_used_fifo the previous request in the used fifo
 *
 * @return the next request in the used FIFO.
 */
static async_netconn_request *async_netconn_free_used_request(async_netconn_request *request, async_netconn_request *previous_request_in_used_fifo) {
	async_netconn_request *next_request;

	// No need to protect with mutex, as this function is only called in async_netconn task context, and the MicroJvm task
	// which is accessing these fifos has lower priority
	//async_netconn_lock();

	next_request = request->next;

	// Remove the request from the used FIFO
	if (previous_request_in_used_fifo != NULL) {
		previous_request_in_used_fifo->next = next_request;
	} else {
		// The request was the first in the used list
		used_requests_fifo = next_request;
	}

	// Add the request into the free FIFO
	request->next = free_requests_fifo;
	free_requests_fifo = request;

	// No need to protect with mutex, as this function is only called in async_netconn task context, and the MicroJvm task
	// which is accessing these fifos has lower priority
	//async_netconn_unlock();

	return next_request;
}

/**
 * @brief Put the given request in the free FIFO.
 * The request must not be in the used FIFO.
 *
 * @param[in/out] request unused async_netconn request to free
 *
 */
static void async_netconn_free_unused_request(async_netconn_request *request) {
	async_netconn_lock();

	// Add the request into the free FIFO
	request->next = free_requests_fifo;
	free_requests_fifo = request;

	async_netconn_unlock();
}

/**
 * @brief Notifies the async_netconn task that a new request must be managed.
 * This function will suspend the execution of the current Java thread using
 * SNI_suspendCurrentJavaThreadWithCallback(). Once the netconn callbacks are triggered the Java
 * thread is resumed and the given SNI callback is called.
 *
 * @param[in/out] request async_netconn request to send
 * @param[in] callback the SNI callback to call when the Java thread is resumed or timeout occurs.
 *
 * @return 0 for success, -1 for error
 */
static int32_t async_netconn_send_new_request(async_netconn_request *request, SNI_callback *callback) {
	bool free_request = false;

	async_netconn_lock();

	// Here we are in a critical section, so the async_netconn_sockets_callback() cannot update the socket fd
	// Do here a non_blocking_select() to check the current socket status, as from the last interrogation it may have received updates
	if (!non_blocking_select(request->fd, request->operation)) {
		// Suspend the Java thread
		SNI_suspendCurrentJavaThreadWithCallback(0, callback, NULL);
		// Add the request in the used FIFO
		request->next = used_requests_fifo;
		used_requests_fifo = request;
	} else {
		// Do not add anymore the request in the used FIFO as it can be processed immediately
		LLNET_DEBUG_TRACE("async_netconn: request done for fd=0x%X operation=%s notify thread 0x%X\n", request->fd, request->operation == SELECT_READ ? "read" : "write", request->java_thread_id);
		free_request = true;
	}

	async_netconn_unlock();

	// Either notify the async_netconn task, or free the request if it was not sent
	if (!free_request) {
		async_netconn_notify_events();
	} else {
		async_netconn_free_unused_request(request);
	}

	return 0;
}

/**
 * @brief Find a free request and returns it.
 * The returned request is not put it in the used requests FIFO.
 * It must be either put in the used requests FIFO using async_netconn_send_new_request()
 * or put back in the free requests FIFO on error using async_netconn_free_unused_request().
 *
 * This function is thread safe.
 *
 * @return allocated request or null if none available
 */
static async_netconn_request *async_netconn_allocate_request(void) {
	async_netconn_lock();

	async_netconn_request *new_request = free_requests_fifo;
	if (new_request != NULL) {
		// Remove the request from the free FIFO
		free_requests_fifo = new_request->next;
	} else {
		// else: no request available
	}

	async_netconn_unlock();

	return new_request;
}

#ifdef __cplusplus
	}
#endif
