/*
 * C
 *
 * Copyright 2017-2018 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */

/**
 * @file
 * @brief Asynchronous network based on netconn implementation over OSAL API.
 * @author MicroEJ Developer Team
 * @version 0.0.1
 * @date 22 May 2018
 */

#include "async_netconn.h"
#include "async_netconn_configuration.h"
#include "osal.h"
#include <stddef.h>

#ifdef __cplusplus
	extern "C" {
#endif

/*
 * See implementations for descriptions.
 */
void async_netconn_lock(void);
void async_netconn_unlock(void);
static int32_t async_netconn_start_task(void);

/**
 * @brief Stack of the async_netconn task.
 */
OSAL_task_stack_declare(async_netconn_task_stack, ASYNC_NETCONN_TASK_STACK_SIZE);

/**
 * @brief async_netconn OS task.
 */
static OSAL_task_handle_t async_netconn_task;

/**
 * @brief Mutex used for critical sections.
 */
static OSAL_mutex_handle_t async_netconn_mutex;

/**
 * @brief Initialize the async_netconn component. This function must be called prior to any call of
 * async_netconn().
 *
 * @return 0 on success, -1 on failure.
 */
int32_t async_netconn_init(void) {
	int32_t res;

	res = async_netconn_start_task();
	if (res == 0) {
		async_netconn_request_fifo_init();
	}
	return res;
}

/**
 * @brief Start RTOS task and init RTOS specific structures.
 */
static int32_t async_netconn_start_task(void) {
	OSAL_status_t status;

	status = OSAL_mutex_create((uint8_t *)ASYNC_NETCONN_MUTEX_NAME, &async_netconn_mutex);

	if (status == OSAL_OK) {
		status = OSAL_task_create((OSAL_task_entry_point_t) async_netconn_task_main, (uint8_t *)ASYNC_NETCONN_TASK_NAME, async_netconn_task_stack, ASYNC_NETCONN_TASK_PRIORITY, NULL, &async_netconn_task);
	}

	if (status == OSAL_OK) {
		return 0;
	} else {
		return -1;
	}
}

/**
 * @brief Enter critical section for the async_netconn component.
 */
void async_netconn_lock(void) {
	OSAL_mutex_take(&async_netconn_mutex, OSAL_INFINITE_TIME);
}

/**
 * @brief Exit critical section for the async_netconn component.
 */
void async_netconn_unlock(void) {
	OSAL_mutex_give(&async_netconn_mutex);
}

/**
 * @brief Wait for an async_netconn event.
 *
 * @param[in] timeout timeout in millisecond to wait for an event
 */
void async_netconn_wait_events(int32_t timeout) {
	OSAL_task_notify_take(timeout);
}

/**
 * @brief Notify the caller than an async_netconn event was received.
 */
void async_netconn_notify_events(void) {
	OSAL_task_notify_give(&async_netconn_task);
}

#ifdef __cplusplus
	}
#endif
