/*
 * C
 *
 * Copyright 2017-2018 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */

/**
 * @file
 * @brief Asynchronous network based on nettconn implementation
 * @author MicroEJ Developer Team
 * @version 0.0.1
 * @date 22 May 2018
 */

#include "async_netconn.h"
#include "async_netconn_configuration.h"
#include <string.h>
#include <lwip/api.h>
#include <lwip/tcp.h>
#include <lwip/udp.h>
#include <lwip/igmp.h>
#include <netinet/in.h>
#include <stdbool.h>
#include <unistd.h>
#include "LLNET_Common.h"
#include "LLNET_ERRORS.h"

#ifdef __cplusplus
	extern "C" {
#endif

/**@brief Define ENABLE_ASYNC_NETCONN_DEBUG in async_netconn_configuration.h to enable debug trace */
#ifdef ENABLE_ASYNC_NETCONN_DEBUG
	#define ASYNC_NETCONN_DEBUG        printf
#else
	#define ASYNC_NETCONN_DEBUG(...)   ((void) 0)
#endif

/**@brief Print an IP address in the log */
#if LWIP_IPV4 && LWIP_IPV6
#define ASYNC_NETCONN_IP_DEBUG(ipaddr) \
if (IP_IS_V6_VAL((ipaddr))) { \
	ASYNC_NETCONN_DEBUG("%x:%x:%x:%x:%x:%x:%x:%x", \
			IP6_ADDR_BLOCK1(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK2(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK3(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK4(ip_2_ip6(&(ipaddr))), \
			IP6_ADDR_BLOCK5(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK6(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK7(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK8(ip_2_ip6(&(ipaddr)))); \
} else { \
	ASYNC_NETCONN_DEBUG("%d.%d.%d.%d", \
			ip4_addr1_16(ip_2_ip4(&(ipaddr))), ip4_addr2_16(ip_2_ip4(&(ipaddr))), ip4_addr3_16(ip_2_ip4(&(ipaddr))), ip4_addr4_16(ip_2_ip4(&(ipaddr)))); \
}
#elif LWIP_IPV6 /* LWIP_IPV4 && LWIP_IPV6 */
#define ASYNC_NETCONN_IP_DEBUG(ipaddr) \
		ASYNC_NETCONN_DEBUG("%x:%x:%x:%x:%x:%x:%x:%x", \
				IP6_ADDR_BLOCK1(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK2(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK3(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK4(ip_2_ip6(&(ipaddr))), \
				IP6_ADDR_BLOCK5(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK6(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK7(ip_2_ip6(&(ipaddr))), IP6_ADDR_BLOCK8(ip_2_ip6(&(ipaddr))))
#else /*-> LWIP_IPV4: LWIP_IPV4 && LWIP_IPV6 */
#define ASYNC_NETCONN_IP_DEBUG(ipaddr) \
		ASYNC_NETCONN_DEBUG("%d.%d.%d.%d", \
				ip4_addr1_16(ip_2_ip4(&(ipaddr))), ip4_addr2_16(ip_2_ip4(&(ipaddr))), ip4_addr3_16(ip_2_ip4(&(ipaddr))), ip4_addr4_16(ip_2_ip4(&(ipaddr))))
#endif /* LWIP_IPV6 */

/**@brief Define ENABLE_ASYNC_NETCONN_ASSERT in async_netconn_configuration.h to enable assert */
#ifdef ENABLE_ASYNC_NETCONN_ASSERT
#define ASYNC_NETCONN_ASSERT(s, x) do { if (!(x)) { ASYNC_NETCONN_DEBUG("%s\n", s); while(1); } } while (0)
#else
#define ASYNC_NETCONN_ASSERT(s, x) do { if (!(x)) { ASYNC_NETCONN_DEBUG("%s\n", s); } } while (0)
#endif

/**@brief IP address manipulation macros */
#if LWIP_IPV4
#define IP4ADDR_PORT_TO_SOCKADDR(sin, ipaddr, port) do { \
      (sin)->sin_len = sizeof(struct sockaddr_in); \
      (sin)->sin_family = AF_INET; \
      (sin)->sin_port = lwip_htons((port)); \
      inet_addr_from_ipaddr(&(sin)->sin_addr, ipaddr); \
      memset((sin)->sin_zero, 0, SIN_ZERO_LEN); }while(0)
#define SOCKADDR4_TO_IP4ADDR_PORT(sin, ipaddr, port) do { \
    inet_addr_to_ipaddr(ip_2_ip4(ipaddr), &((sin)->sin_addr)); \
    (port) = lwip_ntohs((sin)->sin_port); }while(0)
#endif /* LWIP_IPV4 */

#if LWIP_IPV6
#define IP6ADDR_PORT_TO_SOCKADDR(sin6, ipaddr, port) do { \
      (sin6)->sin6_len = sizeof(struct sockaddr_in6); \
      (sin6)->sin6_family = AF_INET6; \
      (sin6)->sin6_port = lwip_htons((port)); \
      (sin6)->sin6_flowinfo = 0; \
      inet6_addr_from_ip6addr(&(sin6)->sin6_addr, ipaddr); \
      (sin6)->sin6_scope_id = 0; }while(0)
#define SOCKADDR6_TO_IP6ADDR_PORT(sin6, ipaddr, port) do { \
    inet6_addr_to_ip6addr(ip_2_ip6(ipaddr), &((sin6)->sin6_addr)); \
    (port) = lwip_ntohs((sin6)->sin6_port); }while(0)
#endif /* LWIP_IPV6 */

#if LWIP_IPV4 && LWIP_IPV6
static void sockaddr_to_ipaddr_port(const struct sockaddr* sockaddr, ip_addr_t* ipaddr, uint16_t* port);

#define IS_SOCK_ADDR_LEN_VALID(namelen)  (((namelen) == sizeof(struct sockaddr_in)) || \
                                         ((namelen) == sizeof(struct sockaddr_in6)))
#define IS_SOCK_ADDR_TYPE_VALID(name)    (((name)->sa_family == AF_INET) || \
                                         ((name)->sa_family == AF_INET6))
#define SOCK_ADDR_TYPE_MATCH(name, sock) \
       ((((name)->sa_family == AF_INET) && !(NETCONNTYPE_ISIPV6((sock)->conn->type))) || \
       (((name)->sa_family == AF_INET6) && (NETCONNTYPE_ISIPV6((sock)->conn->type))))
#define IPADDR_PORT_TO_SOCKADDR(sockaddr, ipaddr, port) do { \
    if (IP_IS_V6(ipaddr)) { \
      IP6ADDR_PORT_TO_SOCKADDR((struct sockaddr_in6*)(void*)(sockaddr), ip_2_ip6(ipaddr), port); \
    } else { \
      IP4ADDR_PORT_TO_SOCKADDR((struct sockaddr_in*)(void*)(sockaddr), ip_2_ip4(ipaddr), port); \
    } } while(0)
#define SOCKADDR_TO_IPADDR_PORT(sockaddr, ipaddr, port) sockaddr_to_ipaddr_port(sockaddr, ipaddr, &(port))
#define DOMAIN_TO_NETCONN_TYPE(domain, type) (((domain) == AF_INET) ? \
  (type) : (enum netconn_type)((type) | NETCONN_TYPE_IPV6))
#elif LWIP_IPV6 /* LWIP_IPV4 && LWIP_IPV6 */
#define IS_SOCK_ADDR_LEN_VALID(namelen)  ((namelen) == sizeof(struct sockaddr_in6))
#define IS_SOCK_ADDR_TYPE_VALID(name)    ((name)->sa_family == AF_INET6)
#define SOCK_ADDR_TYPE_MATCH(name, sock) 1
#define IPADDR_PORT_TO_SOCKADDR(sockaddr, ipaddr, port) \
        IP6ADDR_PORT_TO_SOCKADDR((struct sockaddr_in6*)(void*)(sockaddr), ip_2_ip6(ipaddr), port)
#define SOCKADDR_TO_IPADDR_PORT(sockaddr, ipaddr, port) \
        SOCKADDR6_TO_IP6ADDR_PORT((const struct sockaddr_in6*)(const void*)(sockaddr), ipaddr, port)
#define DOMAIN_TO_NETCONN_TYPE(domain, netconn_type) (netconn_type)
#else /*-> LWIP_IPV4: LWIP_IPV4 && LWIP_IPV6 */
#define IS_SOCK_ADDR_LEN_VALID(namelen)  ((namelen) == sizeof(struct sockaddr_in))
#define IS_SOCK_ADDR_TYPE_VALID(name)    ((name)->sa_family == AF_INET)
#define SOCK_ADDR_TYPE_MATCH(name, sock) 1
#define IPADDR_PORT_TO_SOCKADDR(sockaddr, ipaddr, port) \
        IP4ADDR_PORT_TO_SOCKADDR((struct sockaddr_in*)(void*)(sockaddr), ip_2_ip4(ipaddr), port)
#define SOCKADDR_TO_IPADDR_PORT(sockaddr, ipaddr, port) \
        SOCKADDR4_TO_IP4ADDR_PORT((const struct sockaddr_in*)(const void*)(sockaddr), ipaddr, port)
#define DOMAIN_TO_NETCONN_TYPE(domain, netconn_type) (netconn_type)
#endif /* LWIP_IPV6 */

#define IS_SOCK_ADDR_TYPE_VALID_OR_UNSPEC(name)    (((name)->sa_family == AF_UNSPEC) || \
                                                    IS_SOCK_ADDR_TYPE_VALID(name))
#define SOCK_ADDR_TYPE_MATCH_OR_UNSPEC(name, sock) (((name)->sa_family == AF_UNSPEC) || \
                                                    SOCK_ADDR_TYPE_MATCH(name, sock))
#define IS_SOCK_ADDR_ALIGNED(name)      ((((mem_ptr_t)(name)) % 4) == 0)

/**@brief Set socket errno */
#define async_netconn_sock_set_errno(sk, e) do { \
  const int32_t sockerr = (e); \
  sk->err = (uint8_t)sockerr; \
  async_netconn_set_errno(sockerr); \
} while (0)

/**
 * @brief Set global errno for async_netconn APIs
 */
#define async_netconn_set_errno(err) do { if (err) { async_netconn_global_errno = (err); } } while(0)

#if LWIP_IGMP
/* This is to keep track of IP_ADD_MEMBERSHIP calls to drop the membership when a socket is closed */
struct async_netconn_socket_multicast_pair {
  /** the socket */
  struct async_netconn_sock* sock;
  /** the interface address */
  ip4_addr_t if_addr;
  /** the group address */
  ip4_addr_t multi_addr;
};

struct async_netconn_socket_multicast_pair socket_ipv4_multicast_memberships[MEMP_NUM_NETCONN];

static int32_t async_netconn_socket_register_membership(int32_t s, const ip4_addr_t *if_addr, const ip4_addr_t *multi_addr);
static void async_netconn_socket_unregister_membership(int32_t s, const ip4_addr_t *if_addr, const ip4_addr_t *multi_addr);
static void async_netconn_socket_drop_registered_memberships(int32_t s);
#endif /* LWIP_IGMP */

/**
 * @brief Table to quickly map an lwIP error (err_t) to a socket error
 * by using -err as an index
 */
static const int32_t async_netconn_err_to_errno_table[] = {
  0,             /* ERR_OK          0      No error, everything OK. */
  ENOMEM,        /* ERR_MEM        -1      Out of memory error.     */
  ENOBUFS,       /* ERR_BUF        -2      Buffer error.            */
  EWOULDBLOCK,   /* ERR_TIMEOUT    -3      Timeout                  */
  EHOSTUNREACH,  /* ERR_RTE        -4      Routing problem.         */
  EINPROGRESS,   /* ERR_INPROGRESS -5      Operation in progress    */
  EINVAL,        /* ERR_VAL        -6      Illegal value.           */
  EWOULDBLOCK,   /* ERR_WOULDBLOCK -7      Operation would block.   */
  EADDRINUSE,    /* ERR_USE        -8      Address in use.          */

#if ESP_LWIP
  EALREADY,      /* ERR_ALREADY    -9      Already connected.       */
  EISCONN,		 /* ERR_ISCONN     -10	   Conn already established */
  ECONNABORTED,  /* ERR_ABRT       -11     Connection aborted.      */
  ECONNRESET,    /* ERR_RST        -12     Connection reset.        */
  ENOTCONN,      /* ERR_CLSD       -13     Connection closed.       */
  ENOTCONN,      /* ERR_CONN       -14     Not connected.           */
  EIO,           /* ERR_ARG        -15     Illegal argument.        */
  -1,            /* ERR_IF         -16     Low-level netif error    */
#else
  EALREADY,      /* ERR_ALREADY    -9      Already connecting.      */
  EISCONN,       /* ERR_ISCONN     -10     Conn already established.*/
  ENOTCONN,      /* ERR_CONN       -11     Not connected.           */
  -1,            /* ERR_IF         -12     Low-level netif error    */
  ECONNABORTED,  /* ERR_ABRT       -13     Connection aborted.      */
  ECONNRESET,    /* ERR_RST        -14     Connection reset.        */
  ENOTCONN,      /* ERR_CLSD       -15     Connection closed.       */
  EIO            /* ERR_ARG        -16     Illegal argument.        */
#endif
};

/**
 * @brief LwIP should export the err_to_errno mapping, until that we do it here
 */
#define async_netconn_err_to_errno(err) \
		((err > 0) || (-err >= (err_t)(sizeof(async_netconn_err_to_errno_table) / sizeof(async_netconn_err_to_errno_table[0])))) ? EIO : async_netconn_err_to_errno_table[-err]

/** A struct sockaddr replacement that has the same alignment as sockaddr_in/
 *  sockaddr_in6 if instantiated.
 */
union sockaddr_aligned {
   struct sockaddr sa;
#if LWIP_IPV6
   struct sockaddr_in6 sin6;
#endif /* LWIP_IPV6 */
#if LWIP_IPV4
   struct sockaddr_in sin;
#endif /* LWIP_IPV4 */
};

/**
 * @brief Callback registered in the netconn layer for each socket-netconn.
 *        This callback is called in tcpip_thread context
 *
 * @param[in] conn netconn for which this context is called
 * @param[in] evt callback event
 * @param[in] len unused argument
 */
static void async_netconn_event_callback(struct netconn *conn, enum netconn_evt evt, uint16_t len);

/**
 * @brief Allocate a new socket for a given netconn.
 *
 * @param[in] newconn the netconn for which to allocate a socket
 * @param[in] accepted 1 if socket has been created by async_netconn_accept(),
 *                     0 if socket has been created by async_netconn_socket()
 * @return the index of the new socket; -1 on error
 */
static int32_t async_netconn_alloc_socket(struct netconn *newconn, int32_t accepted);

/**
 * @brief Map a externally used socket index to the internal socket representation.
 *
 * @param s[in] externally used socket index
 * @return struct async_netconn_sock for the socket or NULL if not found
 */
static struct async_netconn_sock * async_netconn_get_socket(int32_t s);

/**
 * @brief Free a socket. The socket's netconn must have been delete before!
 *
 * @param sock[in] the socket to free
 * @param is_tcp[in] != 0 for TCP sockets, used to free lastdata
 */
static void async_netconn_free_socket(struct async_netconn_sock *sock, int32_t is_tcp);

/**
 * @brief Get addr name
 *
 * @param[in] s the socket file descriptor
 * @param[out] name a pointer to a sockaddr structure
 * @param[inout] namelen should be initialized to indicate the amount of space (in bytes) pointed to
 *               by name. On return it contains the actual size of the addr
 * @param[in] local if local or peer addr is requested
 *
 * @return On success, zero is returned.  On error, -1 is returned
 */
static int32_t async_netconn_getaddrname(int32_t s, struct sockaddr *name, socklen_t *namelen, uint8_t local);

/**
 * @brief Available sockets. Used to reserve MEMP_NUM_NETCONN structures.
 */
static async_netconn_sock async_netconn_sockets[MEMP_NUM_NETCONN];

/**
 * @brief Cache of socket timeouts. Used to reserve MEMP_NUM_NETCONN structures.
 */
static async_netconn_sock_timeout async_netconn_sockets_timeout_cache[MEMP_NUM_NETCONN];

/**
 * @brief Global errno for the async_netconn APIs
 */
static int32_t async_netconn_global_errno;

int32_t async_netconn_accept(int32_t s, struct sockaddr *addr, socklen_t *addrlen) {
	struct async_netconn_sock *sock, *nsock;
	struct netconn *newconn;
	ip_addr_t naddr;
	uint16_t port = 0;
	int32_t newsock;
	err_t err;

	ASYNC_NETCONN_DEBUG("async_netconn_accept(%d)...\n", s);
	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (netconn_is_nonblocking(sock->conn) && (sock->rcvevent <= 0)) {
		ASYNC_NETCONN_DEBUG("async_netconn_accept(%d): returning EWOULDBLOCK\n", s);
		async_netconn_sock_set_errno(sock, EWOULDBLOCK);
		return -1;
	}

	/* wait for a new connection */
	err = netconn_accept(sock->conn, &newconn);
	if (err != ERR_OK) {
		ASYNC_NETCONN_DEBUG("async_netconn_accept(%d): netconn_acept failed, err=%d\n", s, err);
		if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) != NETCONN_TCP) {
			async_netconn_sock_set_errno(sock, EOPNOTSUPP);
		} else if (err == ERR_CLSD) {
			async_netconn_sock_set_errno(sock, EINVAL);
		} else {
			async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
		}
		return -1;
	}
	ASYNC_NETCONN_ASSERT("newconn != NULL", newconn != NULL);

	newsock = async_netconn_alloc_socket(newconn, 1);
	if (newsock == -1) {
		netconn_delete(newconn);
		async_netconn_sock_set_errno(sock, ENFILE);
		return -1;
	}

	ASYNC_NETCONN_ASSERT("invalid socket index", (newsock >= 0) && (newsock < (sizeof(async_netconn_sockets) / sizeof(async_netconn_sockets[0]))));
	ASYNC_NETCONN_ASSERT("newconn->callback == async_netconn_event_callback", newconn->callback == async_netconn_event_callback);
	nsock = &async_netconn_sockets[newsock];

	/*
	 * See async_netconn_event_callback: If data comes in right away after an accept, even
	 * though the server task might not have created a new socket yet.
	 * In that case, newconn->socket is counted down (newconn->socket--),
	 * so nsock->rcvevent is >= 1 here!
	 */
	async_netconn_lock();
	nsock->rcvevent += (int16_t)(-1 - newconn->socket);
	newconn->socket = newsock;
	async_netconn_unlock();

	/*
	 * Note that POSIX only requires us to check addr is non-NULL. addrlen must
	 * not be NULL if addr is valid.
	 */
	if (addr != NULL) {
		union sockaddr_aligned tempaddr;

		/* get the IP address and port of the remote host */
		err = netconn_peer(newconn, &naddr, &port);
		if (err != ERR_OK) {
			ASYNC_NETCONN_DEBUG("async_netconn_accept(%d): netconn_peer failed, err=%d\n", s, err);
			netconn_delete(newconn);
			async_netconn_free_socket(nsock, 1);
			async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
			return -1;
		}
		ASYNC_NETCONN_ASSERT("addr valid but addrlen NULL", addrlen != NULL);

		IPADDR_PORT_TO_SOCKADDR(&tempaddr, &naddr, port);
		if (*addrlen > tempaddr.sa.sa_len) {
			*addrlen = tempaddr.sa.sa_len;
		}
		memcpy(addr, &tempaddr, *addrlen);

		ASYNC_NETCONN_DEBUG("async_netconn_accept(%d) returning new sock=%d addr=", s, newsock);
		ASYNC_NETCONN_IP_DEBUG(naddr);
		ASYNC_NETCONN_DEBUG(" port=%d\n", port);
	} else {
		ASYNC_NETCONN_DEBUG("async_netconn_accept(%d) returning new sock=%d", s, newsock);
	}

	async_netconn_sock_set_errno(sock, 0);
	return newsock;
}

int32_t async_netconn_bind(int32_t s, const struct sockaddr *name, socklen_t namelen) {
	struct async_netconn_sock *sock;
	ip_addr_t local_addr;
	uint16_t local_port;
	err_t err;

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (!SOCK_ADDR_TYPE_MATCH(name, sock)) {
		/* sockaddr does not match socket type (IPv4/IPv6) */
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(ERR_VAL));
		return -1;
	}

	/* check size, family and alignment of 'name' */
	if (!(IS_SOCK_ADDR_LEN_VALID(namelen) && IS_SOCK_ADDR_TYPE_VALID(name) && IS_SOCK_ADDR_ALIGNED(name))) {
		ASYNC_NETCONN_ASSERT("async_netconn_bind: invalid address", 0);
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(ERR_ARG));
		return -1;
	}

	SOCKADDR_TO_IPADDR_PORT(name, &local_addr, local_port);

	ASYNC_NETCONN_DEBUG("async_netconn_bind(%d, addr=", s);
	ASYNC_NETCONN_IP_DEBUG(local_addr);
	ASYNC_NETCONN_DEBUG(" port=%d)\n", local_port);

	err = netconn_bind(sock->conn, &local_addr, local_port);

	if (err != ERR_OK) {
		ASYNC_NETCONN_DEBUG("async_netconn_bind(%d) failed, err=%d\n", s, err);
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
		return -1;
	}

	ASYNC_NETCONN_DEBUG("async_netconn_bind(%d) succeeded\n", s);
	async_netconn_sock_set_errno(sock, 0);
	return 0;
}

int32_t async_netconn_close(int32_t s) {
	struct async_netconn_sock *sock;
	int32_t is_tcp = 0;
	err_t err;

	ASYNC_NETCONN_DEBUG("async_netconn_close(%d)\n", s);

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (sock->conn != NULL) {
		is_tcp = NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP;
	} else {
		ASYNC_NETCONN_ASSERT("sock->lastdata == NULL", sock->lastdata == NULL);
	}

#if LWIP_IGMP
	/* drop all possibly joined IGMP memberships */
	async_netconn_socket_drop_registered_memberships(s);
#endif /* LWIP_IGMP */

	err = netconn_delete(sock->conn);
	if (err != ERR_OK) {
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
		return -1;
	}

	async_netconn_free_socket(sock, is_tcp);
	async_netconn_set_errno(0);
	return 0;
}

int32_t async_netconn_connect(int32_t s, const struct sockaddr *name, socklen_t namelen) {
	struct async_netconn_sock *sock;
	err_t err;

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (!SOCK_ADDR_TYPE_MATCH_OR_UNSPEC(name, sock)) {
		/* sockaddr does not match socket type (IPv4/IPv6) */
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(ERR_VAL));
		return -1;
	}

	if (name->sa_family == AF_UNSPEC) {
		ASYNC_NETCONN_DEBUG("async_netconn_connect(%d, AF_UNSPEC)\n", s);
		err = netconn_disconnect(sock->conn);
	} else {
		ip_addr_t remote_addr;
		uint16_t remote_port;

		/* check size, family and alignment of 'name' */
		if (!(IS_SOCK_ADDR_LEN_VALID(namelen) && IS_SOCK_ADDR_TYPE_VALID_OR_UNSPEC(name) && IS_SOCK_ADDR_ALIGNED(name))) {
			ASYNC_NETCONN_ASSERT("async_netconn_connect: invalid address", 0);
			async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(ERR_ARG));
			return -1;
		}

		SOCKADDR_TO_IPADDR_PORT(name, &remote_addr, remote_port);

		ASYNC_NETCONN_DEBUG("async_netconn_connect(%d, addr=", s);
		ASYNC_NETCONN_IP_DEBUG(remote_addr);
		ASYNC_NETCONN_DEBUG(" port=%d)\n", remote_port);

		err = netconn_connect(sock->conn, &remote_addr, remote_port);
	}

	if (err != ERR_OK) {
		ASYNC_NETCONN_DEBUG("async_netconn_connect(%d) failed, err=%d\n", s, err);
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
		return -1;
	}

	ASYNC_NETCONN_DEBUG("async_netconn_connect(%d) succeeded\n", s);
	async_netconn_sock_set_errno(sock, 0);
	return 0;
}

int32_t async_netconn_listen(int32_t s, int32_t backlog) {
	struct async_netconn_sock *sock;
	err_t err;

	ASYNC_NETCONN_DEBUG("async_netconn_listen(%d, backlog=%d)\n", s, backlog);

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	/* limit the "backlog" parameter to fit in an uint8_t */
	if (backlog < 0) {
		backlog = 0;
	} else if (backlog > 0xff) {
		backlog = 0xff;
	}

	err = netconn_listen_with_backlog(sock->conn, (uint8_t)backlog);

	if (err != ERR_OK) {
		ASYNC_NETCONN_DEBUG("async_netconn_listen(%d) failed, err=%d\n", s, err);
		if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) != NETCONN_TCP) {
			async_netconn_sock_set_errno(sock, EOPNOTSUPP);
			return -1;
		}
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
		return -1;
	}

	async_netconn_sock_set_errno(sock, 0);
	return 0;
}

int32_t async_netconn_recvfrom(int32_t s, void *mem, size_t len, int32_t flags, struct sockaddr *from, socklen_t *fromlen) {
	struct async_netconn_sock *sock;
	void *buf = NULL;
	struct pbuf *p;
	uint16_t buflen, copylen;
	int32_t off = 0;
	uint8_t done = 0;
	err_t err;

	ASYNC_NETCONN_DEBUG("async_netconn_recvfrom(%d, %p, %d, 0x%x, ..)\n", s, mem, len, flags);

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	do {
		ASYNC_NETCONN_DEBUG("async_netconn_recvfrom: top while sock->lastdata=%p\n", sock->lastdata);
		/* Check if there is data left from the last recv operation. */
		if (sock->lastdata) {
			buf = sock->lastdata;
		} else {
			/* If this is non-blocking call, then check first */
			if (((flags & MSG_DONTWAIT) || netconn_is_nonblocking(sock->conn)) &&
					(sock->rcvevent <= 0)) {
				if (off > 0) {
					/* already received data, return that */
					async_netconn_sock_set_errno(sock, 0);
					return off;
				}
				ASYNC_NETCONN_DEBUG("async_netconn_recvfrom(%d): returning EWOULDBLOCK\n", s);
				async_netconn_set_errno(EWOULDBLOCK);
				return -1;
			}

			/* No data was left from the previous operation, so we try to get some from the network. */
			if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP) {
				err = netconn_recv_tcp_pbuf(sock->conn, (struct pbuf **)&buf);
			} else {
				err = netconn_recv(sock->conn, (struct netbuf **)&buf);
			}
			ASYNC_NETCONN_DEBUG("async_netconn_recvfrom: netconn_recv err=%d, netbuf=%p\n", err, buf);

			if (err != ERR_OK) {
				if (off > 0) {
					if (err == ERR_CLSD) {
						/* closed but already received data, ensure select gets the FIN, too */
						async_netconn_event_callback(sock->conn, NETCONN_EVT_RCVPLUS, 0);
					}
					/* already received data, return that */
					async_netconn_sock_set_errno(sock, 0);
					return off;
				}
				/* We should really do some error checking here. */
				ASYNC_NETCONN_DEBUG("async_netconn_recvfrom(%d): buf == NULL, error is %d!\n", s, err);
				async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
				if (err == ERR_CLSD) {
					return 0;
				} else {
					return -1;
				}
			}
			ASYNC_NETCONN_ASSERT("buf != NULL", buf != NULL);
			sock->lastdata = buf;
		}

		if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP) {
			p = (struct pbuf *)buf;
		} else {
			p = ((struct netbuf *)buf)->p;
		}
		buflen = p->tot_len;
		ASYNC_NETCONN_DEBUG("async_netconn_recvfrom: buflen=%d len=%d off=%d sock->lastoffset=%d\n", buflen, len, off, sock->lastoffset);

		buflen -= sock->lastoffset;

		if (len > buflen) {
			copylen = buflen;
		} else {
			copylen = (uint16_t)len;
		}

		/* copy the contents of the received buffer into the supplied memory pointer mem */
		pbuf_copy_partial(p, (uint8_t *)mem + off, copylen, sock->lastoffset);

		off += copylen;

		if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP) {
			ASYNC_NETCONN_ASSERT("invalid copylen, len would underflow", len >= copylen);
			len -= copylen;
			if ((len <= 0) ||
					(p->flags & PBUF_FLAG_PUSH) ||
					(sock->rcvevent <= 0) ||
					((flags & MSG_PEEK) != 0)) {
				done = 1;
			}
		} else {
			done = 1;
		}

		/* Check to see from where the data was.*/
		if (done) {
			if (from && fromlen) {
				uint16_t port;
				ip_addr_t tmpaddr;
				ip_addr_t *fromaddr;
				union sockaddr_aligned saddr;

				if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP) {
					fromaddr = &tmpaddr;
					netconn_getaddr(sock->conn, fromaddr, &port, 0);
				} else {
					port = netbuf_fromport((struct netbuf *)buf);
					fromaddr = netbuf_fromaddr((struct netbuf *)buf);
				}

				IPADDR_PORT_TO_SOCKADDR(&saddr, fromaddr, port);

				ASYNC_NETCONN_DEBUG("async_netconn_recvfrom(%d): addr=", s);
				ASYNC_NETCONN_IP_DEBUG(*fromaddr);
				ASYNC_NETCONN_DEBUG(" port=%d len=%d\n", port, off);

				if (*fromlen > saddr.sa.sa_len) {
					*fromlen = saddr.sa.sa_len;
				}
				memcpy(from, &saddr, *fromlen);
			}
		}

		/* If we don't peek the incoming message... */
		if ((flags & MSG_PEEK) == 0) {
			/*
			 * If this is a TCP socket, check if there is data left in the
			 * buffer. If so, it should be saved in the sock structure for next
			 * time around.
			 */
			if ((NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP) && (buflen - copylen > 0)) {
				sock->lastdata = buf;
				sock->lastoffset += copylen;
				ASYNC_NETCONN_DEBUG("async_netconn_recvfrom: lastdata now netbuf=%p\n", buf);
			} else {
				sock->lastdata = NULL;
				sock->lastoffset = 0;
				ASYNC_NETCONN_DEBUG("async_netconn_recvfrom: deleting netbuf=%p\n", buf);
				if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP) {
					pbuf_free((struct pbuf *)buf);
				} else {
					netbuf_delete((struct netbuf *)buf);
				}
				buf = NULL;
			}
		}
	} while (!done);

	async_netconn_sock_set_errno(sock, 0);
	return off;
}

int32_t async_netconn_recv(int32_t s, void *mem, size_t len, int32_t flags) {
	return async_netconn_recvfrom(s, mem, len, flags, NULL, NULL);
}

int32_t async_netconn_send(int32_t s, const void *data, size_t size, int32_t flags) {
	struct async_netconn_sock *sock;
	err_t err;
	uint8_t write_flags;
	size_t written;

	ASYNC_NETCONN_DEBUG("async_netconn_send(%d, data=%p, size=%d, flags=0x%x)\n", s, data, size, flags);

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) != NETCONN_TCP) {
#if (LWIP_UDP || LWIP_RAW)
		return async_netconn_sendto(s, data, size, flags, NULL, 0);
#else /* (LWIP_UDP || LWIP_RAW) */
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(ERR_ARG));
		return -1;
#endif /* (LWIP_UDP || LWIP_RAW) */
	}

	write_flags = NETCONN_COPY |
	((flags & MSG_MORE)     ? NETCONN_MORE      : 0) |
	((flags & MSG_DONTWAIT) ? NETCONN_DONTBLOCK : 0);
	written = 0;
	err = netconn_write_partly(sock->conn, data, size, write_flags, &written);

	ASYNC_NETCONN_DEBUG("async_netconn_send(%d) err=%d written=%"SZT_F"\n", s, err, written);
	async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
	return (err == ERR_OK ? (int32_t)written : -1);
}

int32_t async_netconn_sendto(int32_t s, const void *data, size_t size, int32_t flags, const struct sockaddr *to, socklen_t tolen) {
	struct async_netconn_sock *sock;
	err_t err;
	uint16_t short_size;
	uint16_t remote_port;
	struct netbuf buf;

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) == NETCONN_TCP) {
#if LWIP_TCP
		return async_netconn_send(s, data, size, flags);
#else /* LWIP_TCP */
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(ERR_ARG));
		return -1;
#endif /* LWIP_TCP */
	}

	/* @todo: split into multiple sendto's? */
	ASYNC_NETCONN_ASSERT("async_netconn_sendto: size must fit in uint16_t", size <= 0xffff);
	short_size = (uint16_t)size;
	if (!(((to == NULL) && (tolen == 0)) || (IS_SOCK_ADDR_LEN_VALID(tolen) && IS_SOCK_ADDR_TYPE_VALID(to) && IS_SOCK_ADDR_ALIGNED(to)))) {
		ASYNC_NETCONN_ASSERT("async_netconn_sendto: invalid address", 0);
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(ERR_ARG));
		return -1;
	}

	/* initialize a buffer */
	buf.p = buf.ptr = NULL;
#if LWIP_CHECKSUM_ON_COPY
	buf.flags = 0;
#endif /* LWIP_CHECKSUM_ON_COPY */
	if (to) {
		SOCKADDR_TO_IPADDR_PORT(to, &buf.addr, remote_port);
	} else {
		remote_port = 0;
		ip_addr_set_any(NETCONNTYPE_ISIPV6(netconn_type(sock->conn)), &buf.addr);
	}
	netbuf_fromport(&buf) = remote_port;

	ASYNC_NETCONN_DEBUG("async_netconn_sendto(%d, data=%p, short_size=%d, flags=0x%x to=", s, data, short_size, flags);
	ASYNC_NETCONN_IP_DEBUG(buf.addr);
	ASYNC_NETCONN_DEBUG(" port=%d)\n", remote_port);

	/* make the buffer point to the data that should be sent */
#if LWIP_NETIF_TX_SINGLE_PBUF
	/* Allocate a new netbuf and copy the data into it. */
	if (netbuf_alloc(&buf, short_size) == NULL) {
		err = ERR_MEM;
	} else {
#if LWIP_CHECKSUM_ON_COPY
		if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) != NETCONN_RAW) {
			uint16_t chksum = LWIP_CHKSUM_COPY(buf.p->payload, data, short_size);
			netbuf_set_chksum(&buf, chksum);
		} else
#endif /* LWIP_CHECKSUM_ON_COPY */
		{
			memcpy(buf.p->payload, data, short_size);
		}
		err = ERR_OK;
	}
#else /* LWIP_NETIF_TX_SINGLE_PBUF */
	err = netbuf_ref(&buf, data, short_size);
#endif /* LWIP_NETIF_TX_SINGLE_PBUF */

	if (err == ERR_OK) {
		/* send the data */
		err = netconn_send(sock->conn, &buf);
	}

	/* deallocated the buffer */
	netbuf_free(&buf);

	async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
	return (err == ERR_OK ? short_size : -1);
}

int32_t async_netconn_socket(int32_t domain, int32_t type, int32_t protocol) {
	struct netconn *conn;
	int32_t i;

	/* create a netconn */
	switch (type) {
	case SOCK_RAW:
		conn = netconn_new_with_proto_and_callback(DOMAIN_TO_NETCONN_TYPE(domain, NETCONN_RAW), (uint8_t)protocol, async_netconn_event_callback);
		ASYNC_NETCONN_DEBUG("async_netconn_socket(%s, SOCK_RAW, %d) = ", domain == PF_INET ? "PF_INET" : "UNKNOWN", protocol);
		break;
	case SOCK_DGRAM:
		conn = netconn_new_with_callback(DOMAIN_TO_NETCONN_TYPE(domain, ((protocol == IPPROTO_UDPLITE) ? NETCONN_UDPLITE : NETCONN_UDP)), async_netconn_event_callback);
		ASYNC_NETCONN_DEBUG("async_netconn_socket(%s, SOCK_DGRAM, %d) = ", domain == PF_INET ? "PF_INET" : "UNKNOWN", protocol);
		break;
	case SOCK_STREAM:
		conn = netconn_new_with_callback(DOMAIN_TO_NETCONN_TYPE(domain, NETCONN_TCP), async_netconn_event_callback);
		ASYNC_NETCONN_DEBUG("async_netconn_socket(%s, SOCK_STREAM, %d) = ", domain == PF_INET ? "PF_INET" : "UNKNOWN", protocol);
		break;
	default:
		async_netconn_set_errno(EINVAL);
		ASYNC_NETCONN_DEBUG("async_netconn_socket(%d, %d/UNKNOWN, %d) = -1\n", domain, type, protocol);
		return -1;
	}

	if (!conn) {
		ASYNC_NETCONN_DEBUG("-1 / ENOBUFS (could not create netconn)\n");
		async_netconn_set_errno(ENOBUFS);
		return -1;
	}

	i = async_netconn_alloc_socket(conn, 0);

	if (i == -1) {
		netconn_delete(conn);
		async_netconn_set_errno(ENFILE);
		return -1;
	}

	conn->socket = i;
	ASYNC_NETCONN_DEBUG("%d\n", i);
	async_netconn_set_errno(0);
	return i;
}

int32_t async_netconn_shutdown(int32_t s, int32_t how) {
	struct async_netconn_sock *sock;
	err_t err;
	uint8_t shut_rx = 0, shut_tx = 0;

	ASYNC_NETCONN_DEBUG("async_netconn_shutdown(%d, how=%d)\n", s, how);

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (sock->conn != NULL) {
		if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) != NETCONN_TCP) {
			async_netconn_sock_set_errno(sock, EOPNOTSUPP);
			return -1;
		}
	} else {
		async_netconn_sock_set_errno(sock, ENOTCONN);
		return -1;
	}

	if (how == SHUT_RD) {
		shut_rx = 1;
	} else if (how == SHUT_WR) {
		shut_tx = 1;
	} else if (how == SHUT_RDWR) {
		shut_rx = 1;
		shut_tx = 1;
	} else {
		async_netconn_sock_set_errno(sock, EINVAL);
		return -1;
	}

	err = netconn_shutdown(sock->conn, shut_rx, shut_tx);

	async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
	return (err == ERR_OK ? 0 : -1);
}

int32_t async_netconn_getpeername(int32_t s, struct sockaddr *name, socklen_t *namelen) {
	return async_netconn_getaddrname(s, name, namelen, 0);
}

int32_t async_netconn_getsockname(int32_t s, struct sockaddr *name, socklen_t *namelen) {
	return async_netconn_getaddrname(s, name, namelen, 1);
}

int32_t async_netconn_getsockopt(int32_t s, int32_t level, int32_t optname, void *optval, socklen_t *optlen) {
	struct async_netconn_sock *sock;

	ASYNC_NETCONN_DEBUG("async_netconn_getsockopt(%d, level=%d optname=%d)\n", s, level, optname);

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if ((NULL == optval) || (NULL == optlen)) {
		async_netconn_sock_set_errno(sock, EFAULT);
		return -1;
	}

	if (sock->conn == NULL) {
		async_netconn_sock_set_errno(sock, EINVAL);
		return -1;
	}

	switch (level) {
	case SOL_SOCKET:
		switch (optname) {
		case SO_BROADCAST:
		case SO_KEEPALIVE:
#if SO_REUSE
		case SO_REUSEADDR:
#endif /* SO_REUSE */
			if (*optlen < sizeof(int32_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.ip == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			*(int32_t *)optval = ip_get_option(sock->conn->pcb.ip, optname);
			break;
#if LWIP_SO_RCVTIMEO
		case SO_RCVTIMEO:
			if (*optlen < sizeof(struct timeval)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			((struct timeval *)(optval))->tv_sec = netconn_get_recvtimeout(sock->conn) / 1000U;
			((struct timeval *)(optval))->tv_usec = (netconn_get_recvtimeout(sock->conn) % 1000U) * 1000U;
			break;
#endif /* LWIP_SO_RCVTIMEO */
#if LWIP_SO_RCVBUF
		case SO_RCVBUF:
			if (*optlen < sizeof(int32_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			*(int32_t *)optval = netconn_get_recvbufsize(sock->conn);
			break;
#endif /* LWIP_SO_RCVBUF */
		}
		break;

	case IPPROTO_IP:
		switch (optname) {
		case IP_TOS:
			if (*optlen < sizeof(int32_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.ip == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			*(int32_t *)optval = sock->conn->pcb.ip->tos;
			break;
#if LWIP_MULTICAST_TX_OPTIONS
		case IP_MULTICAST_TTL:
			if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_UDP) {
				async_netconn_sock_set_errno(sock, ENOPROTOOPT);
				return -1;
			}

			if (*optlen < sizeof(uint8_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.udp == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			*(uint8_t *)optval = udp_get_multicast_ttl(sock->conn->pcb.udp);
			break;
		case IP_MULTICAST_IF:
		{
			if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_UDP) {
				async_netconn_sock_set_errno(sock, ENOPROTOOPT);
				return -1;
			}

			if (*optlen < sizeof(struct in_addr)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.udp == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			inet_addr_from_ipaddr((struct in_addr *)optval, udp_get_multicast_netif_addr(sock->conn->pcb.udp));
		}
			break;
		case IP_MULTICAST_LOOP:
			if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_UDP) {
				async_netconn_sock_set_errno(sock, ENOPROTOOPT);
				return -1;
			}

			if (*optlen < sizeof(uint8_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.udp == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if ((sock->conn->pcb.udp->flags & UDP_FLAGS_MULTICAST_LOOP) != 0) {
				*(uint8_t *)optval = 1;
			} else {
				*(uint8_t *)optval = 0;
			}
			break;
#endif /* LWIP_MULTICAST_TX_OPTIONS */
		default:
			async_netconn_sock_set_errno(sock, ENOPROTOOPT);
			return -1;
		}
		break;

	case IPPROTO_TCP:
		if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_TCP) {
			async_netconn_sock_set_errno(sock, ENOPROTOOPT);
			return -1;
		}

		if (*optlen < sizeof(int32_t)) {
			async_netconn_sock_set_errno(sock, EINVAL);
			return -1;
		}

		if (sock->conn->pcb.tcp == NULL) {
			async_netconn_sock_set_errno(sock, EINVAL);
			return -1;
		}

		if (sock->conn->pcb.tcp->state == LISTEN) {
			async_netconn_sock_set_errno(sock, EINVAL);
			return -1;
		}

		switch (optname) {
		case TCP_NODELAY:
			*(int32_t *)optval = tcp_nagle_disabled(sock->conn->pcb.tcp);
			break;
		default:
			async_netconn_sock_set_errno(sock, ENOPROTOOPT);
			return -1;
		}
		break;

	default:
		async_netconn_sock_set_errno(sock, ENOPROTOOPT);
		return -1;
	}

	return 0;
}

int32_t async_netconn_setsockopt(int32_t s, int32_t level, int32_t optname, const void *optval, socklen_t optlen) {
	struct async_netconn_sock *sock;

	ASYNC_NETCONN_DEBUG("async_netconn_setsockopt(%d, level=%d optname=%d)\n", s, level, optname);

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (NULL == optval) {
		async_netconn_sock_set_errno(sock, EFAULT);
		return -1;
	}

	if (sock->conn == NULL) {
		async_netconn_sock_set_errno(sock, EINVAL);
		return -1;
	}

	switch (level) {
	case SOL_SOCKET:
		switch (optname) {
		case SO_BROADCAST:
		case SO_KEEPALIVE:
#if SO_REUSE
		case SO_REUSEADDR:
#endif /* SO_REUSE */
			if (optlen < sizeof(int32_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.ip == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (*(const int32_t *)optval) {
				ip_set_option(sock->conn->pcb.ip, optname);
			} else {
				ip_reset_option(sock->conn->pcb.ip, optname);
			}
			break;
#if LWIP_SO_RCVTIMEO
		case SO_RCVTIMEO:
			if (optlen < sizeof(struct timeval)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			netconn_set_recvtimeout(sock->conn, (int32_t)(((const struct timeval *)(optval))->tv_sec * 1000U) + (((const struct timeval *)(optval))->tv_usec / 1000U));
			break;
#endif /* LWIP_SO_RCVTIMEO */
#if LWIP_SO_RCVBUF
		case SO_RCVBUF:
			if (optlen < sizeof(int32_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			netconn_set_recvbufsize(sock->conn, *(const int32_t *)optval);
			break;
#endif /* LWIP_SO_RCVBUF */
		}
		break;

	case IPPROTO_IP:
		switch (optname) {
		case IP_TOS:
			if (optlen < sizeof(int32_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.ip == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			sock->conn->pcb.ip->tos = (uint8_t)(*(const int32_t *)optval);
			break;
#if LWIP_MULTICAST_TX_OPTIONS
		case IP_MULTICAST_TTL:
			if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_UDP) {
				async_netconn_sock_set_errno(sock, ENOPROTOOPT);
				return -1;
			}

			if (optlen < sizeof(uint8_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.udp == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			udp_set_multicast_ttl(sock->conn->pcb.udp, (uint8_t)(*(const uint8_t *)optval));
			break;
		case IP_MULTICAST_IF:
		{
			ip4_addr_t if_addr;

			if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_UDP) {
				async_netconn_sock_set_errno(sock, ENOPROTOOPT);
				return -1;
			}

			if (optlen < sizeof(struct in_addr)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.udp == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			inet_addr_to_ipaddr(&if_addr, (const struct in_addr *)optval);
			udp_set_multicast_netif_addr(sock->conn->pcb.udp, &if_addr);
		}
			break;
		case IP_MULTICAST_LOOP:
			if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_UDP) {
				async_netconn_sock_set_errno(sock, ENOPROTOOPT);
				return -1;
			}

			if (optlen < sizeof(uint8_t)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.udp == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (*(const uint8_t*)optval) {
				udp_setflags(sock->conn->pcb.udp, udp_flags(sock->conn->pcb.udp) | UDP_FLAGS_MULTICAST_LOOP);
			} else {
				udp_setflags(sock->conn->pcb.udp, udp_flags(sock->conn->pcb.udp) & ~UDP_FLAGS_MULTICAST_LOOP);
			}
			break;
#endif /* LWIP_MULTICAST_TX_OPTIONS */
#if LWIP_IGMP
		case IP_ADD_MEMBERSHIP:
		case IP_DROP_MEMBERSHIP:
		{
			err_t igmp_err;
			const struct ip_mreq *imr = (const struct ip_mreq *)optval;
			ip4_addr_t if_addr;
			ip4_addr_t multi_addr;

			if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_UDP) {
				async_netconn_sock_set_errno(sock, ENOPROTOOPT);
				return -1;
			}

			if (optlen < sizeof(struct ip_mreq)) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			if (sock->conn->pcb.udp == NULL) {
				async_netconn_sock_set_errno(sock, EINVAL);
				return -1;
			}

			inet_addr_to_ipaddr(&if_addr, &imr->imr_interface);
			inet_addr_to_ipaddr(&multi_addr, &imr->imr_multiaddr);
			if (optname == IP_ADD_MEMBERSHIP) {
				if (!async_netconn_socket_register_membership(s, &if_addr, &multi_addr)) {
					/* cannot track membership (out of memory) */
					async_netconn_sock_set_errno(sock, ENOMEM);
					return -1;
				} else {
					igmp_err = igmp_joingroup(&if_addr, &multi_addr);
				}
			} else {
				igmp_err = igmp_leavegroup(&if_addr, &multi_addr);
				async_netconn_socket_unregister_membership(s, &if_addr, &multi_addr);
			}

			if (igmp_err != ERR_OK) {
				async_netconn_sock_set_errno(sock, EADDRNOTAVAIL);
				return -1;
			}
		}
			break;
#endif /* LWIP_IGMP */
		default:
			async_netconn_sock_set_errno(sock, ENOPROTOOPT);
			return -1;
		}
		break;

	case IPPROTO_TCP:
		if (NETCONNTYPE_GROUP(netconn_type((sock)->conn)) != NETCONN_TCP) {
			async_netconn_sock_set_errno(sock, ENOPROTOOPT);
			return -1;
		}

		if (optlen < sizeof(int32_t)) {
			async_netconn_sock_set_errno(sock, EINVAL);
			return -1;
		}

		if (sock->conn->pcb.tcp == NULL) {
			async_netconn_sock_set_errno(sock, EINVAL);
			return -1;
		}

		if (sock->conn->pcb.tcp->state == LISTEN) {
			async_netconn_sock_set_errno(sock, EINVAL);
			return -1;
		}

		switch (optname) {
		case TCP_NODELAY:
			if (*(const int32_t *)optval) {
				tcp_nagle_disable(sock->conn->pcb.tcp);
			} else {
				tcp_nagle_enable(sock->conn->pcb.tcp);
			}
			break;
		default:
			async_netconn_sock_set_errno(sock, ENOPROTOOPT);
			return -1;
		}
		break;

	default:
		async_netconn_sock_set_errno(sock, ENOPROTOOPT);
		return -1;
	}

	return 0;
}

int32_t async_netconn_ioctl(int32_t s, long cmd, void *argp) {
	struct async_netconn_sock *sock = async_netconn_get_socket(s);
	uint8_t val;
#if LWIP_SO_RCVBUF
	uint16_t buflen = 0;
	int32_t recv_avail;
#endif /* LWIP_SO_RCVBUF */

	if (!sock) {
		return -1;
	}

	switch (cmd) {
#if LWIP_SO_RCVBUF || LWIP_FIONREAD_LINUXMODE
	case (long)FIONREAD:
		if (!argp) {
			async_netconn_sock_set_errno(sock, EINVAL);
			return -1;
		}
#if LWIP_FIONREAD_LINUXMODE
		if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) != NETCONN_TCP) {
			struct pbuf *p;
			if (sock->lastdata) {
				p = ((struct netbuf *)sock->lastdata)->p;
				*((int32_t *)argp) = p->tot_len - sock->lastoffset;
			} else {
				struct netbuf *rxbuf;
				err_t err;
				if (sock->rcvevent <= 0) {
					*((int32_t *)argp) = 0;
				} else {
					err = netconn_recv(sock->conn, &rxbuf);
					if (err != ERR_OK) {
						*((int32_t *)argp) = 0;
					} else {
						sock->lastdata = rxbuf;
						sock->lastoffset = 0;
						*((int32_t *)argp) = rxbuf->p->tot_len;
					}
				}
			}
			return 0;
		}
#endif /* LWIP_FIONREAD_LINUXMODE */

#if LWIP_SO_RCVBUF
		/* we come here if either LWIP_FIONREAD_LINUXMODE==0 or this is a TCP socket */
		async_netconn_lock();
		recv_avail = sock->conn->recv_avail;
		async_netconn_unlock();
		if (recv_avail < 0) {
			recv_avail = 0;
		}
		*((int32_t *)argp) = recv_avail;

		/* Check if there is data left from the last recv operation. /maq 041215 */
		if (sock->lastdata) {
			struct pbuf *p = (struct pbuf *)sock->lastdata;
			if (NETCONNTYPE_GROUP(netconn_type(sock->conn)) != NETCONN_TCP) {
				p = ((struct netbuf *)p)->p;
			}
			buflen = p->tot_len;
			buflen -= sock->lastoffset;

			*((int32_t *)argp) += buflen;
		}

		ASYNC_NETCONN_DEBUG("async_netconn_ioctl(%d, FIONREAD, %p) = %d\n", s, argp, *((uint16_t*)argp));
		async_netconn_sock_set_errno(sock, 0);
		return 0;
#else /* LWIP_SO_RCVBUF */
		break;
#endif /* LWIP_SO_RCVBUF */
#endif /* LWIP_SO_RCVBUF || LWIP_FIONREAD_LINUXMODE */

	case (long)FIONBIO:
		val = 0;
		if (argp && *(uint32_t*)argp) {
			val = 1;
		}
		netconn_set_nonblocking(sock->conn, val);
		ASYNC_NETCONN_DEBUG("async_netconn_ioctl(%d, FIONBIO, %d)\n", s, val);
		async_netconn_sock_set_errno(sock, 0);
		return 0;

	default:
		break;
	} /* switch (cmd) */

	ASYNC_NETCONN_DEBUG("async_netconn_ioctl(%d, UNIMPL: 0x%lx, %p)\n", s, cmd, argp);
	async_netconn_sock_set_errno(sock, ENOSYS); /* not yet implemented */
	return -1;
}

int32_t async_netconn_fcntl(int32_t s, int32_t cmd, int32_t val) {
	struct async_netconn_sock *sock = async_netconn_get_socket(s);
	int32_t ret = -1;

	ASYNC_NETCONN_DEBUG("async_netconn_fcntl(%d, cmd=%d val=%d)\n", s, cmd, val);

	if (!sock) {
		return -1;
	}

	switch (cmd) {
	case F_GETFL:
		ret = netconn_is_nonblocking(sock->conn) ? O_NONBLOCK : 0;
		async_netconn_sock_set_errno(sock, 0);
		break;
	case F_SETFL:
		if ((val & ~O_NONBLOCK) == 0) {
			/* only O_NONBLOCK, all other bits are zero */
			netconn_set_nonblocking(sock->conn, val & O_NONBLOCK);
			ret = 0;
			async_netconn_sock_set_errno(sock, 0);
		} else {
			async_netconn_sock_set_errno(sock, ENOSYS); /* not yet implemented */
		}
		break;
	default:
		ASYNC_NETCONN_DEBUG("async_netconn_fcntl(%d, UNIMPL: %d, %d)\n", s, cmd, val);
		async_netconn_sock_set_errno(sock, ENOSYS); /* not yet implemented */
		break;
	}
	return ret;
}

int32_t async_netconn_check_events(int32_t s, uint8_t read_events, uint8_t write_events) {
	struct async_netconn_sock *sock;
	int32_t ret = 0;

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	if (sock->err) {
		return -1;
	}

	if (read_events && ((sock->rcvevent > 0) || (sock->lastdata != NULL))) {
		ret = 1;
	}

	if (write_events && (sock->sendevent != 0)) {
		ret = 1;
	}

	ASYNC_NETCONN_DEBUG("async_netconn_check_events(%d, read_events=%d write_events=%d ret=%d)\n", s, read_events, write_events, ret);

	return ret;
}

void async_netconn_init_socket_timeout_cache(void) {
	for (int32_t i = 0; i < (sizeof(async_netconn_sockets_timeout_cache) / sizeof(async_netconn_sockets_timeout_cache[0])); i++) {
		async_netconn_rem_socket_timeout_from_cache(i);
	}
}

int32_t async_netconn_get_socket_timeout_from_cache(int32_t s) {
	return async_netconn_sockets_timeout_cache[s].timeout;
}

int64_t async_netconn_get_socket_absolute_timeout_from_cache(int32_t s) {
	return async_netconn_sockets_timeout_cache[s].absolute_timeout;
}

void async_netconn_set_socket_timeout_in_cache(int32_t s, int32_t timeout) {
	async_netconn_sockets_timeout_cache[s].timeout = timeout;
}

void async_netconn_set_socket_absolute_timeout_in_cache(int32_t s, int64_t absolute_timeout) {
	async_netconn_sockets_timeout_cache[s].absolute_timeout = absolute_timeout;
}

void async_netconn_rem_socket_timeout_from_cache(int32_t s) {
	async_netconn_sockets_timeout_cache[s].timeout = -1;
	async_netconn_sockets_timeout_cache[s].absolute_timeout = -1;
}

int32_t async_netconn_errno(int32_t s) {
	struct async_netconn_sock *sock;

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	return sock->err;
}

static void async_netconn_event_callback(struct netconn *conn, enum netconn_evt evt, uint16_t len) {
	int32_t s;
	struct async_netconn_sock *sock;
	bool try_notify_events = false;

	/* Get socket */
	if (conn) {
		s = conn->socket;
		if (s < 0) {
			/*
			 * Data comes in right away after an accept, even though
			 * the server task might not have created a new socket yet.
			 * Just count down (or up) if that's the case and we
			 * will use the data later. Note that only receive events
			 * can happen before the new socket is set up.
			 */
			async_netconn_lock();
			if (conn->socket < 0) {
				if (evt == NETCONN_EVT_RCVPLUS) {
					conn->socket--;
				}
				async_netconn_unlock();
				ASYNC_NETCONN_DEBUG("async_netconn_event_callback(%d)\n", s);
				return;
			}
			s = conn->socket;
			async_netconn_unlock();
		}

		sock = async_netconn_get_socket(s);
		if (!sock) {
			ASYNC_NETCONN_DEBUG("async_netconn_event_callback(sock=0)\n");
			return;
		}
	} else {
		ASYNC_NETCONN_DEBUG("async_netconn_event_callback(conn=0)\n");
		return;
	}

	async_netconn_lock();

	/* Set event as required */
	switch (evt) {
	case NETCONN_EVT_RCVPLUS:
		/* In case there was no recv data available, but now it is, notify the async_netconn */
		if (!sock->rcvevent) {
			try_notify_events = true;
		}
		sock->rcvevent++;
		ASYNC_NETCONN_DEBUG("async_netconn_event_callback(%d, evt=%s sock->rcvevent=%d)\n", s, "NETCONN_EVT_RCVPLUS", sock->rcvevent);
		break;
	case NETCONN_EVT_RCVMINUS:
		sock->rcvevent--;
		ASYNC_NETCONN_DEBUG("async_netconn_event_callback(%d, evt=%s sock->rcvevent=%d)\n", s, "NETCONN_EVT_RCVMINUS", sock->rcvevent);
		break;
	case NETCONN_EVT_SENDPLUS:
		/* In case there was no send buffer available, but now it is, notify the async_netconn */
		if (!sock->sendevent) {
			try_notify_events = true;
		}
		sock->sendevent = 1;
		ASYNC_NETCONN_DEBUG("async_netconn_event_callback(%d, evt=%s sock->sendevent=%d)\n", s, "NETCONN_EVT_SENDPLUS", sock->sendevent);
		break;
	case NETCONN_EVT_SENDMINUS:
		sock->sendevent = 0;
		ASYNC_NETCONN_DEBUG("async_netconn_event_callback(%d, evt=%s sock->sendevent=%d)\n", s, "NETCONN_EVT_SENDMINUS", sock->sendevent);
		break;
	case NETCONN_EVT_ERROR:
		/* Always notify the async_netconn in case of socket errors */
		try_notify_events = true;
		sock->errevent = 1;
		ASYNC_NETCONN_DEBUG("async_netconn_event_callback(%d, evt=%s sock->errevent=%d)\n", s, "NETCONN_EVT_ERROR", sock->errevent);
		break;
	default:
		ASYNC_NETCONN_ASSERT("unknown event", 0);
		break;
	}

	if (try_notify_events) {
		async_netconn_try_notify_events();
	}

	async_netconn_unlock();
}

static int32_t async_netconn_alloc_socket(struct netconn *newconn, int32_t accepted) {
	int32_t i;

	/* allocate a new socket identifier */
	for (i = 0; i < (sizeof(async_netconn_sockets) / sizeof(async_netconn_sockets[0])); ++i) {
		/* No need to protect the socket array, as there is only 1 task that can allocate/free socket resources */
		//async_netconn_lock();
		if (!async_netconn_sockets[i].conn) {
			async_netconn_sockets[i].conn       = newconn;
			/*
			 * The socket is not yet known to anyone, so no need to protect
			 * after having marked it as used.
			 */
			//async_netconn_unlock();
			async_netconn_sockets[i].lastdata   = NULL;
			async_netconn_sockets[i].lastoffset = 0;
			async_netconn_sockets[i].rcvevent   = 0;
			/*
			 * TCP sendbuf is empty, but the socket is not yet writable until connected
			 * (unless it has been created by async_netconn_accept()).
			 */
			async_netconn_sockets[i].sendevent  = (NETCONNTYPE_GROUP(newconn->type) == NETCONN_TCP ? (accepted != 0) : 1);
			async_netconn_sockets[i].errevent   = 0;
			async_netconn_sockets[i].err        = 0;
			return i;
		}
		/* No need to protect the socket array, as there is only 1 task that can allocate/free socket resources */
		//async_netconn_unlock();
	}

	return -1;
}

static struct async_netconn_sock * async_netconn_get_socket(int32_t s) {
	struct async_netconn_sock *sock;

	if ((s < 0) || (s >= (sizeof(async_netconn_sockets) / sizeof(async_netconn_sockets[0])))) {
		ASYNC_NETCONN_DEBUG("async_netconn_get_socket(%d): invalid\n", s);
		async_netconn_set_errno(EBADF);
		return NULL;
	}

	sock = &async_netconn_sockets[s];

	if (!sock->conn) {
		ASYNC_NETCONN_DEBUG("async_netconn_get_socket(%d): not active\n", s);
		async_netconn_set_errno(EBADF);
		return NULL;
	}

	return sock;
}

static void async_netconn_free_socket(struct async_netconn_sock *sock, int32_t is_tcp) {
	void *lastdata;

	lastdata         = sock->lastdata;
	sock->lastdata   = NULL;
	sock->lastoffset = 0;
	sock->err        = 0;

	/* No need to protect the socket array, as there is only 1 task that can allocate/free socket resources */
	//async_netconn_lock();
	sock->conn = NULL;
	/* No need to protect the socket array, as there is only 1 task that can allocate/free socket resources */
	//async_netconn_unlock();

	if (lastdata != NULL) {
		if (is_tcp) {
			pbuf_free((struct pbuf *)lastdata);
		} else {
			netbuf_delete((struct netbuf *)lastdata);
		}
	}
}

#if LWIP_IPV4 && LWIP_IPV6
static void sockaddr_to_ipaddr_port(const struct sockaddr* sockaddr, ip_addr_t* ipaddr, uint16_t* port) {
	if ((sockaddr->sa_family) == AF_INET6) {
		SOCKADDR6_TO_IP6ADDR_PORT((const struct sockaddr_in6*)(const void*)(sockaddr), ipaddr, *port);
		ipaddr->type = IPADDR_TYPE_V6;
	} else {
		SOCKADDR4_TO_IP4ADDR_PORT((const struct sockaddr_in*)(const void*)(sockaddr), ipaddr, *port);
		ipaddr->type = IPADDR_TYPE_V4;
	}
}
#endif /* LWIP_IPV4 && LWIP_IPV6 */

#if LWIP_IGMP
static int32_t async_netconn_socket_register_membership(int32_t s, const ip4_addr_t *if_addr, const ip4_addr_t *multi_addr)
{
	struct async_netconn_sock *sock = async_netconn_get_socket(s);
	int32_t i;

	if (!sock) {
		return 0;
	}

	for (i = 0; i < (sizeof(socket_ipv4_multicast_memberships) / sizeof(socket_ipv4_multicast_memberships[0])); i++) {
		if (socket_ipv4_multicast_memberships[i].sock == NULL) {
			socket_ipv4_multicast_memberships[i].sock = sock;
			ip4_addr_copy(socket_ipv4_multicast_memberships[i].if_addr, *if_addr);
			ip4_addr_copy(socket_ipv4_multicast_memberships[i].multi_addr, *multi_addr);
			return 1;
		}
	}
	return 0;
}

static void async_netconn_socket_unregister_membership(int32_t s, const ip4_addr_t *if_addr, const ip4_addr_t *multi_addr) {
	struct async_netconn_sock *sock = async_netconn_get_socket(s);
	int32_t i;

	if (!sock) {
		return;
	}

	for (i = 0; i < (sizeof(socket_ipv4_multicast_memberships) / sizeof(socket_ipv4_multicast_memberships[0])); i++) {
		if ((socket_ipv4_multicast_memberships[i].sock == sock) &&
				ip4_addr_cmp(&socket_ipv4_multicast_memberships[i].if_addr, if_addr) &&
				ip4_addr_cmp(&socket_ipv4_multicast_memberships[i].multi_addr, multi_addr)) {
			socket_ipv4_multicast_memberships[i].sock = NULL;
			ip4_addr_set_zero(&socket_ipv4_multicast_memberships[i].if_addr);
			ip4_addr_set_zero(&socket_ipv4_multicast_memberships[i].multi_addr);
			return;
		}
	}
}

static void async_netconn_socket_drop_registered_memberships(int32_t s) {
	struct async_netconn_sock *sock = async_netconn_get_socket(s);
	int32_t i;

	if (!sock) {
		return;
	}

	for (i = 0; i < (sizeof(socket_ipv4_multicast_memberships) / sizeof(socket_ipv4_multicast_memberships[0])); i++) {
		if (socket_ipv4_multicast_memberships[i].sock == sock) {
			ip_addr_t multi_addr, if_addr;
			ip_addr_copy_from_ip4(multi_addr, socket_ipv4_multicast_memberships[i].multi_addr);
			ip_addr_copy_from_ip4(if_addr, socket_ipv4_multicast_memberships[i].if_addr);
			socket_ipv4_multicast_memberships[i].sock = NULL;
			ip4_addr_set_zero(&socket_ipv4_multicast_memberships[i].if_addr);
			ip4_addr_set_zero(&socket_ipv4_multicast_memberships[i].multi_addr);

			netconn_join_leave_group(sock->conn, &multi_addr, &if_addr, NETCONN_LEAVE);
		}
	}
}
#endif /* LWIP_IGMP */

static int32_t async_netconn_getaddrname(int32_t s, struct sockaddr *name, socklen_t *namelen, uint8_t local) {
	struct async_netconn_sock *sock;
	union sockaddr_aligned saddr;
	ip_addr_t naddr;
	uint16_t port;
	err_t err;

	sock = async_netconn_get_socket(s);
	if (!sock) {
		return -1;
	}

	/* get the IP address and port */
	err = netconn_getaddr(sock->conn, &naddr, &port, local);
	if (err != ERR_OK) {
		async_netconn_sock_set_errno(sock, async_netconn_err_to_errno(err));
		return -1;
	}

	IPADDR_PORT_TO_SOCKADDR(&saddr, &naddr, port);

	ASYNC_NETCONN_DEBUG("async_netconn_getaddrname(%d, addr=", s);
	ASYNC_NETCONN_IP_DEBUG(naddr);
	ASYNC_NETCONN_DEBUG(" port=%d)\n", port);

	if (*namelen > saddr.sa.sa_len) {
		*namelen = saddr.sa.sa_len;
	}
	memcpy(name, &saddr, *namelen);

	async_netconn_sock_set_errno(sock, 0);
	return 0;
}

#ifdef __cplusplus
	}
#endif
